/*
   OpenGUI - Drawing & Windowing library

   Copyright (C) 1996,2002  Marian Krivos

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   nezmar@atlas.sk

   _fastgl.h - internal declarations
 */
#ifndef __FASTGL_H_
#define __FASTGL_H_

#ifdef __cplusplus
class WindowItem
{
	public:
	int x, y, w, h;
	int icon_x, icon_y, flag;
	char *name;
	WindowItem *next;
	  WindowItem();
	  WindowItem(int xx, int yy, int ww, int hh, char *s);
	 ~WindowItem();
};

class WindowDatabase
{
	static int xyCounter;
	friend class WindowItem;
	WindowItem *first;
	WindowItem **endpointer;
	char name[64], exe[64];
	  public:
	  WindowDatabase(char *);
	 ~WindowDatabase();
	void Add(int &, int &, int &, int &, char *, int);
	void DatabaseResize(int, int, char *);
	void DatabaseRelocate(int, int, char *);
	void DatabaseSetIcon(int, int, char *);
	void DatabaseGetIcon(int &, int &, char *);
	WindowItem *findxy(char *s);
};

extern Palette ColorsArray[256];

#ifdef	__WINDOW_H
extern Window *oldidw, *idw;
extern Control *oldidm, *idm;

#endif

extern "C"
{

#endif

#if defined(__linux__) || defined(__rtems__)
#ifndef VGA_H

#define TEXT 	     0			/* Compatible with VGAlib v1.2 */
#define G320x200x256 5
#define G640x480x256 10
#define G800x600x256 11
#define G1024x768x256 12
#define G1280x1024x256 13		/* Additional modes. */
#define G1600x1200x256 45

#define G320x200x16M32 33		/* 32-bit per pixel modes. */
#define G640x480x16M32 34
#define G800x600x16M32 35
#define G1024x768x16M32 36
#define G1280x1024x16M32 37
#define G1600x1200x16M32 49

#define G320x200x32K 14
#define G320x200x64K 15
#define G640x480x32K 17
#define G640x480x64K 18
#define G800x600x32K 20
#define G800x600x64K 21
#define G1024x768x32K 23
#define G1024x768x64K 24
#define G1280x1024x32K 26
#define G1280x1024x64K 27
#define G1600x1200x32K 46
#define G1600x1200x64K 47

#define MOUSE_LEFTBUTTON 4
#define MOUSE_RIGHTBUTTON 1
#define MOUSE_DEFAULTSAMPLERATE 150

#if !defined(__rtems__)
	typedef struct
	{
		int width;
		int height;
		int bytesperpixel;
		int colors;
		int linewidth;			/* scanline width in bytes */
		int maxlogicalwidth;	/* maximum logical scanline width */
		int startaddressrange;	/* changeable bits set */
		int maxpixels;			/* video memory / bytesperpixel */
		int haveblit;			/* mask of blit functions available */
		int flags;				/* other flags */

		/* Extended fields: */

		int chiptype;			/* Chiptype detected */
		int memory;				/* videomemory in KB */
		int linewidth_unit;		/* Use only a multiple of this as parameter for set_logicalwidth and
								   set_displaystart */
		char *linear_aperture;	/* points to mmap secondary mem aperture of card (NULL if unavailable) */
		int aperture_size;		/* size of aperture in KB if size>=videomemory. 0 if unavail */
		void (*set_aperture_page) (int page);
		/* if aperture_size<videomemory select a memory page */
		void *extensions;		/* points to copy of eeprom for mach32 */
		/* depends from actual driver/chiptype.. etc. */
	}
	vga_modeinfo;
#endif  /* __rtems__ */
#endif
#endif  /* __linux & __rtems */

#define FGTEXTBUFSZ		16384

#pragma pack(1)
	typedef struct
	{
	  int fontchar_w;
	  int fontchar_h;
	  unsigned long fontchar_fp;
	} font_t;

	typedef struct
	{
	  char font_name[256];
	  int font_maxcharw;
	  int font_maxcharh;
	  int font_spacesize;
	  font_t ftable[256];
	} fonthdr;

#pragma pack()

	void XWaitRetrace(void);
	void XWaitPeriod(void);
	int	__MouseDraw(int x, int y,	int	a);
	// proportional from file
	int register_font_var(variable_record *source, int size);
	int register_font_image(fonthdr * hdr, char *img);

	typedef struct
	{
		int x, xlast;
		int y, ylast;
		int dx, dy;
		int xmajor;
		int xstep;
		int error;
	}
	polyedge;

	typedef enum
	{
		inactive,
		active,
		passed
	}
	edgestat;

	typedef struct
	{
		edgestat status;
		polyedge e;
	}
	edge;

	typedef struct _scan
	{
		struct _scan *next;
		int x1, x2;
	}
	scan;

//	typedef void (*PixelFillFunc) (int x, int y);
	typedef void (*LineFillFunc) (int x, int y, int dx, int dy, FGPixel ink, unsigned ppop);
	typedef void (*ScanFillFunc) (int x, int y, int w, FGPixel ink, unsigned ppop);

	typedef struct
	{
		LineFillFunc line;
		ScanFillFunc scan;
		int gc_xoffset;
		int gc_yoffset;
		int gc_ycliplo;
		int gc_xcliplo;
		int gc_ycliphi;
		int gc_xcliphi;
	}
	PolygonStruct;

	extern edge edge_array[128+2];
	extern scan scan_array[128+8];
	
	enum Drivers
	{
		FG_VESA, FG_CIRRUS, FG_TRIDENT, FG_S3, FG_S3V2,
		FG_MX, FG_TSENG3, FG_ATI, FG_TSENG4, FG_CHIPS,
		FG_WDIGITAL, FG_BANSHEE, FG_NVIDIA, FG_MATROX, FG_PERMEDIA,
		FG_INTEL740, FG_INTEL810, FG_RENDITION, FG_SIS,
		FG_LAST, FG_KERNEL_ACCEL
	};

struct FGData
{
	unsigned long
		CRTC,
		granularity,
		videobasesegment,
		VideoSelector,
		bpp,
		lfb,
		ppop,
		mmx_state,
		nofb,
		nodga,
		nox11,
		__fgl_mmx;
	int
		clip_x_min,
		clip_y_min,
		clip_x_max,
		clip_y_max,
		driver,
		MOVEABLE,
		verbose,
		synch,
		fulldrag,
		allBitmapCounter;
	FGPixel
		*videobase,
		*font,
		transp_color,
		_CurrBkColor,
		_CurrColor;
};

//extern struct FGData fgdata[];

struct PCICardInfo
{
	unsigned short vendor, device;
	void *MMIO_ADDR_DETECTED;
 	void *LFB_ADDR_DETECTED;
	unsigned LFB_SIZE, MMIO_SIZE;
};

extern struct PCICardInfo CardInfo;
extern FGPixel fgimagebuffer[FGTEXTBUFSZ];

//#if defined(__linux__) || defined(__rtems__) || ( defined(__GNUC__) && defined(__QNX__))
#if defined(__GNUC__)
	extern int CRTC asm("_CRTC");
	extern int granularity asm("_granularity");
	extern FGPixel *videobase asm("_videobase");
	extern unsigned videobasesegment asm("_videobasesegment");
	extern unsigned VideoSelector asm("_VideoSelector");
	void set_palette_map(int cnt, int ind, int *pal) asm("_set_palet1");
	extern int ppop asm("_ppop"), lfb asm("_lfb");
	extern int mmx_state asm("_mmx_state"), __fgl_mmx asm("___fgl_mmx");
	extern int _CurrBkColor asm("__CurrBkColor"), _CurrColor asm("__CurrColor");
	extern int clip_x_min asm("_clip_x_min"), clip_y_min asm("_clip_y_min"),
	  clip_x_max asm("_clip_x_max"), clip_y_max asm("_clip_y_max");
	extern int driver asm("_driver");
	extern FGPixel *font asm("_font");
	extern short _setvideomode(short) asm("__setvideomode");

	extern void __fill_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop) asm("___fill_box");
	extern void __draw_line(int x, int y, int w, int h, FGPixel ink, unsigned ppop) asm("___draw_line");
	extern void __draw_point(int x, int y, FGPixel ink, unsigned ppop) asm("___draw_point");

	extern int lfb;
	extern int bpp asm("_bpp");
	extern FGPixel transp_color asm("_transp_color");
	extern int X_width asm("_X_width"), Y_width asm("_Y_width");	/* maximalne rozlisenie */
	extern int X_virtual asm("_X_virtual"), Y_virtual asm("_Y_virtual");
#else
	extern int CRTC;
	extern int granularity;
	extern FGPixel *videobase;
	extern unsigned videobasesegment;
	extern unsigned VideoSelector;
	void set_palette_map(int cnt, int ind, int *pal);
	extern int ppop;
	extern int mmx_state, __fgl_mmx;
	extern int _CurrBkColor, _CurrColor;
	extern int clip_x_min, clip_y_min, clip_x_max, clip_y_max;
	extern int driver;
	extern FGPixel *font;
	extern short _setvideomode(short);

	extern void __draw_line(int x, int y, int w, int h, FGPixel ink, unsigned ppop);
	extern void __fill_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop);
	extern void __draw_point(int x, int y, FGPixel ink, unsigned ppop);

	extern int lfb;
	extern int X_max, Y_max;	/* maximalne rozlisenie */
	extern int bpp;
	extern FGPixel transp_color;
	extern int X_width, Y_width;	/* maximalne rozlisenie */
	extern int X_virtual, Y_virtual;
#endif

	extern const unsigned int _default_palette[256];
	extern FGPixel _internal_bitmap0[];
	extern unsigned long _fgl_palette[256];
	extern int __force_x, __force_y;
	extern int mouseflag, linear_flag, nofb, noaccel, nodga, nox11;
	extern int force_dac8;
	extern int force_ps2;

	extern unsigned char miro0406[], miro0808[], miro0816[], miro1220[], miro1625[],
	  miro2034[];

	void CloseUserInput(void);
	extern int detect_video(int v);
	extern int peek(short seg, unsigned off);
	extern void poke(short seg, unsigned off, unsigned int value);
	extern int MOVEABLE;
	extern int verbose, synch;
	extern int allBitmapCounter;
	extern int fulldrag;
	extern FGPixel m_save[24*24],
	  c_mask[196], d_mask[196], e_mask[196], f_mask[196];
	extern unsigned int msss16[715*8], msss16b[814*8], msss13[407*8];

	void modify_point(FGPixel * ptr);
	void do_ellipse(int x, int y, int rx, int ry, void (*proc) (int, int, FGPixel,unsigned), FGPixel ink, unsigned ppop);
	void do_fill_ellipse(int x, int y, int rx, int ry, void (*draw_hline) (int, int, int, FGPixel,unsigned), FGPixel ink, unsigned ppop);
	void do_arc(int x, int y, double ang1, double ang2, int r, void (*proc) (int, int, FGPixel,unsigned), FGPixel ink, unsigned ppop);
	void do_polygon(int n, int pt[][2], PolygonStruct * current);
	void do_fill_polygon(int n, int pt[][2], PolygonStruct * current);
	void line(plotter_t, int a1, int b1, int a2, int b2, Pattern * pat);
	void GetPaletteEntry(char *rc, char *gc, char *bc, int i);
	void CreatePaletteEntry(int rc, int gc, int bc, int idx);

#if defined(__GNUC__)
//#if defined(__linux__) || defined(__rtems__) || ( defined(__GNUC__) && defined(__QNX__))
	void L1VideoToRam8(FGPixel *, FGPixel *, int, int, int, int) asm("_L1VideoToRam8");
	void L1RamToVideo2(FGPixel *, FGPixel *, int, int, int, int) asm("_L1RamToVideo2");
	void L1RamToVideo8(FGPixel *, FGPixel *, int, int, int, int) asm("_L1RamToVideo8");
	void L1RamToVideoA(FGPixel *, FGPixel *, int, int, int, int, int) asm("_L1RamToVideoA");
	void L1RamToRamA(FGPixel *, FGPixel *, int, int, int, int, int) asm("_L1RamToRamA");
	void L1RamToRam(FGPixel *, FGPixel *, int, int, int, int) asm("_L1RamToRam");
	void L1RamToRamd(FGPixel *, FGPixel *, int, int, int, int) asm("_L1RamToRamd");
	void L1RamToRamPpop(FGPixel *, FGPixel *, int, int, int, int) asm("_L1RamToRamPpop");
	void L1Box(FGPixel *, int, int, int) asm("_L1Box");
	void L1BoxX(FGPixel *, int, int, int) asm("_L1BoxX");
	void CharOutClip(FGPixel *, FGPixel *, int offset, int w, int h) asm("_CharOutClip");
#else
	void L1VideoToRam8(FGPixel *, FGPixel *, int, int, int, int);
	void L1RamToVideo2(FGPixel *, FGPixel *, int, int, int, int);
	void L1RamToVideo8(FGPixel *, FGPixel *, int, int, int, int);
	void L1RamToVideoA(FGPixel *, FGPixel *, int, int, int, int, int);
	void L1RamToRamA(FGPixel *, FGPixel *, int, int, int, int, int);
	void L1RamToRam(FGPixel *, FGPixel *, int, int, int, int);
	void L1RamToRamPpop(FGPixel *, FGPixel *, int, int, int, int);
	void L1RamToRamd(FGPixel *, FGPixel *, int, int, int, int);
	void L1Box(FGPixel *, int, int, int);
	void L1BoxX(FGPixel *, int, int, int);
	void CharOutClip(FGPixel *, FGPixel *, int offset, int w, int h);
#endif

#define PICTOSIZE		20
#define TITLEH			20
#define BASEBORDER		4
#define MENUH			20
#define STATUSBARH		19

#define BORD1    		CScheme->wnd_bord1
#define BORD2			CScheme->wnd_bord2
#define BORD3			CScheme->wnd_bord3

#define CTITLE   		CScheme->inactive_title		// color of unactive title
#define CTITLEACTIVE 	CScheme->active_title	// color of active title
#define MENUBG	 		CScheme->menu_back

#ifdef __WATCOMC__
#pragma aux CRTC "_*"
#pragma aux set_palette_map "_*"
#pragma aux font "_*"
#pragma aux _setvideomode "_*"
#pragma aux granularity "_*"
#pragma aux lfb "_*"
#pragma aux bpp "_*"
#pragma aux transp_color "_*"
#pragma aux videobase "_*"
#pragma aux videobasesegment "_*"
#pragma aux VideoSelector "_*"
#pragma aux peek "_*"
#pragma aux poke "_*"
#pragma aux driver "_*"
#pragma aux __fgl_mmx "_*"
#pragma aux ppop "_*"
#pragma aux mmx_state "_*"
#pragma aux __fill_box "_*"
#pragma aux __draw_line "_*"
#pragma aux __draw_point "_*"
#pragma aux __dpmi_int "_*"
#pragma aux _CurrBkColor "_*"
#pragma aux _CurrColor "_*"
#pragma aux CharOutClip "_*"
#pragma aux clip_x_min "_*"
#pragma aux clip_y_min "_*"
#pragma aux clip_x_max "_*"
#pragma aux clip_y_max "_*"
#pragma aux L1Box "_*"
#pragma aux L1BoxX "_*"
#pragma aux L1RamToRam "_*"
#pragma aux L1RamToRamd "_*"
#pragma aux L1RamToRamPpop "_*"
#pragma aux L1RamToVideo2 "_*"
#pragma aux L1RamToVideo8 "_*"
#pragma aux L1RamToVideoA "_*"
#pragma aux L1RamToRamA "_*"
#pragma aux L1VideoToRam8 "_*"
#endif

#ifdef __cplusplus
}
#endif

#if defined(__linux__) || defined(__rtems__) 
#include "linkeyb.h"
extern void (*ptr_vga_setpage)(int);
#endif

#endif

