/*
   OpenGUI - Drawing & Windowing library

   Copyright (C) 1996,2003  Marian Krivos

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   nezmar@atlas.sk

   base.cc - base graphics support routines   
 */

#include <stdarg.h>
#include "config.h"
#include "base.h"
#include "_fastgl.h"

#include <sys/timeb.h>

typedef struct
{
	int width, height;
	variable_record *var; // null or ptr
	FGPixel *fontimg;
}
FGFont;

static FGPixel __p1[2] =
{0, 1};
static FGPixel __p2[9] =
{1, 0, 0, 1, 1, 1, 1, 0, 0};
static FGPixel __p3[15] =
{0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1};
static FGPixel __p4[8] =
{0, 0, 0, 0, 1, 1, 1, 1};

Pattern PatternDot(__p1, 2, 1), PatternSlashDot(__p2, 9, 1), PatternSlashDotSlash(__p3, 15, 1),
		PatternSlash(__p4, 8, 1);

// draw clipped to screen line
shapper_t  vector_draw_line;

// draw clipped to screen recty - HW helper = 0;
clipper_t   vector_clip_rect = 0;

// draw clipped to screen box
shapper_t  vector_fill_box;

// draw clipped to screen plot
plotter_t  vector_draw_point;

// set 8 bit palette (shift needed for 6-bit palette)
paletter_t vector_palette;

// bitblit copy
blitter_t   vector_blit_copy = L1RamToVideo8;
blitter_t   vector_blit_op   = L1RamToVideo2;
blitter_a_t vector_blit_a    = L1RamToVideoA;

// mouse cursor drawer
mouser_t   vector_mouse_cursor=__MouseDraw;

// there are current system colors
int _CurrColor, _CurrBkColor;

// clipping co-ordinates
int clip_x_min, clip_y_min, clip_x_max, clip_y_max;

// non-zero if lfb is enabled
int lfb;

// from Vesa info, for paging routines
int granularity;

// offset VRAM in VideoSegment (non-zero in LINUX only)
FGPixel *videobase;

// default 0xA000 for dos 64kb window
unsigned videobasesegment = 0xa000;

// selector for VRAM
unsigned VideoSelector = 0xFFFFFFFF;

// linear address of mapped FRAMEBUFFER (dos only)
unsigned VideoLinear = 0;

// mmx variables
int mmx_state = 0, __fgl_mmx = 0;

// index to driver table (for paging)
int driver = 0;

// size and number of current font
int fonth, fontw, fontn;

//
int synch;

// ptr to font data (each pixel is = 1 byte)
FGPixel *font;

//
variable_record *fontvar;

// current draving mode (_GSET, _GAND ...)
int ppop;

// size of the screen
int X_width, Y_width, X_virtual, Y_virtual;

// internal for polygons
static PolygonStruct ps;

// this is a temp buffer for text drawing
FGPixel fgimagebuffer[FGTEXTBUFSZ];

// TRUE if linear framebuffer is used
int linear_flag = 1;

// current palette [ABGR]
unsigned int long _fgl_palette[256];
Palette ColorsArray[256];

// current bytes per pixel - for future
int bpp = sizeof(FGPixel);

// 
int CRTC = 0x3D4;

//
FGPixel transp_color;

// fix for fbdev palette
int palette_8 = 0;
int noaccel = 0;

//
char *__fgl_version = FG_VERSION;

static FGFont FGFonts[32];
static int font_counter;
edge edge_array[128+2];
scan scan_array[128+8];

int get_colordepth(void)
{
	return FASTGL_BPP;
}

unsigned int get_ppop(void)
{
	return ppop;
}

const char *GetVer(void)
{
	return __fgl_version;
}

void set_transpcolor(FGPixel c)
{
	transp_color = c;
}

int GetXRes(void) { return X_width; }
int GetYRes(void) { return Y_width; }

		void FGScreen::Resize(int xx, int yy)
		{
			X_width = wwrk = w = xx;
			Y_width = hwrk = h = yy;
			::set_clip_rect(w, h, 0, 0);
			set_font(FONT0816);
			set_colors(0, CWHITE);
			set_ppop(_GSET);
			assert(image==0);
		}

inline int Pattern::next(void)
{
	if (pos >= len)
		pos = 0;
	if (masked)
	{
		return data[pos++];
	}
	_CurrColor = data[pos++];
	return 1;
}

// internal for fill circle in VRAM
static void _vsymetry2(int xs, int ys, int x, int y, FGPixel ink, unsigned ppop)
{
	draw_line(xs + x, ys + y, xs - x, ys + y,ink,ppop);
	draw_line(xs + x, ys - y, xs - x, ys - y,ink,ppop);
	draw_line(xs + y, ys + x, xs - y, ys + x,ink,ppop);
	draw_line(xs + y, ys - x, xs - y, ys - x,ink,ppop);
}

// fill circle in VRAM
void fill_circle(int xs, int ys, int r, FGPixel ink, unsigned ppop)
{
	int x = 0;
	int y = r;
	int p = 3 - 2 * r;

	if ((ys + r) < clip_y_min || (xs + r) < clip_x_min || (ys - r) >= clip_y_max || (xs - r) >= clip_x_max || r < 1)
		return;

	while (x < y)
	{
		_vsymetry2(xs, ys, x, y,ink,ppop);
		if (p < 0)
			p += 4 * (x++) + 6;
		else
			p += 4 * ((x++) - (y--)) + 10;
	}
	if (x == y)
		_vsymetry2(xs, ys, x, y,ink,ppop);
}

// internal for draw circle to screen
static void _vsymetry(int xs, int ys, int x, int y, FGPixel ink, unsigned ppop)
{
	vector_draw_point(xs + x, ys + y,ink,ppop);
	vector_draw_point(xs - x, ys + y,ink,ppop);
	vector_draw_point(xs + x, ys - y,ink,ppop);
	vector_draw_point(xs - x, ys - y,ink,ppop);
	vector_draw_point(xs + y, ys + x,ink,ppop);
	vector_draw_point(xs - y, ys + x,ink,ppop);
	vector_draw_point(xs + y, ys - x,ink,ppop);
	vector_draw_point(xs - y, ys - x,ink,ppop);
}

// draw circle to screen
void draw_circle(int xs, int ys, int r, FGPixel ink, unsigned ppop)
{
	int x = 0;
	int y = r;
	int p = 3 - 2 * r;

	if ((ys + r) < clip_y_min || (xs + r) < clip_x_min || (ys - r) >= clip_y_max || (xs - r) >= clip_x_max || r < 1)
		return;

	while (x < y)
	{
		_vsymetry(xs, ys, x, y,ink,ppop);
		if (p < 0)
			p += 4 * (x++) + 6;
		else
			p += 4 * ((x++) - (y--)) + 10;
	}
	if (x == y)
		_vsymetry(xs, ys, x, y,ink,ppop);
}

// this is very important procedure - its must be fast
// put buffer from RAM to VRAM with redraw mode
void RamToVideo(FGPixel * Image, int x, int y, int xm, int ym, int w, int h, int cx_max, int cy_max, int opcia)
{
	FGPixel *src, *dst;

	if (y < 0 || x < 0 || y >= cy_max || x >= cx_max || w < 1 || h < 1)
		return;
	if ((ym + h) < clip_y_min || (xm + w) < clip_x_min || ym >= clip_y_max || xm >= clip_x_max)
		return;

	if ((x + w) > cx_max)
		w -= ((x + w) - cx_max);
	if ((y + h) > cy_max)
		h -= ((y + h) - cy_max);

	if ((xm + w) > clip_x_max)
		w -= ((xm + w) - clip_x_max);
	if (xm < clip_x_min)
	{
		w += (xm - clip_x_min);
		x += clip_x_min - xm;
		xm = clip_x_min;
	}

	if ((ym + h) > clip_y_max)
	{
		h -= ((ym + h) - clip_y_max);
	}
	if (ym < clip_y_min)
	{
		h += (ym - clip_y_min);
		y += clip_y_min - ym;
		ym = clip_y_min;
	}

	if (w < 1 || h < 1)
		return;

	if (xm < 0)
		xm = 0;
	if (ym < 0)
		ym = 0;
	src = Image + y * cx_max + x;
	dst = (FGPixel *) (ym * X_virtual + xm + videobase);

	switch (opcia & 0xF00)
	{
		case BITBLIT_COPY:
			vector_blit_copy(src, dst, w, h, cx_max - w, X_virtual - w);
			break;
		case BITBLIT_WITH_OPERATOR:
			vector_blit_op(src, dst, w, h, cx_max - w, X_virtual - w);
			break;
		case BITBLIT_WITH_ALPHA:
			vector_blit_a(src, dst, w, h, cx_max - w, X_virtual - w, opcia & 0xFF);
			break;
	}
#ifdef X11_DRIVER
	UpdateRect(xm,ym,xm,ym,w,h);
#endif
}

// this very important procedure - its must be fast
// put buffer from VRAM to RAM with redraw mode
void VideoToRam(FGPixel * Image, int x, int y, int xm, int ym, int w, int h, int cx_max, int cy_max)
{
	FGPixel *src, *dst;

	if ((y + h) < clip_y_min || (x + w) < clip_x_min || y >= clip_y_max || x >= clip_x_max || w < 1 || h < 1)
		return;
	if (ym < 0 || xm < 0 || ym >= cy_max || xm >= cx_max)
		return;

	if ((x + w) > clip_x_max)
		w -= ((x + w) - clip_x_max);
	if (x < clip_x_min)
	{
		w += (x - clip_x_min);
		xm += clip_x_min - x;
		x = clip_x_min;
	}
	if ((y + h) > clip_y_max)
		h -= ((y + h) - clip_y_max);
	if (y < clip_y_min)
	{
		h += (y - clip_y_min);
		ym += clip_y_min - y;
		y = clip_y_min;
	}

	if ((xm + w) > cx_max)
		w -= ((xm + w) - cx_max);
	if ((ym + h) > cy_max)
		h -= ((ym + h) - cy_max);

	if (w < 1 || h < 1)
		return;
	dst = Image + ym * cx_max + xm;
	src = (FGPixel *) (y * X_virtual + x + videobase);
	L1VideoToRam8(src, dst, w, h, cx_max - w, X_virtual - w);
}

// this is a very important procedure - its must be fast
// put buffer from RAM to RAM with redraw mode
// The rectanles must be clipped !!!
void RamToRam(int x, int y, int xmax, int ymax, int xdst, int ydst, int xmaxdst, int ymaxdst, int w, int h, FGPixel * from, FGPixel * to, int alpha)
{
	FGPixel *src, *dst;

	if (alpha & BITBLIT_WITH_CLIPPING)	// clip right now!
	{
		if (w < 1 || h < 1)
			return;
		if (y >= ymax || x >= xmax)
			return;
		if (ydst >= ymaxdst || xdst >= xmaxdst)
			return;
		if (y + h < 0 || x + h < 0)
			return;
		if (ydst + h < 0 || xdst + w < 0)
			return;

		if ((x + w) > xmax)
			w -= ((x + w) - xmax);
		if ((y + h) > ymax)
			h -= ((y + h) - ymax);
		if (xdst < 0)
		{
			w += xdst;
			xdst = 0;
		}
		if (ydst < 0)
		{
			h += ydst;
			ydst = 0;
		}
		if ((xdst + w) > xmaxdst)
			w -= ((xdst + w) - xmaxdst);
		if ((ydst + h) > ymaxdst)
			h -= ((ydst + h) - ymaxdst);
	}
	if (w < 1 || h < 1)
		return;

	if (alpha & BITBLIT_WITH_DECREMENT)
	{
		dst = to + ((ydst+h-1) * xmaxdst + xdst+w-1);
		src = from + ((y+h-1) * xmax + x+w-1);
		L1RamToRamd(src, dst, w, h, xmax, xmaxdst);
	}
	else
	{
		dst = to + ydst * xmaxdst + xdst;
		src = from + y * xmax + x;
		if (alpha & 255)
			L1RamToRamA(src, dst, w, h, xmaxdst - w, xmax - w, alpha & 255);
		else
		{
			if (alpha & BITBLIT_WITH_OPERATOR)
				L1RamToRamPpop(src, dst, w, h, xmaxdst - w, xmax - w);
			else
				L1RamToRam(src, dst, w, h, xmaxdst - w, xmax - w);
		}
	}
}

// draw text directly to VRAM
int text_out(int x, int y, char *txt, FGPixel ink, FGPixel paper, unsigned ppop, int f)
{
	FGPixel *dst, *dst2;
	int tc;
	int alloc=0;

	int	width =	textwidth(txt), ww = fontw;

	if ((y + fonth) < clip_y_min || (x + width) < clip_x_min || y >= clip_y_max || x >= clip_x_max)
		return x;

	unsigned  bytes = (width+ww)*fonth;

	if (bytes < FGTEXTBUFSZ)
		dst	= fgimagebuffer;
	else
	{
		alloc = 1;
		dst = (FGPixel *) malloc(bytes*sizeof(FGPixel));
	}

	dst2 = dst;
	tc = get_bcolor();

	if (ppop == _GTRANSP)
	{
		set_bcolor(transp_color);
	}
#ifndef	NO_MMX
	set_mmx();
#endif
	while (*txt)
	{
		if (fontvar)
		{
			ww = fontvar->w[*(unsigned char *)txt];
		}
		CharOutClip(dst,
			fontvar ? fontvar->off[*(unsigned char *)txt++] + font:
				*(unsigned char *)txt++ * (fontw*fonth) + font,
			width+fontw,
			(ww+3)&~3,
			fonth);
		dst	+= ww;
	}
	set_bcolor(tc);
#ifndef	NO_MMX
	reset_mmx();
#endif
	RamToVideo(dst2, 0, 0, x, y, width, fonth, width+fontw, fonth, BITBLIT_WITH_OPERATOR);
	if (alloc) free(dst2);
	return x + width;
}

// fill box directly to VRAM
void fill_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop)
{
	if ((y + h) < clip_y_min || (x + w) < clip_x_min || y >= clip_y_max || x >= clip_x_max || w < 1 || h < 1)
		return;
	if ((x + w) > clip_x_max)
		w -= ((x + w) - clip_x_max);
	if ((y + h) > clip_y_max)
		h -= ((y + h) - clip_y_max);
	if (x < clip_x_min)
	{
		w += (x - clip_x_min);
		x = clip_x_min;
	}
	if (y < clip_y_min)
	{
		h += (y - clip_y_min);
		y = clip_y_min;
	}
	if (w < 1 || h < 1)
		return;
	vector_fill_box(x, y, w, h,ink,ppop);
#ifdef X11_DRIVER
	UpdateRect(x,y,x,y,w,h);
#endif
}

// draw box directly to VRAM
void draw_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop)
{
	draw_line(x, y, x + w - 1, y,ink,ppop);
	draw_line(x, y + h - 1, x + w - 1, y + h - 1,ink,ppop);
	draw_line(x, y, x, y + h - 1,ink,ppop);
	draw_line(x + w - 1, y, x + w - 1, y + h - 1,ink,ppop);
}

// set clipping for screen (no window!)
void set_clip_rect(int w, int h, int x, int y)
{
	clip_x_min = x;
	clip_y_min = y;
	if (clip_x_max < 0)
		clip_x_min = 0;
	if (clip_y_max < 0)
		clip_y_min = 0;
	clip_x_max = x + w;
	clip_y_max = y + h;
	if (clip_x_max > X_width)
		clip_x_max = X_width;
	if (clip_y_max > Y_width)
		clip_y_max = Y_width;

	// for polygon routines
	ps.gc_xoffset = 0;
	ps.gc_yoffset = 0;
	ps.gc_xcliplo = clip_x_min;
	ps.gc_ycliplo = clip_y_min;
	ps.gc_xcliphi = clip_x_max - 1;
	ps.gc_ycliphi = clip_y_max - 1;

	ps.line = vector_draw_line;
	ps.scan = draw_hline;
	if (vector_clip_rect)
		vector_clip_rect(clip_x_min, clip_y_min, clip_x_max, clip_y_max);
}

// put block from RAM to VRAM with current draw mode
void put_block(int x, int y, int w, int h, FGPixel * p, unsigned ppop)
{
	if (ppop == _GSET)
		RamToVideo(p, 0, 0, x, y, w, h, w, h, BITBLIT_COPY);
	else
		RamToVideo(p, 0, 0, x, y, w, h, w, h, BITBLIT_WITH_OPERATOR);
}

// gut block from VRAM to RAM
void get_block(int x, int y, int w, int h, FGPixel * p)
{
	VideoToRam(p, x, y, 0, 0, w, h, w, h);
}

// return size of buffer for block
unsigned int areasize(int w, int h)
{
	return w * h * bpp;
}

// clear screen
void clear_frame_buffer(FGPixel color)
{
	fill_box(0, 0, X_width, Y_width,color,_GSET);
#ifdef _WIN32
	delay(5);	// empiric constant for VSYNC
#endif
}

// register fixed-binary font
int register_font_fix(unsigned char *source, int width, int height, int count, int offset)
{
	return _register_font(source, width, height, count, offset, 0, -1);
}

// register variable font from memory
// img and data must be allocated with malloc
int register_font_var(variable_record *data, char *img, int size)
{
	img = (char *)realloc(img, size*sizeof(FGPixel));
	char *p1 = img+size-1;
	FGPixel *p2=(FGPixel *)img+size-1;
	while(size--)
		*p2-- = *p1--; // copy with color expand
	return _register_font((unsigned char *)img, data->mw, data->mh, 256, 0, data, -1);
}

// register variable font from file
int register_font_file(const char *source)
{
	fonthdr hdr;
	int size;
	char *img;
	FILE *f = fopen(source, "rb");

	if (f==0)
	{
		perror("register_font ");
		return -1;
	}
	fread(&hdr, sizeof(hdr), 1, f);
	
	fseek(f,0,SEEK_END);
	size=ftell(f)-sizeof(fonthdr);
	fseek(f,sizeof(fonthdr),SEEK_SET);

	if ((img=(char *)malloc(size))==0)
	{
		perror("font malloc ");
		fclose(f);
		return -1;
	}
	fread(img,1,size,f);
	fclose(f);
	int rc = register_font_image(&hdr, img);
	free(img);
	return rc;
}

// register variable font from compiled font
int register_font_image(fonthdr * hdr, char *img)
{
	int w,ww,hh,ww32;
	char *img4;

	// alokuj vlastny font
	hdr->font_maxcharw = (hdr->font_maxcharw+3)&~3;
	variable_record *img2 = (variable_record *)calloc(sizeof(variable_record), 1);
	char *img3 = (char *) calloc(hdr->font_maxcharw*(hdr->font_maxcharh+1)*256, 1), *fimg=img3;
	// nastav pevnu max. velkost jedneho znaku
	img2->mw = hdr->font_maxcharw;
	hh = img2->mh = hdr->font_maxcharh+1;
	//
	img4 = img+hdr->ftable[0].fontchar_fp-sizeof(hdr->ftable);
	for(int i=0;i<256;i++)
	{
		// one char, oofset from the total begin to begin of the char
		img2->off[i] = img3-fimg;
		ww = img2->w[i] = hdr->ftable[i].fontchar_w;
		ww32 = (ww+3)&~3;
//printf("%x %x %x %d %d\n",img3,img4,img3-(char *)img2,img2->off[i],i);
		// test for SPACE
		if (ww==0)
		{
			img2->w[i] = hdr->font_spacesize;
			img3 += hh * ((img2->w[i]+3)&~3);
			continue;
		}
		
		for(int h=0; hdr->ftable[i].fontchar_h > h; h++)
		{
			for(w=0; ww > w; w++)
			{
				if (*img4++)
					*img3 = 255;
				img3++;
			}
			while (ww32>w) { img3++; w++; }
		}
//printf("- %d %d %d %d %d\n", hdr.ftable[i].fontchar_w, ((hdr.ftable[i].fontchar_w+3)&~3), img2->mh, hdr.ftable[i].fontchar_h,((hdr.ftable[i].fontchar_w+3)&~3) * (img2->mh-(hdr.ftable[i].fontchar_h)));
		img3 += ww32 * (hh-hdr->ftable[i].fontchar_h);
	}
	return register_font_var(img2, fimg, int(img3-fimg));
}

int _register_font(unsigned char *source, int width, int height, int count, int offset, variable_record * type, int index)
{
	int i;
	register int c;

	if (index==-1) // calculate first free slot
	{
		for (index=FONT2034+1; index<=FONTLAST; index++)
			if (FGFonts[index].fontimg==0) break;
		if (index>FONTLAST) return -1;
	}

	FGFonts[index].height = height;
	FGFonts[index].width = width;
	FGFonts[index].var = 0;

	if (type == 0) // fixed width
	{
		FGPixel *dst;
		dst = (FGPixel *) calloc(width * height * (count+offset) * bpp, 1);
		assert(dst);
		FGFonts[index].fontimg = dst; // ????
		dst = dst + width*height*offset;
		width = width / 4;
		for (; count--;)
		for (i = 0; i < height; i++)
			switch (width)
			{
				case 6:
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					break;
				case 5:
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					break;
				case 4:
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					break;
				case 3:
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					break;
				case 2:
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					*dst++ = c & 8 ? -1 : 0;
					*dst++ = c & 4 ? -1 : 0;
					*dst++ = c & 2 ? -1 : 0;
					*dst++ = c & 1 ? -1 : 0;
					break;
				case 1:
					c = *source++;
					*dst++ = c & 128 ? -1 : 0;
					*dst++ = c & 64 ? -1 : 0;
					*dst++ = c & 32 ? -1 : 0;
					*dst++ = c & 16 ? -1 : 0;
					break;
			}
	}
	else	// proportional width
	{
		FGFonts[index].var = type;
		FGFonts[index].fontimg = (FGPixel *)source;
	}

	++font_counter;
	return index;
}

void deregister_font(void)
{
}

void deregister_fonts(void)
{
	int i;

	for (i = 0; i < font_counter; i++)
	{
		if (FGFonts[i].var)
			free(FGFonts[i].var);
		free(FGFonts[i].fontimg);
		FGFonts[i].fontimg = 0;
	}
	font_counter=0;
}

int get_font(void)
{
	return fontn;
}

// set current font
int set_font(int f)
{
	int rc = fontn;

	switch (f)
	{
		// load dynamically - save memory usage
		case FONT1222:
		case FONT1625:
		case FONT2034:
			if (FGFonts[f].fontimg == 0)
				switch (f)
				{
					case FONT1222:
						_register_font(miro1220, 12, 20, 95, ' ', 0, f);
						break;
					case FONT1625:
						_register_font(miro1625, 16, 25, 95, ' ', 0, f);
						break;
					case FONT2034:
						_register_font(miro2034, 20, 34, 95, ' ', 0, f);
						break;
				}
		default:
			// test for user defined fonts
			if (!FGFonts[f].fontimg) break;
			// OK -> setup font
			font = FGFonts[f].fontimg;
			fontw = FGFonts[f].width;
			fonth = FGFonts[f].height;
			fontvar = FGFonts[f].var;
			fontn = f;
			break;
	}
	return rc;
}

#if !defined(  __linux__) && !defined(  __rtems__) && !defined(__sun__) && !defined(_WIN32)
// sound a HZ tone for b msec
void Snd(int a, int b)
{
	sound(a);
	delay(b);
	nosound();
}
#endif

// sound efect
void Puk(void)
{
	Snd(100, 10);
}

// for drawing sync
void XWaitRetrace(void)
{
#if !defined(_WIN32) && !defined(__sun__)
#ifdef __linux__
	if (GetPriv())	// when ROOT privs
#endif
	{
//		while ((inp(0x3da) & 0x08) != 0);
//		while ((inp(0x3da) & 0x08) == 0);
	}
#endif
}

void XWaitPeriod(void)
{
#if !defined(_WIN32) && !defined(__sun__)
#ifdef __linux__
	if (GetPriv())	// when ROOT privs
#endif
	{
		while ((inpb(0x3da) & 0x08) == 0);
		while ((inpb(0x3da) & 0x08) != 0);
	}
#endif
}

//
// A traditional printf to the screen + position
//
int gprintf(FGPixel ink, FGPixel paper, int x, int y, const char *format,...)
{
	char s[256];

	va_list arglist;

	va_start(arglist, format);
	vsprintf(s, format, arglist);
	va_end(arglist);
	text_out(x, y, s,ink,paper,get_ppop(),FONT0816);
	return strlen(s);
}

/**
 ** SCANCNVX.C ---- scan fill a convex polygon
 **
 ** Copyright (c) 1995 Csaba Biegl, 820 Stirrup Dr, Nashville, TN 37221
 ** [e-mail: csaba@vuse.vanderbilt.edu] See "doc/copying.cb" for details.
 **/
void do_fill_polygon(int n, int pt[][2], PolygonStruct * current, FGPixel ink, unsigned ppop)
{
	edge *edges = edge_array, *ep;
	scan *scans = scan_array, *sp, *points, *segments;
	int xmin, xmax, ymin, ymax;
	int ypos, nedges, mem=0;

	if ((n > 1) &&
		(pt[0][0] == pt[n - 1][0]) &&
		(pt[0][1] == pt[n - 1][1]))
	{
		n--;
	}
	if (n < 1)
	{
		return;
	}
	if (n > 128)
	{
		edges = (edge *) malloc(sizeof(edge) * (n + 2));
		scans = (scan *) malloc(sizeof(scan) * (n + 8));
		mem = 1;
	}
	if (edges && scans)
	{
		int prevx = xmin = xmax = pt[0][0];
		int prevy = ymin = ymax = pt[0][1];

		nedges = 0;
		ep = edges;
		while (--n >= 0)
		{
			if (pt[n][1] >= prevy)
			{
				ep->e.x = prevx;
				ep->e.y = prevy;
				ep->e.xlast = prevx = pt[n][0];
				ep->e.ylast = prevy = pt[n][1];
			}
			else
			{
				ep->e.xlast = prevx;
				ep->e.ylast = prevy;
				ep->e.x = prevx = pt[n][0];
				ep->e.y = prevy = pt[n][1];
			}
			if ((ep->e.y > (current->gc_ycliphi)) || (ep->e.ylast < (current->gc_ycliplo)))
				continue;
			{
				if (ep->e.y < current->gc_ycliplo)
				{
					ep->e.x += (((int) (((long) (((int) ((ep->e.xlast - ep->e.x)) << 1)) * (long) ((current->gc_ycliplo - ep->e.y))) / (long) ((ep->e.ylast - ep->e.y))) + (((int) ((ep->e.xlast - ep->e.x)) ^ (int) ((current->gc_ycliplo - ep->e.y)) ^ (int) ((ep->e.ylast - ep->e.y))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

					ep->e.y = current->gc_ycliplo;;
				}
			};
			if (ymin > ep->e.y)
				ymin = ep->e.y;
			if (ymax < ep->e.ylast)
				ymax = ep->e.ylast;
			if (xmin > ep->e.x)
				xmin = ep->e.x;
			if (xmax < ep->e.xlast)
				xmax = ep->e.xlast;
			{
				(&ep->e)->dy = (&ep->e)->ylast - (&ep->e)->y;
				(&ep->e)->dx = (&ep->e)->xlast - (&ep->e)->x;
				if ((&ep->e)->dx < 0)
				{
					(&ep->e)->xstep = (-1);
					(&ep->e)->dx = (-(&ep->e)->dx);
				}
				else
				{
					(&ep->e)->xstep = 1;
				}
				if ((&ep->e)->dx > (&ep->e)->dy)
				{
					(&ep->e)->xmajor = 1;
					(&ep->e)->error = (&ep->e)->dx >> 1;
				}
				else
				{
					(&ep->e)->xmajor = 0;
					(&ep->e)->error = ((&ep->e)->dy - ((1 - (&ep->e)->xstep) >> 1)) >> 1;
				}
			};
			ep->status = inactive;
			nedges++;
			ep++;
		}
		if ((nedges > 0) && (xmin <= (current->gc_xcliphi)) && (xmax >= (current->gc_xcliplo)))
		{
			if (xmin < (current->gc_xcliplo))
				xmin = (current->gc_xcliplo);
			if (ymin < (current->gc_ycliplo))
				ymin = (current->gc_ycliplo);
			if (xmax > (current->gc_xcliphi))
				xmax = (current->gc_xcliphi);
			if (ymax > (current->gc_ycliphi))
				ymax = (current->gc_ycliphi);

			for (ypos = ymin; ypos <= ymax; ypos++)
			{
				sp = scans;
				points = 0;
				segments = 0;
				for (n = nedges, ep = edges; --n >= 0; ep++)
				{
					switch (ep->status)
					{
						case inactive:
							if (ep->e.y != ypos)
								break;
							if (ep->e.dy == 0)
							{
								ep->status = passed;
								xmin = ep->e.x;
								xmax = ep->e.xlast;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int _swap_tmpval_ = (xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
								{
									scan *prev = 0;
									scan *work = segments;
									int overlap = 0;

									while (work != 0)
									{
										if ((work->x1 <= xmax) && (xmin <= work->x2))
										{
											overlap = 1;
											if (xmin < work->x1)
												work->x1 = xmin;
											if (xmax > work->x2)
											{
												prev = work;
												while ((work = work->next) != 0)
												{
													if (work->x1 > xmax)
														break;
													if (work->x2 > xmax)
														xmax = work->x2;
												}
												prev->x2 = xmax;
												prev->next = work;
											}
											break;
										}
										if (work->x1 > xmax)
											break;
										prev = work;
										work = work->next;
									}
									if (!overlap)
									{
										sp->x1 = xmin;
										sp->x2 = xmax;
										sp->next = work;
										if (prev)
											prev->next = sp;
										else
											segments = sp;
									}
								};
								sp++;
								break;
							}
							ep->status = active;
						case active:
							xmin = xmax = ep->e.x;
							if (ep->e.ylast == ypos)
							{
								ep->status = passed;
								xmax = ep->e.xlast;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int _swap_tmpval_ = (xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
								{
									scan *prev = 0;
									scan *work = points;

									while (work != 0)
									{
										if (work->x1 > xmin)
											break;
										prev = work;
										work = work->next;
									}
									sp->x1 = xmin;
									sp->x2 = xmax;
									sp->next = work;
									if (prev)
										prev->next = sp;
									else
										points = sp;
								};
								sp++;
							}
							else if (ep->e.xmajor)
							{
								for (;;)
								{
									(&ep->e)->x += (&ep->e)->xstep;
									if (((&ep->e)->error -= (&ep->e)->dy) < 0)
									{
										(&ep->e)->error += (&ep->e)->dx;
										break;
									}
								};
								xmax = ep->e.x - ep->e.xstep;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int _swap_tmpval_ = (xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
							}
							else
							{
								{
									if (((&ep->e)->error -= (&ep->e)->dx) < 0)
									{
										(&ep->e)->x += (&ep->e)->xstep;
										(&ep->e)->error += (&ep->e)->dy;
									}
								};
							}
							{
								scan *prev = 0;
								scan *work = points;

								while (work != 0)
								{
									if (work->x1 > xmin)
										break;
									prev = work;
									work = work->next;
								}
								sp->x1 = xmin;
								sp->x2 = xmax;
								sp->next = work;
								if (prev)
									prev->next = sp;
								else
									points = sp;
							};
							sp++;
							break;
						default:
							break;
					}
				}
				while (points != 0)
				{
					scan *nextpt = points->next;

					if (!nextpt)
						break;
					xmin = points->x1;
					xmax = nextpt->x2;
					points = nextpt->next;
					{
						scan *prev = 0;
						scan *work = segments;
						int overlap = 0;

						while (work != 0)
						{
							if ((work->x1 <= xmax) && (xmin <= work->x2))
							{
								overlap = 1;
								if (xmin < work->x1)
									work->x1 = xmin;
								if (xmax > work->x2)
								{
									prev = work;
									while ((work = work->next) != 0)
									{
										if (work->x1 > xmax)
											break;
										if (work->x2 > xmax)
											xmax = work->x2;
									}
									prev->x2 = xmax;
									prev->next = work;
								}
								break;
							}
							if (work->x1 > xmax)
								break;
							prev = work;
							work = work->next;
						}
						if (!overlap)
						{
							nextpt->x1 = xmin;
							nextpt->x2 = xmax;
							nextpt->next = work;
							if (prev)
								prev->next = nextpt;
							else
								segments = nextpt;
						}
					};
				}
				while (segments != 0)
				{
					xmin = segments->x1;
					xmax = segments->x2;
					segments = segments->next;
					{
						if (xmin > current->gc_xcliphi)
						{
							continue;
						}
						if (xmax < current->gc_xcliplo)
						{
							continue;
						}
						if (xmin < current->gc_xcliplo)
						{
							xmin = current->gc_xcliplo;;
						}
						if (xmax > current->gc_xcliphi)
						{
							xmax = current->gc_xcliphi;;
						}
					};
					(*current->scan) (
										 (xmin + current->gc_xoffset),
										 (ypos + current->gc_yoffset),
										 (xmax - xmin + 1), ink,ppop);
				}
			}
		}
	}
	if (mem)
	{
		free(edges);
		free(scans);
	}
}

void do_polygon(int n, int pt[][2], PolygonStruct * current, FGPixel ink, unsigned ppop)
{
	int i, px, py, x1, y1, x2, y2, doClose = 1;

	if (n <= 0)
		return;
	if (n == 1)
		doClose = 1;
	x1 = x2 = pt[0][0];
	y1 = y2 = pt[0][1];
	for (i = 1; i < n; i++)
	{
		int *ppt = pt[i];

		if (x1 > ppt[0])
			x1 = ppt[0];
		if (x2 < ppt[0])
			x2 = ppt[0];
		if (y1 > ppt[1])
			y1 = ppt[1];
		if (y2 < ppt[1])
			y2 = ppt[1];
	}
	px = pt[n - 1][0];
	py = pt[n - 1][1];
	for (i = 0; i < n; i++)
	{
		x1 = px;
		y1 = py;
		x2 = px = pt[i][0];
		y2 = py = pt[i][1];
		if (i | doClose)
		{
			if (y1 > y2)
			{
				{
					int _swap_tmpval_ = (x1);

					(x1) = (x2);
					(x2) = _swap_tmpval_;
				};
				{
					int _swap_tmpval_ = (y1);

					(y1) = (y2);
					(y2) = _swap_tmpval_;
				};
			}
			{
				if (x1 < x2)
				{
					if (x2 < current->gc_xcliplo)
					{
						continue;
					}
					if (x1 > current->gc_xcliphi)
					{
						continue;
					}
					{
						if (x1 < current->gc_xcliplo)
						{
							y1 += (((int) (((long) (((int) ((y2 - y1)) << 1)) * (long) (current->gc_xcliplo - x1)) / (long) ((x2 - x1))) + (((int) ((y2 - y1)) ^ (int) (current->gc_xcliplo - x1) ^ (int) ((x2 - x1))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							x1 = current->gc_xcliplo;
						}
					};
					{
						if (x2 > current->gc_xcliphi)
						{
							y2 -= (((int) (((long) (((int) ((y2 - y1)) << 1)) * (long) (x2 - current->gc_xcliphi)) / (long) ((x2 - x1))) + (((int) ((y2 - y1)) ^ (int) (x2 - current->gc_xcliphi) ^ (int) ((x2 - x1))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							x2 = current->gc_xcliphi;
						}
					};
				}
				else
				{
					if (x1 < current->gc_xcliplo)
					{
						continue;
					}
					if (x2 > current->gc_xcliphi)
					{
						continue;
					}
					{
						if (x2 < current->gc_xcliplo)
						{
							y2 += (((int) (((long) (((int) ((y1 - y2)) << 1)) * (long) (current->gc_xcliplo - x2)) / (long) ((x1 - x2))) + (((int) ((y1 - y2)) ^ (int) (current->gc_xcliplo - x2) ^ (int) ((x1 - x2))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							x2 = current->gc_xcliplo;
						}
					};
					{
						if (x1 > current->gc_xcliphi)
						{
							y1 -= (((int) (((long) (((int) ((y1 - y2)) << 1)) * (long) (x1 - current->gc_xcliphi)) / (long) ((x1 - x2))) + (((int) ((y1 - y2)) ^ (int) (x1 - current->gc_xcliphi) ^ (int) ((x1 - x2))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							x1 = current->gc_xcliphi;
						}
					};
				}
				if (y1 < y2)
				{
					if (y2 < current->gc_ycliplo)
					{
						continue;
					}
					if (y1 > current->gc_ycliphi)
					{
						continue;
					}
					{
						if (y1 < current->gc_ycliplo)
						{
							x1 += (((int) (((long) (((int) ((x2 - x1)) << 1)) * (long) (current->gc_ycliplo - y1)) / (long) ((y2 - y1))) + (((int) ((x2 - x1)) ^ (int) (current->gc_ycliplo - y1) ^ (int) ((y2 - y1))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							y1 = current->gc_ycliplo;
						}
					};
					{
						if (y2 > current->gc_ycliphi)
						{
							x2 -= (((int) (((long) (((int) ((x2 - x1)) << 1)) * (long) (y2 - current->gc_ycliphi)) / (long) ((y2 - y1))) + (((int) ((x2 - x1)) ^ (int) (y2 - current->gc_ycliphi) ^ (int) ((y2 - y1))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							y2 = current->gc_ycliphi;
						}
					};
				}
				else
				{
					if (y1 < current->gc_ycliplo)
					{
						continue;
					}
					if (y2 > current->gc_ycliphi)
					{
						continue;
					}
					{
						if (y2 < current->gc_ycliplo)
						{
							x2 += (((int) (((long) (((int) ((x1 - x2)) << 1)) * (long) (current->gc_ycliplo - y2)) / (long) ((y1 - y2))) + (((int) ((x1 - x2)) ^ (int) (current->gc_ycliplo - y2) ^ (int) ((y1 - y2))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							y2 = current->gc_ycliplo;
						}
					};
					{
						if (y1 > current->gc_ycliphi)
						{
							x1 -= (((int) (((long) (((int) ((x1 - x2)) << 1)) * (long) (y1 - current->gc_ycliphi)) / (long) ((y1 - y2))) + (((int) ((x1 - x2)) ^ (int) (y1 - current->gc_ycliphi) ^ (int) ((y1 - y2))) >> ((sizeof(int) * 8) - 1)) + 1) >> 1);

							y1 = current->gc_ycliphi;
						}
					};
				}
			};
			(*current->line) (
								 (x1 + current->gc_xoffset),
								 (y1 + current->gc_yoffset),
								 (x2 + current->gc_xoffset),
								 (y2 + current->gc_yoffset),
								 ink,
								 ppop);
		}
	}
}

// fill polygon
void fill_convex(int num, int pole[][2], FGPixel ink, unsigned ppop)
{
	do_fill_polygon(num, pole, &ps,ink,ppop);
}

// draw polygon
void draw_convex(int num, int pole[][2], FGPixel ink, unsigned ppop)
{
	do_polygon(num, pole, &ps,ink,ppop);
}

//
// this code is from allegro lib
//

/* do_ellipse:
 *  Helper function for the ellipse drawing routines. Calculates the points
 *  in an ellipse of radius rx and ry around point x, y, and calls the
 *  specified routine for each one. The output proc will be passed first a
 *  copy of the bmp parameter, then the x, y point, then a copy of the d
 *  parameter (so putpixel() can be used as the callback).
 */
void do_ellipse(int x, int y, int rx, int ry, void (*proc) (int, int, FGPixel,unsigned), FGPixel ink, unsigned ppop)
{
	int ix, iy;
	int h, i, j, k;
	int oh, oi, oj, ok;

	if (rx < 1)
		rx = 1;

	if (ry < 1)
		ry = 1;

	h = i = j = k = 0xFFFF;

	if (rx > ry)
	{
		ix = 0;
		iy = rx * 64;

		do
		{
			oh = h;
			oi = i;
			oj = j;
			ok = k;

			h = (ix + 32) >> 6;
			i = (iy + 32) >> 6;
			j = (h * ry) / rx;
			k = (i * ry) / rx;

			if (((h != oh) || (k != ok)) && (h < oi))
			{
				proc(x + h, y + k, ink, ppop);
				if (h)
					proc(x - h, y + k, ink, ppop);
				if (k)
				{
					proc(x + h, y - k, ink, ppop);
					if (h)
						proc(x - h, y - k, ink, ppop);
				}
			}

			if (((i != oi) || (j != oj)) && (h < i))
			{
				proc(x + i, y + j, ink, ppop);
				if (i)
					proc(x - i, y + j, ink, ppop);
				if (j)
				{
					proc(x + i, y - j, ink, ppop);
					if (i)
						proc(x - i, y - j, ink, ppop);
				}
			}

			ix = ix + iy / rx;
			iy = iy - ix / rx;

		}
		while (i > h);
	}
	else
	{
		ix = 0;
		iy = ry * 64;

		do
		{
			oh = h;
			oi = i;
			oj = j;
			ok = k;

			h = (ix + 32) >> 6;
			i = (iy + 32) >> 6;
			j = (h * rx) / ry;
			k = (i * rx) / ry;

			if (((j != oj) || (i != oi)) && (h < i))
			{
				proc(x + j, y + i, ink, ppop);
				if (j)
					proc(x - j, y + i, ink, ppop);
				if (i)
				{
					proc(x + j, y - i, ink, ppop);
					if (j)
						proc(x - j, y - i, ink, ppop);
				}
			}

			if (((k != ok) || (h != oh)) && (h < oi))
			{
				proc(x + k, y + h, ink, ppop);
				if (k)
					proc(x - k, y + h, ink, ppop);
				if (h)
				{
					proc(x + k, y - h, ink, ppop);
					if (k)
						proc(x - k, y - h, ink, ppop);
				}
			}

			ix = ix + iy / ry;
			iy = iy - ix / ry;

		}
		while (i > h);
	}
}

/* ellipsefill:
 *  Draws a filled ellipse.
 */
void do_fill_ellipse(int x, int y, int rx, int ry, void (*draw_hline) (int, int, int, FGPixel ink, unsigned ppop), FGPixel ink, unsigned ppop)
{
	int ix, iy;
	int a, b, c, d;
	int da, db, dc, dd;
	int na, nb, nc, nd;

	if (rx > ry)
	{
		dc = -1;
		dd = 0xFFFF;
		ix = 0;
		iy = rx * 64;
		na = 0;
		nb = (iy + 32) >> 6;
		nc = 0;
		nd = (nb * ry) / rx;

		do
		{
			a = na;
			b = nb;
			c = nc;
			d = nd;

			ix = ix + (iy / rx);
			iy = iy - (ix / rx);
			na = (ix + 32) >> 6;
			nb = (iy + 32) >> 6;
			nc = (na * ry) / rx;
			nd = (nb * ry) / rx;

			if ((c > dc) && (c < dd))
			{
				draw_hline(x - b, y + c, b * 2,ink,ppop);
				if (c)
					draw_hline(x - b, y - c, b * 2,ink,ppop);
				dc = c;
			}

			if ((d < dd) && (d > dc))
			{
				draw_hline(x - a, y + d, a * 2,ink,ppop);
				draw_hline(x - a, y - d, a * 2,ink,ppop);
				dd = d;
			}

		}
		while (b > a);
	}
	else
	{
		da = -1;
		db = 0xFFFF;
		ix = 0;
		iy = ry * 64;
		na = 0;
		nb = (iy + 32) >> 6;
		nc = 0;
		nd = (nb * rx) / ry;

		do
		{
			a = na;
			b = nb;
			c = nc;
			d = nd;

			ix = ix + (iy / ry);
			iy = iy - (ix / ry);
			na = (ix + 32) >> 6;
			nb = (iy + 32) >> 6;
			nc = (na * rx) / ry;
			nd = (nb * rx) / ry;

			if ((a > da) && (a < db))
			{
				draw_hline(x - d, y + a, d * 2,ink,ppop);
				if (a)
					draw_hline(x - d, y - a, d * 2,ink,ppop);
				da = a;
			}

			if ((b < db) && (b > da))
			{
				draw_hline(x - c, y + b, c * 2,ink,ppop);
				draw_hline(x - c, y - b, c * 2,ink,ppop);
				db = b;
			}

		}
		while (b > a);
	}
}

void fill_ellipse(int x, int y, int rx, int ry, FGPixel ink, unsigned ppop)
{
	if ((y + ry) < clip_y_min || (x + rx) < clip_x_min || (y - ry) >= clip_y_max || (x - rx) >= clip_x_max || rx < 1 || ry < 1)
		return;
	do_fill_ellipse(x, y, rx, ry, draw_hline, ink, ppop);
}

/* ellipse:
 *  Draws an ellipse.
 */
void draw_ellipse(int xs, int ys, int rx, int ry, FGPixel ink, unsigned ppop)
{
	if ((ys + ry) < clip_y_min || (xs + rx) < clip_x_min || (ys - ry) >= clip_y_max || (xs - rx) >= clip_x_max || rx < 1 || ry < 1)
		return;
	do_ellipse(xs, ys, rx, ry, vector_draw_point, ink, ppop);
}

// ***********************************************************************
/* bez_split:
 *  Calculates a weighted average between x1 and x2.
 */
static float bez_split(float mu, float x1, float x2)
{
	return (1.0 - mu) * x1 + mu * x2;
}


/* bezval:
 *  Calculates a point on a bezier curve.
 */
static float bezval(float mu, int *coor)
{
	float work[4];
	int i;
	int j;

	for (i = 0; i < 4; i++)
		work[i] = (float) coor[i * 2];

	for (j = 0; j < 3; j++)
		for (i = 0; i < 3 - j; i++)
			work[i] = bez_split(mu, work[i], work[i + 1]);

	return work[0];
}

/* calc_spline:
 *  Calculates a set of pixels for the bezier spline defined by the four
 *  points specified in the points array. The required resolution
 *  is specified by the npts parameter, which controls how many output
 *  pixels will be stored in the x and y arrays.
 */
void calc_spline(int points[8], int npts, int *x, int *y)
{
	int i;
	float denom;

	for (i = 0; i < npts; i++)
	{
		denom = (float) i / ((float) npts - 1.0);
		x[i] = (int) bezval(denom, points);
		y[i] = (int) bezval(denom, points + 1);
	}
}


/* spline:
 *  Draws a bezier spline onto the specified bitmap in the specified color.
 */
void draw_spline(int points[8], FGPixel ink, unsigned ppop)
{
#define NPTS	  64

	int xpts[NPTS], ypts[NPTS];
	int i;

	calc_spline(points, NPTS, xpts, ypts);

	for (i = 1; i < NPTS; i++)
	{
		vector_draw_line(xpts[i - 1], ypts[i - 1], xpts[i], ypts[i], ink, ppop);

		if (ppop == _GXOR)
			vector_draw_point(xpts[i], ypts[i], ink, ppop);
	}
}

#ifndef __BORLANDC__
// for compatibility
int stricmp(const char *s1, const char *s2)
{
	while (tolower(*s1) == tolower(*s2))
	{
		if (*s1 == 0)
			return 0;
		s1++;
		s2++;
	}
	return (int) tolower(*s1) - (int) tolower(*s2);
}
#endif

int strnicmp(const char *s1, const char *s2, size_t n)
{

	if (n == 0)
		return 0;
	do
	{
		if (tolower(*s1) != tolower(*s2++))
			return (int) tolower(*s1) - (int) tolower(*--s2);
		if (*s1++ == 0)
			break;
	}
	while (--n != 0);
	return 0;
}

char *strupr(register char *s)
{
	register int c;
	char *s2 = s;

	while ((c = *s) != 0)
	{
		if (islower(c))
			*s = toupper(c);
		s++;
	}
	return s2;
}

char *strlwr(char *_s)
{
	char *rv = _s;

	while (*_s)
	{
		*_s = tolower(*_s);
		_s++;
	}
	return rv;
}

#if 0 //(defined(__GNUC__) && (defined(__QNX__)) && !defined(NO_MMX))
// no MMX for QNX
int test_mmx(void)
{
	return 0;
}

void set_mmx(void)
{
}

void reset_mmx(void)
{
}
#endif

//
// NOTE! this runs with pixels not bytes !!!
//
void FGmemset(FGPixel * to, FGPixel data, unsigned c)
{
#if !(defined(__WATCOMC__) || defined(__BORLANDC__) || defined(_MSC_VER) || defined(__sun__))
	__asm__("cld");
#ifdef INDEX_COLORS
	int cc;

	while (int (to) & 3)
	{
		if (--c & 0x80000000)
			return;
		*to++ = data;
	}
	cc = c >> 2;
	if (cc < 4)					// if less than 16 bytes

		asm volatile ("rep\n\t stosb":
					  :"c" (c), "D"(to), "a"(data));

	else
	{
		asm volatile ("mov %%al,%%ah\n pushw %%ax\n pushw %%ax\n popl %%eax\n rep\n\t stosl" : : "c" (cc), "D"(to), "a"(data));

		if (c &= 3)
			asm volatile ("rep\n\t stosb": :"c" (c), "a"(data));
	}
#endif
#ifdef DIRECT_COLORS
	asm volatile ("rep\n\t stosw": :"c" (c), "D"(to), "a"(data));

#endif
#ifdef TRUE_COLORS
	asm volatile ("rep\n\t stosl": :"c" (c), "D"(to), "a"(data));

#endif
#else
	while (c--)
		*to++ = data;
#endif
}

// *******************************************************************

// set foreground color
void set_fcolor(FGPixel c)
{
	_CurrColor = c;
}

// set background color
void set_bcolor(FGPixel c)
{
	_CurrBkColor = c;
}

// get foreground color
FGPixel get_fcolor(void)
{
	return _CurrColor;
}

// get background color
FGPixel get_bcolor(void)
{
	return _CurrBkColor;
}

// set back and foreground color
void set_colors(FGPixel f, FGPixel b)
{
	set_fcolor(f);
	set_bcolor(b);
}

#ifdef INDEX_COLORS

static int colorsFree = 256, startFree;
static int range;

// set color in palette
void _palette(unsigned col, unsigned rgb)
{
	col &= 255;
	// call palette trought vector
	vector_palette(col, rgb);
	_fgl_palette[col] = rgb;
}

// absolete
void set_palette_map(int cnt, int ind, int *pal)
{
	unsigned int color, x = 0;

	for (; cnt; cnt--)
	{
		color = (unsigned int) pal[x++];
		_palette(ind++, color);
	}
}

// get color from palette
unsigned int get_palette(unsigned int i)
{
	return _fgl_palette[i];
}

// refresh current palette
void _set_fgl_palette(void)
{
	int i;

	for (i = 0; i < 256; i++)
		_palette(i, _fgl_palette[i]);
}

// set startup palette
void _set_default_palette(void)
{
	memmove(_fgl_palette, _default_palette, sizeof(_fgl_palette));
	_set_fgl_palette();
}

void GetPaletteEntry(char *rc, char *gc, char *bc, int i)
{
	if (i < 0 || i > 255)
		return;
	*rc = ColorsArray[i].r;
	*gc = ColorsArray[i].g;
	*bc = ColorsArray[i].b;
}

void DeleteColor(int idx)
{
	if (idx < 0 || idx > 255)
		return;
	if (ColorsArray[idx].alfa == 1)		// is free
	{
		ColorsArray[idx].r = (char) idx;	// create an gray color
		ColorsArray[idx].g = (char) idx;
		ColorsArray[idx].b = (char) idx;
		ColorsArray[idx].alfa = 0;
		colorsFree++;
	}
	else if (ColorsArray[idx].alfa > 1)
		ColorsArray[idx].alfa--;
}

void SetColorFuzzy(int a)
{
	range = (a*4) & 15;
}

int GetFreeColors(void)
{
	return colorsFree;
}

void CreatePaletteEntry(int rc, int gc, int bc, int idx)
{
	if (idx < 0 || idx > 255)
		return;
	ColorsArray[idx].r = (char) rc;
	ColorsArray[idx].g = (char) gc;
	ColorsArray[idx].b = (char) bc;
	ColorsArray[idx].alfa++;	// incr used counter

	_palette(idx, bc + gc * 256U + rc * 65536U);
}

static FGPixel CreateColor2(int r, int g, int b)
{
	// first we look at current colors for identical (+-3%) color
	int i, r0, g0, b0, first = 0;

	for (i = 0; i < 256; i++)
	{
		r0 = ColorsArray[i].r;
		g0 = ColorsArray[i].g;
		b0 = ColorsArray[i].b;
		if (ColorsArray[i].alfa)	// is existing colors ?

		{
			if ((r0 - range <= r && r0 + range >= r)
				&& (g0 - range <= g && g0 + range >= g)
				&& (b0 - range <= b && b0 + range >= b))
			{
				// WOW !!! we founds loved color ...
				first = i;
				if (r0 == r && g0 == g && b0 == b)	// must be equivalent

				{
					ColorsArray[i].alfa++;
					return i;	// return palette index

				}
			}
		}
	}

	if (first)
	{
		ColorsArray[first].alfa++;
		return first;			// return palette index

	}

	// we must create new palette entry for your color

	if (colorsFree == 0)
		return (FGPixel) - 1;	// ERROR I haven't any palette entry for one

	colorsFree--;
	for (i = startFree; i < 256; i++)
	{
		if (ColorsArray[i].alfa == 0)	// is free color ?

		{
			CreatePaletteEntry(r, g, b, i);
			return (FGPixel) i;
		}
	}
	return (FGPixel) - 2;		// ERROR - data mismatch !!!

}

//  WARNING colors are in range 0..255 !!!

FGPixel CreateColor(int r, int g, int b, int ind)
{
	if (ind < 0 || ind > 255)
	{
		return CreateColor2(r, g, b);
	}
	CreatePaletteEntry(r, g, b, ind);
	if (ColorsArray[ind].alfa > 1)
		return (FGPixel) ind;	// return palette index

	if (colorsFree == 0)
		return (FGPixel) - 1;	// ERROR I haven't any palette entry for one

	colorsFree--;
	return (FGPixel) ind;		// ERROR - data mismatch !!!

}

#else

int GetFreeColors(void)
{
	return 0xFFFFFF;
}

void SetColorFuzzy(int a)
{
}

FGPixel CreateColor(int r, int g, int b, int)
{
	return FGPixel(FGDirectColor(r + (g << 8) + (b << 16)));
}
#endif

#if 1
#define swap(a,b)           {a^=b; b^=a; a^=b;}
#define	MIN(a,b)	( (a) < (b) ? (a) : (b) )
#define MAX(a,b)	( (a) > (b) ? (a) : (b) )

//
// patterned line
//
void line(plotter_t draw, int a1, int b1, int a2, int b2, Pattern * pat)
{
	int xend, yend, dx, dy, c1, c2, step;
	register int p, x, y;

	dx = abs(a2 - a1);
	dy = abs(b2 - b1);

	if (dx > 32000)
		return;					// bulgar const

	if (dy > 32000)
		return;
	if (pat->reset)
		pat->pos = 0;

	if (dx > dy)
	{							/* slope < 1 => step in x direction */
		x = MIN(a1, a2);
		xend = MAX(a1, a2);
		if (x == a1)
		{
			y = b1;
			step = ((b2 - y) < 0 ? -1 : 1);
		}
		else
		{
			y = b2;
			step = ((b1 - y) < 0 ? -1 : 1);
		}

		p = 2 * dy - dx;
		c1 = 2 * dy;
		c2 = 2 * (dy - dx);

		if (pat->next())
			draw(x, y, _CurrColor, get_ppop());
		while (x < xend)
		{
			x++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				y += step;
				p += c2;
			}
			if (pat->next())
				draw(x, y, _CurrColor, get_ppop());
		}
	}
	else
	{							/* slope > 1 => step in y direction */
		y = MIN(b1, b2);
		yend = MAX(b1, b2);
		if (y == b1)
		{
			x = a1;
			step = ((a2 - x) < 0 ? -1 : 1);
		}
		else
		{
			x = a2;
			step = ((a1 - x) < 0 ? -1 : 1);
		}

		p = 2 * dx - dy;
		c1 = 2 * dx;
		c2 = 2 * (dx - dy);

		if (pat->next())
			draw(x, y, _CurrColor, get_ppop());
		while (y < yend)
		{
			y++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				x += step;
				p += c2;
			}
			if (pat->next())
				draw(x, y, _CurrColor, get_ppop());
		}
	}
}

static int xold, yold;

void draw_pattern_line(int x1, int y1, int x2, int y2, Pattern * p)
{
	line(vector_draw_point, x1, y1, xold = x2, yold = y2, p);
}

void drawto_pattern_line(int x2, int y2, Pattern * p)
{
	line(vector_draw_point, xold, yold, x2, y2, p);
	xold = x2;
	yold = y2;
}

void draw_pattern_box(int x, int y, int a, int b, Pattern *pat)
{
	int tmp = pat->reset;
	pat->reset=1;
	draw_pattern_line(x, y, x+a-1, y, pat);
	pat->reset=0;
	draw_pattern_line(x+a-1, y+1, x+a-1, y+b-1, pat);
	draw_pattern_line(x+a-2, y+b-1, x, y+b-1, pat);
	draw_pattern_line(x, y+b-2, x, y+1, pat);
	pat->reset = tmp;
}

#endif

/* do_arc:
 *  Helper function for the arc function. Calculates the points in an arc
 *  of radius r around point x, y, going anticlockwise from fixed point
 *  binary angle ang1 to ang2, and calls the specified routine for each one. 
 *  The output proc will be passed first a copy of the bmp parameter, then 
 *  the x, y point, then a copy of the d parameter (so putpixel() can be 
 *  used as the callback).
 */
void do_arc(int x, int y, double ang1, double ang2, int r, void (*proc) (int, int, FGPixel ink, unsigned ppop), FGPixel ink, unsigned ppop)
{
	int px, py;
	int ex, ey;
	int px1, px2, px3;
	int py1, py2, py3;
	int d1, d2, d3;
	int ax, ay;
	int q, qe;
	double tg_cur, tg_end;
	int done = 0;
	double rr1, rr2, rr3;
	int rr = (r * r);

	rr1 = r;
	rr2 = x;
	rr3 = y;

	/* evaluate the start point and the end point */
	px = (int) (rr2 + rr1 * cos(ang1));
	py = (int) (rr3 - rr1 * sin(ang1));
	ex = (int) (rr2 + rr1 * cos(ang2));
	ey = (int) (rr3 - rr1 * sin(ang2));

	/* start quadrant */
	if (px >= x)
	{
		if (py <= y)
			q = 1;				/* quadrant 1 */
		else
			q = 4;				/* quadrant 4 */
	}
	else
	{
		if (py < y)
			q = 2;				/* quadrant 2 */
		else
			q = 3;				/* quadrant 3 */
	}

	/* end quadrant */
	if (ex >= x)
	{
		if (ey <= y)
			qe = 1;				/* quadrant 1 */
		else
			qe = 4;				/* quadrant 4 */
	}
	else
	{
		if (ey < y)
			qe = 2;				/* quadrant 2 */
		else
			qe = 3;				/* quadrant 3 */
	}

#define loc_tg(_y, _x)  (_x-x) ? (double)(_y-y)/(_x-x) : (double)(_y-y)

	tg_end = loc_tg(ey, ex);

	while (!done)
	{
		proc(px, py,ink,ppop);

		/* from here, we have only 3 possible direction of movement, eg.
		 * for the first quadrant:
		 *
		 *    OOOOOOOOO
		 *    OOOOOOOOO
		 *    OOOOOO21O
		 *    OOOOOO3*O
		 */

		/* evaluate the 3 possible points */
		switch (q)
		{

			case 1:
				px1 = px;
				py1 = py - 1;
				px2 = px - 1;
				py2 = py - 1;
				px3 = px - 1;
				py3 = py;

				/* 2nd quadrant check */
				if (px != x)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
				}
				/* fall through */

			case 2:
				px1 = px - 1;
				py1 = py;
				px2 = px - 1;
				py2 = py + 1;
				px3 = px;
				py3 = py + 1;

				/* 3rd quadrant check */
				if (py != y)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
				}
				/* fall through */

			case 3:
				px1 = px;
				py1 = py + 1;
				px2 = px + 1;
				py2 = py + 1;
				px3 = px + 1;
				py3 = py;

				/* 4th quadrant check */
				if (px != x)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
				}
				/* fall through */

			case 4:
				px1 = px + 1;
				py1 = py;
				px2 = px + 1;
				py2 = py - 1;
				px3 = px;
				py3 = py - 1;

				/* 1st quadrant check */
				if (py == y)
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;

					q = 1;
					px1 = px;
					py1 = py - 1;
					px2 = px - 1;
					py2 = py - 1;
					px3 = px - 1;
					py3 = py;
				}
				break;

			default:
				return;
		}

		/* now, we must decide which of the 3 points is the right point.
		 * We evaluate the distance from center and, then, choose the
		 * nearest point.
		 */
		ax = x - px1;
		ay = y - py1;
		rr1 = ax * ax + ay * ay;

		ax = x - px2;
		ay = y - py2;
		rr2 = ax * ax + ay * ay;

		ax = x - px3;
		ay = y - py3;
		rr3 = ax * ax + ay * ay;

		/* difference from the main radius */
		if (rr1 > rr)
			d1 = (int) (rr1 - rr);
		else
			d1 = (int) (rr - rr1);
		if (rr2 > rr)
			d2 = (int) (rr2 - rr);
		else
			d2 = (int) (rr - rr2);
		if (rr3 > rr)
			d3 = (int) (rr3 - rr);
		else
			d3 = (int) (rr - rr3);

		/* what is the minimum? */
		if (d1 <= d2)
		{
			px = px1;
			py = py1;
		}
		else if (d2 <= d3)
		{
			px = px2;
			py = py2;
		}
		else
		{
			px = px3;
			py = py3;
		}

		/* are we in the final quadrant? */
		if (qe == q)
		{
			tg_cur = loc_tg(py, px);

			/* is the arc finished? */
			switch (q)
			{

				case 1:
					/* end quadrant = 1? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 2:
					/* end quadrant = 2? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 3:
					/* end quadrant = 3? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 4:
					/* end quadrant = 4? */
					if (tg_cur <= tg_end)
						done = 1;
					break;
			}
		}
	}

	/* draw the last evaluated point */
	proc(px, py,ink,ppop);
}

/* arc:
 *  Draws an arc.
 */
void draw_arc(int x, int y, double ang1, double ang2, int r, FGPixel ink, unsigned ppop)
{
	do_arc(x, y, ang1, ang2, r, vector_draw_point, ink, ppop);
}

void AlphaBlit(FGPixel * dst_data, FGPixel * src_data, int len, int alpha);

#if FASTGL_BPP==32
#define ALPHA_BLIT		res = ((src_pix & 0xFF00FF) - (dst_pix & 0xFF00FF)) * alpha / 256 + dst_pix; \
		src_pix &= 0xFF00; \
		dst_pix &= 0xFF00; \
		g = (src_pix - dst_pix) * alpha / 256 + dst_pix; \
		res &= 0xFF00FF; \
		g &= 0xFF00;
#endif

#if FASTGL_BPP==16
#define ALPHA_BLIT		res = ((src_pix & 0xF81F) - (dst_pix & 0xF81F)) * alpha / 256 + dst_pix; \
		src_pix &= 0x7E0;\
		dst_pix &= 0x7E0;\
		g = (src_pix - dst_pix) * alpha / 256 + dst_pix;\
		res &= 0xF81F;\
		g &= 0x7E0;
#endif

#if FASTGL_BPP==15
#define ALPHA_BLIT		res = ((src_pix & 0x7c1F) - (dst_pix & 0x7c1F)) * alpha / 256 + dst_pix;\
		src_pix &= 0x3E0;\
		dst_pix &= 0x3E0;\
		g = (src_pix - dst_pix) * alpha / 256 + dst_pix;\
		res &= 0x7c1F;\
		g &= 0x3E0;
#endif

//#if FASTGL_BPP==8
//#define ALPHA_BLIT		res = g = src_pix;
//#endif

void AlphaBlit(FGPixel * dst_data, FGPixel * src_data, int len, int alpha)
{
#if FASTGL_BPP != 8
	unsigned long dst_pix, src_pix;
	unsigned long res, g;

	if (ppop != _GTRANSP)
		while (len--)
		{
			src_pix = *src_data++;
			dst_pix = *dst_data;
			ALPHA_BLIT
			* dst_data++ = g | res;
		}
	else
		while (len--)
		{
			if ((src_pix = *src_data++) != (unsigned)transp_color)
			{
				dst_pix = *dst_data;
				ALPHA_BLIT
				* dst_data = g | res;
			}
			dst_data++;
		}
#endif
}

void L1RamToVideoA(FGPixel * src, FGPixel * dst, int w, int h, int offsrc, int offdst, int alpha)
{
	while (h--)
	{
		AlphaBlit(dst, src, w, alpha);
		dst += offdst + w;
		src += offsrc + w;
	}
}

void L1RamToRamA(FGPixel * src, FGPixel * dst, int w, int h, int offdst, int offsrc, int alpha)
{
	while (h--)
	{
		AlphaBlit(dst, src, w, alpha);
		dst += offdst + w;
		src += offsrc + w;
	}
}

static int _oldx, _oldy;

void moveto(int x, int y)
{
	_oldx = x;
	_oldy = y;
}

void drawto_line(int x, int y, FGPixel ink, unsigned ppop)
{
	vector_draw_line(_oldx, _oldy, x, y,ink,ppop);
	moveto(x,y);
}

unsigned long FGClock(void)
{
	static int first = 1, res;
	static struct timeb tbuf, tcur, tlast;
	if (first)
	{
		first = 0;
		ftime(&tbuf);
	}
	ftime(&tcur);
	tlast = tcur;
	
	res = (tcur.time - tbuf.time) * 1000 + (tcur.millitm - tbuf.millitm);
	// test for time change
	if (res<0)
	{
		tbuf = tcur;
		res = 0;
	}
	return res;
}

int GetOS(void)
{
#ifdef __QNX__
	return OSTYPE_QNX;
#endif
#ifdef __linux__
	return OSTYPE_LINUX;
#endif
#ifdef __MSDOS__
	return OSTYPE_MSDOS;
#endif
#ifdef _WIN32
	return OSTYPE_WIN32;
#endif
}
