/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2003  Marian Krivos

    This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    nezmar@atlas.sk

    fastgl.cc - graphics engine
*/

#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>

#ifdef __WATCOMC__
#ifdef __DOS__
#include <direct.h>
#else
#include <dirent.h>
#endif // dos
#elif __linux__
#include <dirent.h>
#include <sys/time.h>
#include <unistd.h>
#elif _MSC_VER
#include <direct.h>

#endif // watcom

#ifdef __BORLANDC__
#include <dir.h>
#endif

#pragma implementation "fastgl.h"

#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"

#include "fglist.h"
#ifdef FG_TTF
#include "fgttf.h"
#endif
#include "listbox.h"

extern void ReleaseFd(void);
extern void ReleaseCd(void);

//
// Global definitions & declaration
//

int Window::allControls = 0;	// numbers of all buttons
const char *DrawBuffer::empty_string="";
int Window::rebuild = 1;		// flag
int Window::optimize = 0;		// flag
int Window::line_x1 = 0;		// flag
int Window::line_y1 = 0;		// flag
int Window::line_x2 = 0;		// flag
int Window::line_y2 = 0;		// flag
int Window::nWnd = 0;			// numbers of all windows
Window *Window::current = (Window *) 0;
Window *Window::first = (Window *) 0;
void (*App::DelayProc) (void) = 0;
void (*App::OnEverySecond)(int) = 0;
static Window *Colise = 0;
static Window *_Note_Ptr, *close_adept;
char EditBox::buf[128], EditBox::clip[128];
int EditBox::is_clip;
GuiEvent App::queque[MAX_EVENT_QUEQUE];
MainHwnd App::appHandler;
int 	App::quequeIndex = 0;
char	App::currdir[128];
char 	*App::homedir = 0;
Window	*App::Root = 0;
int		App::flags;
int		App::video;
int		App::Argc;
int		App::background;
char	**App::Argv;
char	* App::name;
unsigned long App::ticks;
App *cApp;
int WindowDatabase::xyCounter = 0;
WindowDatabase *entryPoint;
static BaseMenu *inmenu;
Control * call_handler = 0; // set if need call handler (in translateevent)
MenuWindow *MenuWindow::CurrentMenu=0;
Control *Control::cControl=0;
GuiHwnd MenuWindow::Proc=0;
Config *cCfg;
static unsigned int CRCTab[256];
int allBitmapCounter=0;
int MOVEABLE = 1, verbose;
static int DoRootWindow = 0;
static int shut_down = 0;
static Window *errWindow = 0;
static int alty[26] = { 
ALT_A,ALT_B,ALT_C,ALT_D,ALT_E,ALT_F,ALT_G,ALT_H,
ALT_I,ALT_J,ALT_K,ALT_L,ALT_M,ALT_N,ALT_O,ALT_P,
ALT_Q,ALT_R,ALT_S,ALT_T,ALT_U,ALT_V,ALT_W,ALT_X,
ALT_Y,ALT_Z };
int __fgl_log_level = 0;
int __force_x, __force_y;
int	force_dac8 = 0;
int	force_ps2 = 0;
FGScreen FGFrameBuffer;
int App::ttf_support = 0;
extern int __fg_fps, __fg_fps_total;
static void InitCRC(void);

#define		NOTEGAP		8
#define NOTEBOOKH		18

class FGTabPage
{
		BaseGui		*self;
		const char	*idstring;
		int			width, ypos, flag;
	public:
		FGTabPage() { self = 0; idstring = 0; width = 0; ypos = 0; flag = 0;}
		FGTabPage(const char *s, BaseGui *p, int y) : self(p), idstring(s), ypos(y)
		{
			int f = set_font(FONTSYSLIGHT);
			width = textwidth((char *)s) + NOTEGAP*2;
			set_font(f);
			flag = 0;
		}
		const char	*GetString(void) { return idstring; }
		int			GetWidth(void) { return width; }
		int			GetY(void) { return ypos; }
		int			GetFlag(void) { return flag; }
		void		SetFlag(int f) { flag = f; }
		BaseGui 	*GetControls(void) { return self; }
};

typedef FGList<FGTabPage> TabPageList;
typedef FGList<FGTabPage>::iterator TabPageListPtr;

static ColorScheme _FGCS =
{
	CWHITED,
	CBLACK,		// window colors
	CBLUELIGHT, // active title
	CBROWN,		// inactive title
	CWHITE,		// wnd border 1
	CGRAYED,	// wnd border 2
	CGRAY2,		// wnd border 3
	CGRAY2,		// statusbar
	CGRAY2,		// menu bg
	CBLUE,		// menu fore
	CGRAY2,		// menu bg active
	CWHITE,		// menu fore active
	CBLACK,		// pbutton fore
	CGRAY1,		// pbutton back
	CDARK,		// pbutton fore - pushed
	CGRAY3,		// pbutton back - pushed
	CWHITE,		// control::frame1
	CBLACK,		// control::frame2
	CGRAYED,	// control::frame3
	CWHITE,		// edit back
	CBLACK,		// edit fore
	CDARK, 		// edit bord1
	CGRAY3,		// edit bord2
	CGRAY1,		// slider
	CGRAY2,		// menuwindow back
	CBLACK,		// menuwindow fore
	CBLUELIGHT,		// pdmenu back
	CWHITE,		// pdmenu fore
	CGRAY3,		// pdmenu gray
	CGRAY1,		// editbox disable
	CBLUELIGHT, // notebook active
	CGRAY3,		// slider disable
};

ColorScheme * CScheme = &_FGCS;

class IconWindow : public Window
{
		Window *original;
		static void Proc(GuiEvent *p)
		{
			IconWindow *pp = (IconWindow *)p->wnd;
			if (p->GetType()==CLICKRIGHTEVENT)
				pp->original->WindowIconize();
			if (p->GetType()==KEYEVENT && p->GetKey()==ALT_F02)
				pp->original->WindowIconize();
			if (p->GetType()==WINDOWMOVEEVENT)
			{
				if (entryPoint) entryPoint->DatabaseSetIcon(p->GetX(), p->GetY(), pp->original->GetName());
				pp->original->SetIconPosition(p->GetX(), p->GetY());
			}
		}
	public:
		IconWindow(Window *parent, int xs, int ys, DrawBuffer *ico)
			: Window(0, xs, ys, ico?ico->GetW()+4:48, ico?ico->GetH()+4:48, "Icon", Proc, 0, CWHITE, WNOPICTO|WICONWINDOWTYPE|WCLICKABLE)
		{
			icon = ico;
			if (icon==0)
				icon = new DrawBuffer(44, 44, BMP_MEM, CGRAYED);
			original = parent;
			original->WindowAttachIcon(icon);
			WindowBox(0,0,w,h,CBLUELIGHT);
			WindowSetWorkRect(2,2,w-4,h-4);
			set_ppop(_GSET);
			WindowText(4,24,"ICON",CWHITE,CGRAYED);
			draw();
		}
		virtual void draw(void)
		{
			if (icon)
			{
//				get_block(x, y, w,h,image);
//				set_ppop(_GTRANSP);
				WindowPutBitmap(0,0,0,0,w,h,icon);
				set_ppop(_GSET);
			}
		}
};

void RootResize(int ww, int hh)
{
	Window::WindowVirtualRoot(0,0,ww,hh);
}

//
// Source code
// --------------------------------------------------------------------------
//
// Internal
//
// kresli ikonky na titulok okna
void Window::WindowPiktograms(int x, int y)
{
	int action= status&WMINIMIZE?1:0;
	action |= status&WNOPICTO?0:2;
	switch(action)
	{
		case 0:	// nothing
			break;
		case 3:	// all (minimize+close)
			RamToRam(0, 0, PICTOSIZE*2, PICTOSIZE, x, y, w,h, PICTOSIZE*2, PICTOSIZE, _internal_bitmap0, image, BITBLIT_WITH_CLIPPING);
			break;
		case 1:	// (minimize)
			RamToRam(0, 0, PICTOSIZE*2, PICTOSIZE, x+PICTOSIZE, y, w,h, PICTOSIZE, PICTOSIZE, _internal_bitmap0, image, BITBLIT_WITH_CLIPPING);
			break;
		case 2:	// (close)
			RamToRam(PICTOSIZE, 0, PICTOSIZE*2, PICTOSIZE, x+PICTOSIZE, y, w,h, PICTOSIZE, PICTOSIZE, _internal_bitmap0, image, BITBLIT_WITH_CLIPPING);
			break;
	}
}

//
// A traditional printf in window form
//

int Window::printf(const char *format, ...)
{
	char s[256];
	va_list arglist;

	va_start(arglist, format);
	vsprintf(s, format, arglist);
	va_end(arglist);
	puts(s);
	return strlen(s);
}

//
// A traditional printf in window form + position
//

int Window::printf(int xx, int yy, const char *format, ...)
{
	char s[256];
	move(xx, yy);
	va_list arglist;

	va_start(arglist, format);
	vsprintf(s, format, arglist);
	va_end(arglist);
	int len = strlen(s);
	if (s[len-1] == '\n') 
	{
	    s[--len] = 0;
	}
	WindowText(xx,yy,s, get_fcolor(), get_bcolor());
	return len;
}

//
// Put char to window at current print position
//
void Window::wputc(int c)
{
	if (c == -1)
	{
		WindowRepaintUser(0, ypos, xpos, fonth);
		return;
	}
	if (xpos == -1)
	{
		WindowScrollUp(0, fonth, wwrk, hwrk - fonth - hwrk % fonth, fonth);
		xpos = 1;
	}
	if (xpos < (wwrk - fontw))
	{
		if (c == '\n')
		{
			if (ypos > (hwrk - fonth * 2))
			{
				WindowRepaintUser(0, ypos, xpos, fonth);
				xpos = -1;
			}
			else
			{
				WindowRepaintUser(0, ypos, xpos, fonth);
				ypos += fonth;
				xpos = 1;
			}
		}
		else
		{
			text(xpos, ypos, (char *) &c);
			xpos += fontw;
		}
	}
	else
	{
		xpos -= fontw;
		wputc('\n');
		wputc(c);
	}
}

//
// Internal
//
void BaseGui::PositionCheck(int &xx, int &yy, int &ww, int &hh)
{
	if (ww > X_width) ww = X_width;
	if (hh > Y_width) hh = Y_width;
	if (ww < 8)	ww = 8;
	if (hh < 8)	hh = 8;
	if (status&WALIGNED)
	{
		xx &= ~3;
		yy &= ~3;
		ww &= ~3;
		hh &= ~3;
	}
	if ((xx + ww) >= X_width) xx = X_width - ww;
	if (xx < 0)	xx = 0;
	if ((yy + hh - 1) >= Y_width) yy = Y_width - hh;
	if (yy < 0)	yy = 0;
}

/*
 * return 0 if in window area
 * else 1 for title
 * 2 for close
 * 3 for minimize
 * 4 for resize
 */
int Window::TitleFind(GuiEvent * e)
{
	if (!IsVisible()) return 0;
	
	if ((x <= e->GetX()) && ((x + w) >= e->GetX()))
	{
		if (((x + w - 13) <= e->GetX()) &&
			((y + h - 13 <= e->GetY()) && ((y + h) >= e->GetY())) && status & WSIZEABLE)
				return 4;
		if (status & WTITLED)
		{
			if ((y <= e->GetY()) && ((y + 3+TITLEH) >= e->GetY()))
			{
				if (!(status & WNOPICTO)) // close or close&minimize
				{
					if (e->GetX() >= (x + w - 3 - PICTOSIZE))
						return 2;
					if (e->GetX() >= (x + w - 3 - PICTOSIZE*2))
						if (status&WMINIMIZE) return 3;
						else return 0;
				}
				else if (status & WMINIMIZE)
				{
					if (e->GetX() >= (x + w - 3 - PICTOSIZE))
						return 3;
				}
				return 1;
			}
		}
		else if (status & WUNMOVED) return 0;
		else if ((y <= e->GetY()) && ((y + h) >= e->GetY()) && status&WICONWINDOWTYPE)
			return 1;
		else if ((y <= e->GetY()) && ((y + TITLEH+3) >= e->GetY()))
			return 1;
	}
	return 0;
}

volatile void DestroyWindow(void)
{
	int c = Window::nWnd;
	if (c && verbose) ::printf("WARNING: %d window(s) not closed at exit\n", c);
}

int	Window::_over(int xa1, int ya1, int xa2, int ya2, int xx, int yy, int ww, int hh)
{
	register int xb1 = xx, xb2 = xx + ww, yb1 = yy, yb2 = yy + hh;

	if ((xa1 <= xb1 && xa2 >= xb1) || (xa1 <= xb2 && xa2 >= xb2))	// x sa prekryva
	{
		if (ya1 <= yb1 && ya2 >= yb1)
		{
			return 1;
		}
		if (ya1 <= yb2 && ya2 >= yb2)
		{
			return 2;
		}
		if (yb1 <= ya1 && yb2 >= ya1)
		{
			return 3;
		}
		if (yb1 <= ya2 && yb2 >= ya2)
		{
			return 4;
		}
	}
	else if ((xb1 <= xa1 && xb2 >= xa1) || (xb1 <= xa2 && xb2 >= xa2))
	{
		if ((yb1 <= ya1 && yb2 >= ya1))
		{
			return 5;
		}
		if ((yb1 <= ya2 && yb2 >= ya2))
		{
			return 6;
		}
		if (ya1 <= yb1 && ya2 >= yb1)
		{
			return 7;
		}
		if (ya1 <= yb2 && ya2 >= yb2)
		{
			return 8;
		}
	}
	return 0;
}

//
// redraw block unconditionally
//
void Window::WindowUpdateBlock(int x, int y, int w, int h)
{
	clip_shape(x,y,w,h);
	RepaintBlock(x,y,w,h);
}

//
// params are real size & position of cleared window
//
void Window::intersect(int _x, int _y, int _w, int _h)
{
	Window *p = first;
	int rc;

	if (!(cApp->flags&APP_ROOTWINDOW)) fill_box(_x,_y,_w,_h,cApp->background,_GSET);
	
	// now we trace all windows from down to *this*
	while (p)
	{
		if (!(p->status&WHIDEN) && p!=this)
		{
			rc=_over(p->x, p->y, p->x+p->w, p->y+p->h, _x, _y, _w, _h);
			if (rc)
			{
				p->WindowUpdateBlock(_x-p->GetX(), _y-p->GetY(), _w, _h);
			}
		}
		p =	(Window *) ((BaseGui *)p)->GetNext();
	}
}

//
// redraw window intersection with original background
// dx,dy are deltas from original
void Window::WindowUpdate(int dx, int dy, int all)
{
	if (all || abs(dx) > w || abs(dy) > h)
	{
		intersect(x,y,w,h);
		return;
	}
	if (dx>0) // right side
	{
		intersect(x,y,dx,h);
	}
	if (dy>0) // up side
	{
		intersect(x,y,w,dy);
	}

	if (dx<0) // left side
	{
		intersect(x+w+dx,y,-dx,h);
	}
	if (dy<0) // down side
	{
		intersect(x,y+h+dy,w,-dy);
	}
}

//
// Test overlaping window, called from OdkryteOkno only
//
BaseGui *BaseGui::OverSprite(BaseGui * od, int xx, int yy, int ww, int hh)
{
	register int xa1, xa2, ya1, ya2;
	BaseGui *v = 0;
	int xb1 = xx, xb2 = xx + ww, yb1 = yy, yb2 = yy + hh;

	while (od != this)
	{
		if (!(od->status & WHIDEN))
		{
			xa1 = od->x,
				xa2 = od->x + od->w,
				ya1 = od->y,
				ya2 = od->y + od->h;

			if ((xa1 <= xb1 && xa2 >= xb1) ||
				(xa1 <= xb2 && xa2 >= xb2))	// x sa prekryva
			{
				if ((ya1 <= yb1 && ya2 >= yb1) ||
					(ya1 <= yb2 && ya2 >= yb2))
				{
					v = od;
					break;
				}
				else if (yb1 <= ya1 && yb2 >= ya2)
				{
					v = od;
					break;
				}
			}
			else if ((xb1 <= xa1 && xb2 >= xa1) ||
				(xb1 <= xa2 && xb2 >= xa2))
			{
				if ((yb1 <= ya1 && yb2 >= ya1) ||
					(yb1 <= ya2 && yb2 >= ya2))
				{
					v = od;
					break;
				}
				else if (ya1 <= yb1 && ya2 >= yb2)
				{
					v = od;
					break;
				}
			}
		}
		od = od->last;
	}
	return v;
}

//
// Called from WindowRepaint() for the window overlap testing
//
int Window::OdkryteOkno(Window * pokial, int x, int y, int w, int h)
{
	if (!IsVisible())
		return -1;				// je skryte

	if (!GetNext())
		return 0;				// je aktivne

	while (pokial != this)
	{
		if ((Colise = (Window *) OverSprite(pokial, x, y, w, h)) != 0)
		{
			return 1;
		}
		pokial = (Window *) pokial->GetLast();
	}
	return 0;
}

#define Ladenie ::printf
#define area(a,b,c,d) *out++=a,*out++=b,*out++=c,*out++=d; //, debug==1?(Ladenie("a,b,c,d - %d,%d,%d,%d\n",a,b,c,d)):0;

int Window::Rozporcuj(int X1, int Y1, int X2, int Y2, int *in, int *out)
{
	int x1, y1, x2, y2, f;

//	FGL_DBG(2,"  ROZPORCUJ %d %d %d %d\n", X1, Y1, X2, Y2);
	if (*in == -1) return 0;
	while (*in != -1)
	{
		x1 = *in++;
		y1 = *in++;
		x2 = *in++;
		y2 = *in++;
		f = 0;
		if (x1 >= X1 && y1 >= Y1 && x2 <= X2 && y2 <= Y2)
			continue;				// cely prekryty

		if (x2 < X1 || y2 < Y1 || x1 > X2 || y1 > Y2)	// cely mimo
		{
			area(x1, y1, x2, y2);
			continue;
		}

		if (x1 <= X1 && x2 >= X1)
		{
			area(x1, y1, X1, y2);
			x1 = X1;
		}

		if (y1 <= Y1 && y2 >= Y1)
		{
			area(x1, y1, x2, Y1);
			y1 = Y1;
		}

		if (x2 >= X2 && x1 <= X2)
		{
			area(X2, y1, x2, y2);
		}

		if (y2 >= Y2 && y1 <= Y2)
		{
			area(x1, Y2, x2, y2);
		}
	}
	*out = -1;
	return 1;
}

//
// aktualizovane okno a relativne suradnice v tomto okne
//

void Window::WindowRepaint(int xr, int yr, int w, int h)
{
	static int zoznam1[4096], zoznam2[4096], *p1, *p2, *p3;
	BaseGui *posledne = this;
	int opt=optimize;

	optimize=0;	// cancel accel hint
	if (status & WLOCKED)
	{
		status |= WDIRTY;
		return;
	}
	// clipping
	if (xr < 0)
	{
		w += xr;
		xr = 0;
		opt = 0;
		if (w <= 0)
			return;
	}
	if (yr < 0)
	{
		h += yr;
		yr = 0;
		opt = 0;
		if (h <= 0)
			return;
	}
	for (; posledne->GetNext(); posledne = posledne->GetNext()); // at the end
	// test overlap
	switch (OdkryteOkno((Window *) posledne, x + xr, y + yr, w, h))
	{
		case -1:
			status |= WDIRTY;
			return;
		case 0:
			switch(opt)
			{
				case 1:
					draw_line(x+xoff+line_x1, y+yoff+line_y1, x+xoff+line_x2, y+yoff+line_y2, get_fcolor(), _GSET);
					break;
				default:
				case 0:
					RepaintBlock(xr, yr, w, h);
					break;
				case 2:
					fill_box(x+xr, y+yr, w, h, get_fcolor(), _GSET);
					break;
				case 3:
					draw_point(x+xr, y+yr, get_fcolor(), _GSET);
					break;
			}
			return;
	}
	// chyba je v okne Colise a mozno aj v oknach medzi nimy
	posledne = Colise;
	p1 = zoznam1;
	p1[0] = x + xr;
	p1[2] = p1[0] + w;
	p1[1] = y + yr;
	p1[3] = p1[1] + h;
	p1[4] = -1;
	p2 = zoznam2;		// inicializacia poli
	p2[0] = -1;

	status |= WDIRTY;
	for (;;)
	{
		if (posledne->IsVisible())
		{
			if (!Rozporcuj(posledne->GetX(), posledne->GetY(), posledne->GetX() + posledne->GetW(), posledne->GetY() + posledne->GetH(), p1, p2))
				return; // ???
			p3 = p1;
			p1 = p2;
			p2 = p3;
		}
		posledne = posledne->GetLast();
		if (posledne == this) // finito?
			break;
		if (OdkryteOkno((Window *) posledne, x + xr, y + yr, w, h) < 1)
			break;
	}
//	FGL_DBG(1," zoznamy %x %x %x %x\n", zoznam1, zoznam2, p1, p2);
	while (*p1 != -1)
	{
		RepaintBlock(p1[0] - x, p1[1] - y, p1[2] - p1[0], p1[3] - p1[1]);
//		FGL_DBG(3,"   REPAINT in WINDOWREPAINT %d %d %d %d\n", p1[0] - x, p1[1] - y, p1[2] - p1[0], p1[3] - p1[1]);
		p1 += 4;
	}
}

//
// redraw user space only
//
void Window::WindowRepaintUser(int xr, int yr, int wr, int hr)
{
	if (clip(xr,yr,wr,hr))
	{
		xr+=xoff;
		yr+=yoff;
		WindowRepaint(xr,yr,wr,hr);
	}
   	optimize = 0; // be carefull!
}

void Window::WindowHide(void)
{
	WindowUpdate(0,0,1); // zmaze ho cele podkladom
}

void Window::WindowShow(void)
{
	if (!iconized)
	{
		RepaintBlock(0, 0, w, h);
		SetStatus(GetStatus() & ~(WDIRTY));
	}
}

void Window::WindowRecopy(void)
{
	if (!IsVisible()) WindowHide();
	else WindowShow();
}


BaseGui *BaseGui::Over(BaseGui * od, BaseGui * po)
{
	int xa1, xa2, ya1, ya2;
	BaseGui *v = 0;
	int xb1 = po->GetX(), xb2 = po->GetX() + po->GetW(), yb1 = po->GetY(),
	  yb2 = po->GetY() + po->GetH();
	BaseGui *t;

	while (od != po && od)
	{
		if (od->IsVisible())
		{
			xa1 = od->GetX(),
				xa2 = od->GetX() + od->GetW(),
				ya1 = od->GetY(),
				ya2 = od->GetY() + od->GetH();

			if ((xa1 <= xb1 && xa2 > xb1) || (xa1 < xb2 && xa2 >= xb2))	// x sa prekryva
			{
				if ((ya1 <= yb1 && ya2 > yb1) || (ya1 < yb2 && ya2 >= yb2))
				{
					v = od;
					break;
				}
				else if (yb1 <= ya1 && yb2 >= ya2)
				{
					v = od;
					break;
				}
			}
			else if ((xb1 <= xa1 && xb2 > xa1) || (xb1 < xa2 && xb2 >= xa2))
			{
				if ((yb1 <= ya1 && yb2 > ya1) || (yb1 < ya2 && yb2 >= ya2))
				{
					v = od;
					break;
				}
				else if (ya1 <= yb1 && ya2 >= yb2)
				{
					v = od;
					break;
				}
			}
		}
		t = od;
		od = od->GetLast();
	}
	return v;
}

/*
** Simply test to Window intersection
*/
int Window::OverWnd(void)
{
	BaseGui *pokial = current;

	if (!GetNext())	return 0;				// is active
	while (pokial != this && pokial)
	{
		if (Over(pokial, this) != 0) return 1;
		pokial = pokial->GetLast();
	}
	return 0; // no window over this
}

long Window::WindowStatus(long stat)
{
	int tmp;
	tmp = status;

	switch (stat)
	{
		case WHIDE:
			if (!IsVisible()) break;
			status |= WHIDEN;
			WindowRecopy();
			break;

		case WVISIBLE:
			if (IsVisible()) break;
			status &= (~WHIDEN);
			WindowRecopy();
			break;

		case WDEACTIVE:
			WindowDrawTitle(CTITLE);
			break;

		case WACTIVE:
			WindowDrawTitle(CTITLEACTIVE);
			break;
	}
	return tmp;
}

// prepne okno na popredie
void Window::WindowFocus(void)
{
	Window *old = current, *novy = this;
	GuiEvent a1(LOSTFOCUSEVENT), a2(GETFOCUSEVENT);

	if (iconized && iconized!=(Window *)-1) novy = iconized;
	FlushInput();
	old->SendToWindow(&a1);
	old = current;

	if (nWnd <= 1) // no more window
		return;
	if (novy == old) // no more to switch
		return;

	old->WindowStatus(WDEACTIVE);	// deactive old

	if (novy == first)				// if switch to first in queque
	{
		first = (Window *) novy->GetNext();
		first->SetLast(0);
	}
	else // switch to other than first
	{
		(novy->GetLast())->SetNext(novy->GetNext());
		(novy->GetNext())->SetLast(novy->GetLast());
	}
	current = novy;
	old->SetNext((BaseGui *) novy);	  // exchange
	novy->SetLast((BaseGui *) old);
	novy->SetNext(0);
	rebuild = 1;
	novy->WindowStatus(WACTIVE); // active new
	oldidm = idm = 0;
	oldidw = idw = 0;
	novy->RepaintBlock(0,0,w,h); // refresh all
	SendToWindow(&a2);
}

static char *skipstr(char *s)
{
	while (*s)
	{
		if (*s > ' ')
			return s;
		s++;
	}
	return s;
}

Control::Control(int xs, int ys, Type typ, int keycode, int flag, Window * own, Bitmap * bm, ControlCall f):
  BaseGui(xs += own->xoff, ys += own->yoff, bm->GetW() + 4, bm->GetH() + 4, 0, typ, 0, 0, flag | WBITMAP)
{
	icon = bm;
	owner = own;
	grp = 0;
	mask = 1;
	iomode = IOSTATE_NORMAL;
	if (keycode >= 'a' && keycode <= 'z')
		keycode = toupper(keycode);
	key = keycode;
	fnc = f;
	Window::allControls++;
	local_id = owner->nControls++;
	xr = owner->GetX() + xs;
	yr = owner->GetY() + ys;
	SetLast( owner->GetLastButton() );
//	owner->lastButton = this;

	if (!owner->Buttony)		// first control
	{
		owner->Buttony = this;
	}
	else
	{
		GetLast()->SetNext(this);
	}
}

Control::Control(int xs, int ys, int ws, int hs, char *nm, Type typ, int keycode, int flag, Window * own, ControlCall f, FGPixel p, FGPixel i):
  BaseGui(xs += own->xoff, ys += own->yoff, ws, hs, nm, typ, i, p, flag)
{
	icon = 0;
	owner = own;
	grp=0;
	mask = 1;
	iomode = IOSTATE_NORMAL;
	if (keycode >= 'a' && keycode <= 'z')
		keycode = toupper(keycode);
	key = keycode;
	fnc = f;
	Window::allControls++;
	local_id = owner->nControls++;
	xr = owner->GetX() + xs;
	yr = owner->GetY() + ys;
	SetLast(owner->GetLastButton() );
//	owner->lastButton = this;
	if (!owner->Buttony)		// first control
	{
		owner->Buttony = this;
	}
	else
	{
		GetLast()->SetNext(this);
	}
}

Control::~Control()
{
	BaseGui *l, *n;

	if (!owner) return;
	if (this==cControl) cControl = 0; // no more valid
	if (this==call_handler) call_handler = 0;
	l = GetLast();
	n = GetNext();

	// erase from visible !
	owner->WindowBox(x-owner->xoff, y-owner->yoff, w, h, owner->GetPaper());

	if (n == (Control *) 0 && l == (Control *) 0)	// no more windows ...
	{
		owner->Buttony = 0;
	}
	else if (n == (BaseGui *) 0)	// last in queqe
	{
//		owner->lastButton = l;
	}
	else if (l == (BaseGui *) 0)	// first in queqe
	{
		owner->Buttony = n;
	}
	else
	{
	}
	Window::allControls--;
	owner->nControls--;
	owner = 0;
}

void Control::Enable(void)
{
	SetStatus((GetStatus() & (~7)) | BNORMAL);
	draw();
}

void Control::Disable(void)
{
	SetStatus((GetStatus() & (~7)) | BDISABLED);
	draw();
}

void EditBox::Disable(void)
{
	FlushInput();
	SetStatus((GetStatus() & (~7)) | BDISABLED);
	draw();
}

void Control::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	status ^= WTRIGGER;
	draw();
	if (a) 
	{
		SetActive();
	}
}

void BaseMenu::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	draw();
	if (a)
	{
		SetActive();
	}
}

Control *Control::ButtonFind(GuiEvent * e, int n)
{
	Window *curr = Window::GetCurrent();
	Control *p = (Control *) curr->GetCurrentControls();
	int c;

	// test for notebools and its bookmarks
	if ( e->GetButtons()&FG_BUTTON_LEFT )
		if (curr->ClickTabPage(e->GetX(), e->GetY()))
			return 0;
		
	if (!p || curr->IsIconized())
		return 0;				// niesu buttony alebo je to ikona

	if (e->GetKey() >= 'a' && e->GetKey() <= 'z')
		c = toupper(e->GetKey());
	else c = e->GetKey();

	if (e->GetType() == KEYEVENT)
	{
		for (;;)
		{
			if ((p->GetKey() == c) && (p->GetStatus() & n & 7) == n)
				return p;
			if ((p = (Control *) p->GetNext()) == 0)
				return 0;
		}
	}
	else
	{
		do
		{
			if ((p->owner->GetX() + p->owner->GetW() >= e->GetX()) && ((p->owner->GetY() + p->owner->GetH()) >= e->GetY()))
				if ((p->GetXr() <= e->GetX()) && ((p->GetXr() + p->GetW()) >= e->GetX()))
					if ((p->GetYr() <= e->GetY()) && ((p->GetYr() + p->GetH()) >= e->GetY()))
						if ((p->GetStatus() & n & 7) == n)
						{
							if (p->GetType() == MENUBUTTON || p->GetType() == MENUBUTTON2)
								inmenu = (BaseMenu *) p;
							return p;
						}
						else if (p->GetStatus() == BDISABLED)
						{
							return 0;
						}
		}
		while ((p = (Control *) p->GetNext()) != 0);
	}
	return 0;
}

// najde okno, ktoremu patrila mysia udalost
Window *Window::WindowFind(GuiEvent * e)
{
	return (Window *) BaseGui ::Find(first, e);
}

BaseGui *BaseGui::Find(BaseGui * p, GuiEvent * e)
{
	BaseGui *x = 0;
	if (!p) return p;
	if (p->next)
			x = Find(p->next, e);
	if (x)
		return x;
	if ((p->x <= e->x) && ((p->x + p->w) >= e->x))
	{
		if ((p->y <= e->y) && ((p->y + p->h) >= e->y))
			if (!(p->status & (WHIDEN|WTEMPHIDEN)) && (p->type!=ROOTWINDOW))
				x = p;
	}
	return x;
}

// najde okno podla nazvu
BaseGui *BaseGui::FindName(BaseGui * p, char *s)
{
	for (; p;)
	{
		if (!strcmp(p->name, s))
			break;
		if (p->next == 0)
			return 0;
		p = p->next;
	}
	return p;
}

// najde okno podla idw
BaseGui *BaseGui::FindId(BaseGui * p, int idw)
{
	for (; p;)
	{
		if (p->GetId() == idw)
			break;
		if (p->next == 0)
			return 0;
		p = p->next;
	}
	return p;
}

void Control::frame(void) 
{
	owner->clipping_state(0);
	set_fcolor(CScheme->button_bord1);
	owner->vline(x, y, h);
	owner->hline(x, y, w);
	set_fcolor(CScheme->button_bord3); // CGRAYED;
	owner->vline(x + w - 2, y + 1, h - 2);
	owner->hline(x + 1, y + h - 2, w - 2);
	set_fcolor(CScheme->button_bord2); // CBLACK;
	owner->vline(x + w - 1, y, h - 0);
	owner->hline(x, y + h - 1, w - 1);
	owner->clipping_state(1);
}

//
// nakresli button
//
void PushButton::draw(void)
{
	int a = set_font(FGFONT_BUTTON);
	int dl= textwidth(name);	// maximalny zobrazitelny pocet znakov

	int xx = w / 2 - dl / 2,
		yy = h / 2 - fonth / 2;

	set_ppop(_GSET);
	frame();
	owner->clipping_state(0);
	switch (status & 7)
	{
		case BNORMAL:
			if (status & WBITMAP)
			{
				owner->WindowPutBitmap(x + 2, y + 2, 0, 0, icon->GetW(), icon->GetH(), icon);
			}
			else
			{
				set_fcolor(CScheme->button_back);
				owner->box(x + 1, y + 1, w - 3, h - 3);
				set_bcolor(CScheme->button_back);
				set_fcolor(CScheme->button_fore);
				owner->text(x + xx, y  + yy, name);
				Underscore(xx , yy, CDARK);
				frame();
			}
			break;

		case BPUSHED:
			set_fcolor(CScheme->button_bord2);
			owner->box(x, y, w, h);
			set_fcolor(CScheme->button_bord1);
			owner->vline(x + w - 2, y + 1, h - 2);
			owner->hline(x + 1, y + h - 2, w - 2);
			if (status & WBITMAP)
			{
				owner->WindowPutBitmap(x + 2, y + 2, 0, 0, icon->GetW(), icon->GetH(), icon);
			}
			else
			{
				set_fcolor(CScheme->button_back_pushed);
				owner->box(x + 2, y + 2, w - 4, h - 4);
				set_bcolor(CScheme->button_back_pushed);
				set_fcolor(CScheme->button_fore_pushed);
				owner->text(x + xx, y + yy, name);
				Underscore(xx , yy, CDARK);
			}
			break;

		case BDISABLED:
			set_fcolor(CScheme->button_bord1);
			owner->box(x,y,w,h);
			set_fcolor(CScheme->button_fore_pushed);
			owner->vline(x + w - 2, y + 1, h - 2);
			owner->hline(x + 1, y + h - 2, w - 2);
			owner->rect(x, y, w, h);
			set_fcolor(CScheme->button_back);
			owner->box(x + 2, y + 2, w - 4, h - 4);
			set_bcolor(CScheme->button_back);
			set_fcolor(CScheme->button_bord3);
			owner->text(x + xx, y + yy, name);
			break;

		case BUNVISIBLE:
			set_fcolor(GetPaper());
			owner->box(x, y, w, h);
			break;
	}
	owner->clipping_state(1);
	set_font(a);
	owner->WindowRepaint(x, y, w, h);
}

void EditBox::MoveIn(void)
{
	set_fcolor(CYELLOW);
	set_ppop(_GSET);
	owner->clipping_state(0);
	owner->rect(GetX() + w1, GetY(), w2, GetH());
	owner->clipping_state(1);
	owner->WindowRepaint(GetX() + w1, GetY(), w2, GetH());
}

void EditBox::MoveOut(void)
{
	set_fcolor(CScheme->edit_bord1);
	set_ppop(_GSET);
	owner->clipping_state(0);
	owner->rect(GetX() + w1, GetY(), w2, GetH());
	owner->clipping_state(1);
	owner->WindowRepaint(GetX() + w1, GetY(), w2, GetH());
}

void PushButton::MoveIn(void)
{
	set_fcolor(CYELLOW);
	set_ppop(_GSET);
	owner->clipping_state(0);
	owner->rect(x,y,w,h);
	owner->clipping_state(1);
	owner->WindowRepaint(x,y,w,h);
}

void PushButton::MoveOut(void)
{
	frame();
	owner->WindowRepaint(x,y,w,h);
}

void PushButton::Push(void)
{
	SetStatus((GetStatus() & (~7)) | BPUSHED | WTRIGGER);
	draw();
}

void PushButton::ClickDown(void)
{
	Push();
}

void PushButton::Release(void)
{
	SetStatus((GetStatus() & (~(7|WTRIGGER))) | BNORMAL);
	draw();
}

void PushButton::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	Release();
	if (a) 
	{
		// if error, set cControl to 0
		SetActive();
	}
}

void CheckButton::draw(void)
{
	FGPixel *ptr;
	int a = set_font(FGFONT_BUTTON);
	if (strlen(name))
	{
		w = textwidth(name)+24;
		h = fonth;
	}
	set_ppop(_GSET);
	set_bcolor(paper);
	owner->clipping_state(0);
	switch (status & 7)
	{
		case BNORMAL:
			set_fcolor(ink);
			owner->text(x + 24, y, name);
			Underscore(24, 0, ink);
			break;
		case BDISABLED:
//			set_fcolor(CGRAY1);
			set_fcolor(CScheme->edit_disable);
			owner->text(x + 24, y, name);
			Underscore(24, 0, CScheme->edit_disable);
			break;
	}

	if (status & WTRIGGER)
	{
		ptr = c_mask;
		if (variable) *variable |= mask;
	}
	else
	{
		ptr = d_mask;
		if (variable) *variable &= ~mask;
	}
	RamToRam(0, 0, 14, 14, x, y, owner->GetW(), owner->GetH(), 14, 14, ptr, owner->GetArray(), BITBLIT_WITH_CLIPPING);
	set_font(a);
	owner->WindowRepaint(x, y, w, h);
	owner->clipping_state(1);
}

#if 0
void PointButton::draw(void)
{
	int a = set_font(FGFONT_BUTTON);

	if (strlen(name))
	{
		w = textwidth(name)+24;
		h = fonth;
	}
	set_ppop(_GSET);
	owner->clipping_state(0);
	set_bcolor(paper);
	if ((status&7)==BDISABLED) set_fcolor(CScheme->edit_disable);
	else set_fcolor(GetInk());
	owner->text(x + 24, y, name);
	Underscore(24, 0, get_fcolor());
	int xx=x+2, yy=y+2;
	if (status & WTRIGGER)
	{
		if (variable) *variable |= mask;
		set_fcolor(CWHITE);
		owner->box(xx,yy,10,10);
		set_fcolor(CBLACK);
		owner->hline(xx,yy,10);
		owner->hline(xx+1,yy+1,8);
		owner->vline(xx,yy,10);
		owner->vline(xx+1,yy+1,8);
	}
	else
	{
		if (variable) *variable &= ~mask;
		set_fcolor(CBLACK);
		owner->box(xx,yy,10,10);
		set_fcolor(CWHITE);
		owner->hline(xx,yy,10);
		owner->hline(xx+1,yy+1,8);
		owner->vline(xx,yy,10);
		owner->vline(xx+1,yy+1,8);
	}
	set_fcolor(CGRAY2);
	owner->box(xx+2,yy+2,6,6);
	set_font(a);
	owner->WindowRepaint(x,y,w,h);
	owner->clipping_state(1);
}
#else
void PointButton::draw(void)
{
	int a = set_font(FGFONT_BUTTON), col;

	if (strlen(name))
	{
		w = textwidth(name)+24;
		h = fonth;
	}
	set_ppop(_GSET);
	owner->clipping_state(0);
	set_bcolor(paper);
	if ((status&7)==BDISABLED) 
		col = CScheme->edit_disable;
	else 
		col = GetInk();
	set_fcolor(col);
	owner->text(x + 24, y, name);
	Underscore(24, 0, col);
	int xx=x+2, yy=y+2;

	set_fcolor(CWHITE);
	owner->fcircle(xx+8,yy+4,6);
	owner->circle(xx+8,yy+4,6);
	set_fcolor(CDARK);
	owner->circle(xx+8,yy+4,6);

	if (status & WTRIGGER)
	{
		if (variable) *variable |= mask;
		owner->fcircle(xx+8,yy+4,2);
		owner->circle(xx+8,yy+4,2);
	}
	else
	{
		if (variable) *variable &= ~mask;
	}
	set_font(a);
	owner->WindowRepaint(x,y,w,h);
	owner->clipping_state(1);
}
#endif

void Window::WindowText(int x, int y, char *s)
{
	set_fcolor(ink);
	set_bcolor(paper);
	text(x, y, s);
	WindowRepaintUser(x, y, textwidth(s), fonth);
}

void Window::WindowText(int x, int y, char *s, int color)
{
	set_bcolor(GetPaper());
	set_fcolor(color);
 	text(x, y, s);
	WindowRepaintUser(x, y, textwidth(s), fonth);
}

void Window::WindowText(int x, int y, char *s, int color, int bk)
{
	set_fcolor(color);
	set_bcolor(bk);
	text(x, y, s);
	WindowRepaintUser(x, y, textwidth(s), fonth);
}

void Window::WindowBox(int x, int y, int a, int b)
{
	if (box(x, y, a, b))
   {
		optimize = 2; // as box
		WindowRepaintUser(x, y, a, b);
   }
}

void Window::WindowBox(int x, int y, int a, int b, int c)
{
	set_fcolor(c);
	if (box(x, y, a, b))
	{
		optimize = 2; // as box
		WindowRepaintUser(x, y, a, b);
	}
}

void Window::WindowRect(int x, int y, int a, int b)
{
	WindowLine(x, y, x+a-1, y);
	WindowLine(x+a-1, y+1, x+a-1, y+b-1);
	WindowLine(x+a-2, y+b-1, x, y+b-1);
	WindowLine(x, y+b-2, x, y+1);
}

void Window::WindowRect(int x, int y, int a, int b, int c)
{
	set_fcolor(c);
	WindowLine(x, y, x+a-1, y);
	WindowLine(x+a-1, y+1, x+a-1, y+b-1);
	WindowLine(x+a-2, y+b-1, x, y+b-1);
	WindowLine(x, y+b-2, x, y+1);
}

void Window::WindowPatternRect(int x, int y, int a, int b, Pattern *pat)
{
	int tmp = pat->reset;
	pat->reset=1;
	WindowPatternLine(x, y, x+a-1, y, pat);
	pat->reset=0;
	WindowPatternLine(x+a-1, y+1, x+a-1, y+b-1, pat);
	WindowPatternLine(x+a-2, y+b-1, x, y+b-1, pat);
	WindowPatternLine(x, y+b-2, x, y+1, pat);
	pat->reset = tmp;
}

void Window::WindowPatternRect(int x, int y, int a, int b, int c, Pattern *pat)
{
	int tmp = pat->reset;
	pat->reset=1;
	set_fcolor(c);
	WindowPatternLine(x, y, x+a-1, y, pat);
	pat->reset=0;
	WindowPatternLine(x+a-1, y+1, x+a-1, y+b-1, pat);
	WindowPatternLine(x+a-2, y+b-1, x, y+b-1, pat);
	WindowPatternLine(x, y+b-2, x, y+1, pat);
	pat->reset = tmp;
}

void Window::WindowLine(int x, int y, int a, int b, int c)
{
	int x1, x2, y1, y2;

	set_fcolor(c);
	if (x < a)
	{
		x1 = x;
		x2 = a;
	}
	else
	{
		x1 = a;
		x2 = x;
	}
	if (y < b)
	{
		y1 = y;
		y2 = b;
	}
	else
	{
		y1 = b;
		y2 = y;
	}
	if (line(line_x1=x, line_y1=y, line_x2=a, line_y2=b)==0)
		optimize = 1; // as not clipped line
	WindowRepaintUser(x1, y1, x2 - x1+1, y2 - y1+1);
}

void Window::WindowLine(int x, int y, int a, int b)
{
	int x1, x2, y1, y2;

	if (x < a)
	{
		x1 = x;
		x2 = a;
	}
	else
	{
		x1 = a;
		x2 = x;
	}
	if (y < b)
	{
		y1 = y;
		y2 = b;
	}
	else
	{
		y1 = b;
		y2 = y;
	}
	if (line(line_x1=x, line_y1=y, line_x2=a, line_y2=b)==0)
		optimize = 1; // as not clipped line
	WindowRepaintUser(x1, y1, x2 - x1+1, y2 - y1+1);
}

void Window::WindowPixel(int x, int y, int c)
{
	if (y <	0 || x < 0 || y >= hwrk || x >= wwrk)
	{
		return;
	}
	set_fcolor(c);
	modify_point(image + w * (y+yoff) + x+xoff);
	optimize = 3; // as point
	WindowRepaintUser(x, y, 1, 1);
}

void Window::WindowPixel(int x, int y)
{
	if (y <	0 || x < 0 || y >= hwrk || x >= wwrk)
	{
		return;
	}
	modify_point(image + w * (y+yoff) + x+xoff);
	WindowRepaintUser(x, y, 1, 1);
	optimize = 3; // as point
}

void Window::WindowPatternLine(int x, int y, int a, int b, Pattern *p)
{
	int x1, x2, y1, y2;
	x1 = (x < a) ? x : a;
	x2 = (x < a) ? a : x;
	y1 = (y < b) ? y : b;
	y2 = (y < b) ? b : y;
	pline(x, y, a, b, p);
	WindowRepaintUser(x1, y1, x2 - x1+1, y2 - y1+1);
}

void Window::WindowPatternLine(int x, int y, int a, int b, int c, Pattern *p)
{
	int x1, x2, y1, y2;
	set_fcolor(c);
	x1 = (x < a) ? x : a;
	x2 = (x < a) ? a : x;
	y1 = (y < b) ? y : b;
	y2 = (y < b) ? b : y;
	pline(x, y, a, b, p);
	WindowRepaintUser(x1, y1, x2 - x1+1, y2 - y1+1);
}

void Window::WindowDrawCircle(int x, int y, int r)
{
	circle(x, y, r);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
}

void Window::WindowDrawCircle(int x, int y, int r, int c)
{
	set_fcolor(c);
	circle(x, y, r);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
}

void Window::WindowFillCircle(int x, int y, int r)
{
	fcircle(x, y, r);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
}

void Window::WindowFillCircle(int x, int y, int r, int c)
{
	set_fcolor(c);
	fcircle(x, y, r);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
}

void Window::WindowDrawEllipse(int x, int y, int rx, int ry)
{
	ellipse(x, y, rx, ry);
	WindowRepaintUser(x - rx, y - ry, rx + rx + 1, ry + ry + 1);
}

void Window::WindowDrawEllipse(int x, int y, int rx, int ry, int c)
{
	set_fcolor(c);
	ellipse(x, y, rx, ry);
	WindowRepaintUser(x - rx, y - ry, rx + rx + 1, ry + ry + 1);
}

void Window::WindowFillEllipse(int x, int y, int rx, int ry, int c)
{
	set_fcolor(c);
	fellipse(x, y, rx, ry);
	WindowRepaintUser(x - rx, y - ry, rx + rx + 1, ry + ry + 1);
}

void Window::WindowFillEllipse(int x, int y, int rx, int ry)
{
	fellipse(x, y, rx, ry);
	WindowRepaintUser(x - rx, y - ry, rx + rx + 1, ry + ry + 1);
}

void Window::WindowDrawArc(int x, int y, double a, double b, int r, int c)
{
	set_fcolor(c);
	arc(x, y, a, b, r);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
}

void Window::WindowDrawArc(int x, int y, double a, double b, int r)
{
	arc(x, y, a, b, r);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
}

void Window::WindowScrollDown(int p1, int p2, int p3, int p4, int p5)
{
	int a = get_fcolor();
	set_fcolor(paper);
	bitblit(p1,p2+p5,p1,p2,p3,p4,this,BITBLIT_WITH_DECREMENT);
	box(p1, p2, p3, p5);
	set_fcolor(a);
	WindowRepaintUser(p1, p2, p3, p4 + p5);
}

void Window::WindowScrollUp(int p1, int p2, int p3, int p4, int p5)
{
	int a = get_fcolor();
	set_fcolor(paper);
	bitblit(p1,p2-p5,p1,p2,p3,p4,this);
	box(p1, p2 + p4 - p5, p3, p5);
	set_fcolor(a);
	WindowRepaintUser(p1, p2 - p5, p3, p4 + p5);
}

void Window::WindowScrollLeft(int p1, int p2, int p3, int p4, int p5)
{
	int a = get_fcolor();
	set_fcolor(paper);
	bitblit(p1 - p5, p2, p1, p2, p3, p4, this);
	box(p1 + p3 - p5, p2, p5, p4);
	set_fcolor(a);
	WindowRepaintUser(p1 - p5, p2, p3 + p5, p4);
}

void Window::WindowScrollRight(int p1, int p2, int p3, int p4, int p5)
{
	int a = get_fcolor();
	set_fcolor(paper);
	bitblit(p1 + p5, p2, p1, p2, p3, p4, this, BITBLIT_WITH_DECREMENT);
	box(p1, p2, p5, p4);
	set_fcolor(a);
	WindowRepaintUser(p1, p2, p3 + p5, p4);
}

void Window::WindowUnLock(void)
{
//	SetDebug(9);
	SetStatus(GetStatus() & (~WLOCKED));
	if (GetStatus() & WDIRTY && IsVisible())
	{
		SetStatus(GetStatus() & (~WDIRTY));
		WindowRepaint(0,0,w,h);
	}
//	SetDebug(0);
}

App::App(int m, int &argc, char **&argv, int bck, int f)
{
	volatile int curr = 0, par = 0;
	char *apn=argv[0], *apn2, **cmdl=argv+1;
	video = m;
	driver = 0;

	background = bck;
	flags = f;
	if (cApp) IError("Multiple App class not allowed !", 1);
	cApp = this;
	homedir = 0;
	if ((f&APP_LOCALDIR) == 0)
		homedir = getenv("HOME");
#ifndef _MSC_VER
	getcwd(currdir, 127);
#else
	_getcwd(currdir, 127);
#endif
	if (homedir==NULL) homedir = ".";
	apn2 = apn;
	apn = apn+strlen(apn)-1;
	if (strcmp(apn-3, ".exe")==0 || strcmp(apn-3, ".EXE")==0)
	{
		apn = apn-4;
		apn[1]=0;
	}
	while ((*apn!='/' && *apn!='\\' && *apn!=':') && apn2<=apn) apn--;
	apn++;
	name = strdup(apn);
#ifdef __linux__
	setenv("EXENAME", name, 1);
#endif
	cCfg = new Config(apn, f&APP_CFG);
	cCfg->ReadInt("default_videomode", video);

	if (argc > 1)
	{
		while (curr<(argc-1))
		{
			curr++;
			if (argv[curr][0] == '-') switch (tolower(argv[curr][1]))
			{
				case 'p':
					if (!memcmp(&argv[curr][1], "ps2", 3))
					{
						force_ps2 = 1;
						if (verbose) printf("PS2 fallback mouse initialization!\n");
						par ++;
					}
					else *cmdl++ = argv[curr];
					break;
				case 'r':
					if (!memcmp(&argv[curr][1], "res", 3))
					{
						curr++;
						if (argc)
							video = argv[curr][0] - '0';
						par += 2;
					}
					else *cmdl++ = argv[curr];
					break;
				case 'd':
					if (!memcmp(&argv[curr][1], "dac", 3))
					{
						force_dac8 = 1;
						par ++;
					}
					else
#ifndef __linux__
					if (!memcmp(&argv[curr][1], "dri", 3))
					{
						curr++;
						if (argc)
							driver = atoi(&argv[curr][0]);
						if (driver == 0)
							driver = -1;	// no detect -> VESA
						par += 2;
					}
					else
#endif
						*cmdl++ = argv[curr];
					break;
				case 's':
#ifdef __linux__
					if (!memcmp(&argv[curr][1], "svg", 3))
					{
						nofb = 1;	// no detect FRAMEBUFFER
						nodga = 1;
						par++;
					}
					else if (!memcmp(&argv[curr][1], "syn", 3))
					{
						curr++;
						if (argc)
							synch = atoi(&argv[curr][0]);
						par += 2;
					}
               else
#endif
					   if (!memcmp(&argv[curr][1], "saf", 3))
					{
						video = G640x480;
	                  verbose = 1;
    	              driver = -1;
        	          __fgl_mmx = 0;
            	      flags &= ~APP_WINDOWDATABASE;
						if (verbose) printf("Safe mode entered ..\a\n");
						par++;
					}
					else *cmdl++ = argv[curr];
					break;
#ifndef __QNX__
				case 'g':
					if (!memcmp(&argv[curr][1], "geo", 3))
					{
						curr++;
						if (argc>1)
						{
							__force_x = atoi(&argv[curr++][0]);
							__force_y = atoi(&argv[curr][0]);
							video 	  = GCUSTOM;
							par+=3;
						} else par++;
					}
					else *cmdl++ = argv[curr];
					break;
#endif
				case 'n':
					if (!memcmp(&argv[curr][1], "nom", 3))
					{
						__fgl_mmx = -1;
						if (verbose) printf("MMX instruction disabled\n");
						par++;
					}
					else
                    	if (!memcmp(&argv[curr][1], "nof", 3))
					{
						fulldrag = 0;
						if (verbose) printf("Full window drag enabled\n");
						par++;
					}
					else if (!memcmp(&argv[curr][1], "noa", 3))
					{
						noaccel = 1;
						if (verbose) printf("HW acceleration disabled\n");
						par++;
					}
					else
#ifdef __linux__
						if (!memcmp(&argv[curr][1], "nox", 3))
					{
						nofb = 1;	// no detect FRAMEBUFFER
						nox11 = 1;
						par++;
					}
					else if (!memcmp(&argv[curr][1], "nod", 3))
					{
						noaccel = 1;
						if (verbose) printf("DGA disabled\n");
						par++;
					}
					else
#endif
						*cmdl++ = argv[curr];
					break;
				case 'b':
					if (!memcmp(&argv[curr][1], "ban", 3))
					{
						linear_flag=0;
						if (verbose) printf("Linear addresing disabled!\n");
						par++;
					}
					else *cmdl++ = argv[curr];
					break;
				case 'h':
					printf("\nOpenGUI %s - (c) Copyright 1996,2003 Marian Krivos\n\n", FG_VERSION);
					printf(" supported image file formats: BMP GIF PCX TGA");
#ifdef FG_JPEG
					printf(" JPG");
#endif
#ifdef FG_PNG
					printf(" PNG");
#endif
#ifdef FG_TTF
					printf(" TRUETYPE");
#endif
					printf("\n");
#ifdef __linux__
					printf(" backend: FRAMEBUFFER SVGALIB");
#ifdef X11_DRIVER
					printf(" X11");
#endif
#ifdef DGA_DRIVER
					printf(" DGA2");
#endif
#endif
#ifdef __QNX__
					printf(" backend: VESA");
#endif
#ifdef __MSDOS__
					printf(" backend: VESA");
#endif
#ifdef _WIN32
					printf(" backend: M$_DirectX");
#endif
					printf(" with %d BPP\n\n", FASTGL_BPP);
					printf(" -resolution N\n");
					printf("          :force scr. resolution 'N', where 'N' = \n");
					printf("           1=320x200, 2=640x480, 3=800x600, 4=1024x768, 5=1280x1024\n");
					printf("           6=1600x1200\n");
#ifndef __linux__
					printf(" -driver N\n");
					printf("          :force display driver to 'N' (if autodetect failed), where 'N' = \n");
					printf("           0=vesa, 1=cirrus, 2=trident, 3=old s3, 4=new s3, 5=mxic\n");
					printf("           6=tseng3000, 7=ati, 8=tseng4000, 9=chips&tech, 10=wd, 11=banshee\n");
					printf("           12=riva 128, 13=matrox, 14=permedia, 15=intel, 16=sis\n");
#else
					printf(" -svga    :don't use FBDEV driver\n");
					printf(" -nodga   :don't use DGA2 driver\n");
					printf(" -nox11   :don't use X11 driver\n");
					printf(" -sync    :monitor refresh\n");
#endif
					printf(" -geometry:set custom resolutions x y\n");
					printf(" -nommx   :don`t use MMX instruction set\n");
					printf(" -noaccel :don`t use HW accelerations\n");
					printf(" -nofull  :use faster window moving\n");
					printf(" -verbose :verbose output\n");
					printf(" -safe    :safe mode\n");
					printf(" -banked  :disable linear mode\n");
					printf(" -dac8    :force 8bit DAC/palette\n");
					printf(" -ps2     :force PS/2 mouse driver (LINUX/QNX)\n");
					getchar();
					par++;
					break;
				case 'v':
					if (!memcmp(&argv[curr][1], "ver", 3))
					{
						verbose = 1;
						par++;
					}
					else *cmdl++ = argv[curr];
					break;
				default:
					*cmdl++ = argv[curr];
					break;
			}
			else *cmdl++ = argv[curr];
		}
	}
	argc -= par;
	Argv = argv;
	Argc = argc;

	if (flags&APP_WINDOWDATABASE) entryPoint = new WindowDatabase(apn);
	if (video < 1 || video > GCUSTOM)
		video = 1;
	if (driver < 0 || driver >= FG_LAST)
		driver = -1;

	if (!graph_set_mode(video))
	{
		printf("Fatal error: can't set graphics mode %d\n\a", video);
		exit(-2);
	}

	if (video>=4)
		idc_normal = &IDC_NORMAL_LARGE;
	else
		idc_normal = &IDC_NORMAL_SMALL;
	CursorLoad(idc_normal);

	SetColorFuzzy(3);
	if (background!=255)
		clear_frame_buffer(background);
	GuiEvent::InitUserEvent();
	SetDelayProc(0);
#ifndef _WIN32
	SetTimerProc(0, 1000);
#endif
	if (f&APP_ROOTWINDOW) Root = new Window(&Root, background);
#if ( !defined( __MSDOS__ ) && !defined( __rtems__ ) ) && !defined(_WIN32)
	setuid(getuid());
	setgid(getgid());
	seteuid(getuid());
	setegid(getgid());
#endif
#if defined( __linux__ ) || defined( __rtems__ )
	FGTimer::TimerInit();
#endif
#ifdef FG_TTF
	if ( TTF_Init() < 0 )
		fprintf(stderr, "Couldn't initialize TTF\n");
	else
	{
		ttf_support = 1;
		atexit(TTF_Quit);
	}
#endif
}

void ClrScr(int ako)
{
	switch(ako)
	{
		default:
		case 0:
			{
				for(int i=0; i<(GetYRes()-1); i++)
				{
					draw_line(i,i,GetXRes(), i,0,_GSET);
					draw_line(i,i,i, GetYRes(),0,_GSET);
					if (i%32 ==0) delay(1);
				}
			}
			break;
	}
}

App::~App()
{
#if defined( __linux__ ) || defined( __rtems__ )
	FGTimer::TimerQuit();
#endif
	ClrScr(0);
//	clear_frame_buffer(CBLACK);
	SetDelayProc(0);
#ifndef _WIN32
	SetTimerProc(0);
#endif
	if (allBitmapCounter && verbose) IError("WARNING: Some non-released bitmaps encoutered !", 0);
	ReleaseFd();
	ReleaseCd();
#ifdef INDEX_COLORS
	_set_default_palette();
#endif	
#ifndef __WATCOMC__
	if (mmx_state) printf("WARNING: MMX set/reset calling mismatch!\n");
#endif	
	if (flags&APP_WINDOWDATABASE) delete entryPoint;
	delete cCfg;
	if (Root) delete Root;
	DestroyWindow();
	free(name);
	cApp = 0;
}

// send message to all windows
void App::BroadcastMessage(GuiEvent *e)
{
	BaseGui *first = Window::GetCurrent();
	while(first)
	{
		((Window *)first)->SendToWindow(e);
		first = first->GetNext();
	}
}

extern int mouseflag;

void App::FGYield(void)
{
	GuiEvent event;
	int tmp = mouseflag;
	mouseflag = 0;
	event.GetUserEvent(1);	// do while ALT+X no pressed
	mouseflag = tmp;
}

void App::Run(MainHwnd hwnd)
{
	GuiEvent *e=queque, event;

	if (hwnd)
		appHandler = hwnd;
		
	if (shut_down ==0) for (;;)
	{
		event.GetUserEvent(-1);	// do while ALT+X no pressed
		event.TranslateUserEvent();
		for (; quequeIndex;)
		{
			quequeIndex--;
			e = queque+quequeIndex;

			// call event handler
			if (appHandler)
				appHandler(e);

			// call event methods
			switch(e->GetType())
			{
				case KEYEVENT:
					OnKeyPress(e->key);
					break;
				case MOVEEVENT:
					OnMouseMove(e->x, e->y);
					break;
				case CLICKLEFTEVENT:
					OnClick(e->x, e->y);
					break;
				case CLICKRIGHTEVENT:
					OnContextPopup(e->x, e->y);
					break;
				case CURSOROUTEVENT:
					OnCursorOut(e->key);
					break;
				case STARTDRAGLEFTEVENT:
					OnStartDrag(1, e->x, e->y);
					break;
				case STARTDRAGRIGHTEVENT:
					OnStartDrag(0, e->x, e->y);
					break;
				case DRAGLEFTEVENT:
					OnEndDrag(1, e->x, e->y, e->w, e->h);
					break;
				case DRAGRIGHTEVENT:
					OnEndDrag(0, e->x, e->y, e->w, e->h);
					break;
				case QUITEVENT:
					shut_down = 1;
					return; // force ret if QUITEVENT
			}
			e->type = NOEVENT;
		}
	}
}

int App::RunModal(Window *which)
{
	GuiEvent *e=queque, event;

    for (;;)
	{
		event.GetUserEvent(-1);	// do while ALT+X no pressed
		event.TranslateUserEvent();
		for (; quequeIndex;)
		{
			quequeIndex--;
			e = queque+quequeIndex;

			// call event handler
			if (appHandler)
				appHandler(e);

			// call event methods
			switch(e->GetType())
			{
				case TERMINATEEVENT:
					if (e->wnd == which)
						return mrClose;
					break;
				case ACCELEVENT:
					if (which == e->wnd)
						if (e->accel->GetParam() != mrNone)
    							return e->accel->GetParam();
					break;
				case QUITEVENT:
					shut_down = 1;
					return mrQuit; // force ret if QUITEVENT
			}
			e->type = NOEVENT;
		}
	}
	return mrNone;
}

/*
   ** prepne na dalsie okno a ak treba tak aj na dalsiu aplikaciu
 */
void Window::nextwnd(void)
{
	BaseGui *p = 0;
	static int cnt = 0;
	int i;

	if (nWnd <= 1 || current->GetStatus() & WMODAL)
	{
		Snd(1000,80);
		return;
	}
	for (i=nWnd; i; i--)
	{
		if (GetCounter() > cnt) cnt++;
		else cnt = 0;
		p = BaseGui::FindId(first, cnt);
		if (!p)
			continue;
		if (current == p || p->GetType()==ROOTWINDOW || (p->GetStatus()&WHIDEN))
			continue;
		break;
	}
	if (!i) for (i=nWnd; i; i--)
	{
		if (GetCounter() > cnt) cnt++;
		else cnt = 0;
		p = BaseGui::FindId(first, cnt);
		if (!p)
			continue;
		if (current == p || p->GetType()==ROOTWINDOW || (p->GetStatus()&WHIDEN))
			continue;
		break;
	}
	if (!i) for (i=nWnd; i; i--)
	{
		if (GetCounter() > cnt) cnt++;
		else cnt = 0;
		p = BaseGui::FindId(first, cnt);
		if (!p)
			continue;
		if (current == p || p->GetType()==ROOTWINDOW || (p->GetStatus()&WHIDEN))
			continue;
		break;
	}
	if (!i) return; // inner loop
	((Window *) p)->WindowFocus();
}

// draw only input text with/without cursor

void EditBox::CapsMode(int m)
{
	caps = m;
	char *s=buf;
	if (caps==EBOX_LOW)
	{
		while(*s)
		{
			if(isupper(*s)) *s+=32;
			s++;
		}
	}
	else if (caps==EBOX_UPR)
	{
		while(*s)
		{
			if(islower(*s)) *s-=32;
			s++;
		}
	}
	draw();
}

// redraw input string
void EditBox::draw(int f)
{
	char str[130];
	unsigned a = set_font(FONT0816);

	set_ppop(_GSET);
	owner->clipping_state(0);
	set_fcolor(CScheme->edit_back);
	owner->box(GetX() + 2 + w1, GetY() + 2, w2 - 4, GetH() - 4);
	strcpy(str, buf);
	if (passwd) 
	{
		char *x=str;
		while (*x)
		{
			if (*x != ' ') *x = '*';
			x++;
		}
	}
	set_colors(CScheme->edit_fore, CScheme->edit_back);
	str[maxpos+offset] = 0;
	if (f)
	{
		owner->text(GetX() + w1 + pos * 8 + 2 + 8, GetY() + 1 + (GetH() / 2 - fonth / 2), str + offset + pos);
		owner->box(GetX() + w1 + pos * 8 + 8, GetY() + 3, 2, GetH() - 6);
		str[pos+offset] = 0;
		owner->text(GetX() + w1 + 8, GetY() + 1 + (GetH() / 2 - fonth / 2), str+offset);
	}
	else
	{
		owner->text(GetX() + w1 + 8, GetY() + 1 + (GetH() / 2 - fonth / 2), str+offset);
	}
	set_font(a);
	owner->WindowRepaint(x, y, w, h);
	owner->clipping_state(1);
}

// draw all

void EditBox::draw(void)
{
	char str[132], *spt=0;
	unsigned a = set_font(FGFONT_BUTTON), cp;
	unsigned ndl=w1-textwidth(name)-8;

	set_ppop(_GSET);
	owner->clipping_state(0);
	offset = pos = 0;
	TestRange();
	
	cp = ((GetStatus() == BDISABLED) ? CScheme->edit_disable : (FGPixel)GetInk());
	set_fcolor(paper);
	owner->box(x, y, w1, h);
	set_colors(cp, paper);
	owner->text(x+ndl, y + 1 + (h / 2 - fonth / 2), name);
	Underscore(x+ndl, 3 + h / 2 - fonth / 2, paper);
	
	switch (status & 7)
	{
		case BNORMAL:
			Underscore(ndl, 3 + h / 2 - fonth / 2, ink);
		case BDISABLED:
			set_font(FONT0816);
			set_fcolor(CScheme->edit_fore);
			owner->box(x + w1, y, w2, h);
			set_fcolor(CScheme->edit_back);
			owner->vline(x + w1 + w2 - 2, y + 1, h - 2);
			owner->hline(x + 1 + w1, y - 2 + h, w2 - 2);
			set_fcolor(CGRAY3);
			owner->box(x + 2 + w1, y + 2, w2 - 4, h - 4);
			set_fcolor(CScheme->edit_bord1);
			owner->rect(x + w1, y, w2, h);
			set_bcolor(CScheme->edit_bord2);
			set_fcolor((status == BDISABLED) ? cp : CScheme->edit_fore);
			switch (data_type)
			{
				case 0:
					if (!hex) sprintf(str, "%d", *(int *) ptr);
					else sprintf(str, "0x%x", *(int *) ptr);
					spt = str;
					break;
				case 1:
					 strcpy(str,(char *)ptr);
					spt = str;
					break;
				case 2:
					sprintf(str, "%f", *(double *) ptr);
					if (*str=='0') strcpy(str, str+1);
					spt = str;
					break;
			}
			spt[maxpos] = 0;	// print visible part only
			if (passwd) 
			{
				char *x=spt;
				while (*x)
				{
					if (*x != ' ') *x = '*';
					x++;
				}
			}
			owner->text(x + w1 + 8, y + 1 + (h / 2 - fonth / 2), spt);
			break;

		case BUNVISIBLE:
			set_fcolor(paper);
			owner->box(x,y,w,h);
			break;
	}
	set_font(a);
	owner->WindowRepaint(x,y,w,h);
	owner->clipping_state(1);
}

//
// vracia 1 na konci
//
int EditBox::inputproc(unsigned c)
{
	char *s, *d;
	int i;

	// input filter
	if (c>= ' ' && c<127) switch(data_type)
	{
		case 0:
			if (!isdigit(c&255) && c != '-')
				if (hex)
				{
					if (c<'A' || c>'f') return 0;
					else if (c>'F' && c<'a') return 0;
				}
				else return 0;
			break;
		case 2:
			if (!isdigit(c&255) && c != '-')
				if (c!= '.' && c!='e' && c!='E') return 0;
			break;
	}

	if (caps==EBOX_UPR && islower(c&255)) c-=32;
	else if (caps==EBOX_LOW && isupper(c&255)) c+=32;

	draw(0); // cursor off
	if (first)
	{
		first = 0;
		if ((isprint(c&255) || c == DEL || c == BACKSP) && c < 256)
		{
			for (i = 0; i < size; buf[i++] = ' ');
			offset = pos = 0;
		}
	}

	switch (c)
	{
		case CTRL_INSERT:
			memcpy(clip,buf,sizeof(clip));
			is_clip = 1;
			break;
		case INSERT:
			memcpy(buf,clip,sizeof(clip));
			break;
		case ESC:
			ininput = 0;
			SetStatus(BNORMAL);
			draw(); // draw final
			return -1;
		case CR:
			ininput = 0;
			switch (data_type)
			{
				case 0:
					int n;
					if (hex) n = strtol(buf, 0, 16); else n = atoi(buf);
					{
						*(int *) ptr = n;
						TestRange();
					}
					break;
				case 1:
					pos = size;
					while (pos)
					{
						if (buf[pos - 1] != ' ')
							break;
						pos--;
					}
					if (buf[pos-1]=='\"' && pos>0) pos--;
					buf[pos]=0;
					strcpy((char *) ptr, buf);
					break;
				case 2:
					double d;
					d = atof(buf);
					{
						*(double *) ptr = d;
                        TestRange();
					}
					break;
			}
			SetStatus(BNORMAL);
			draw(); // draw final
			SetActive();
			return 1;
		case HOME:
			offset = pos = 0;
			break;
		case END:
			{
				pos = size;
				while (pos)
				{
					if (buf[pos - 1] != ' ')
						break;
					pos--;
				}
				offset = pos>maxpos?pos-maxpos:0;
				pos = pos-offset;
			}
			break;
		case BACKSP:
			if (pos+offset)
			{
				s = &buf[offset+pos--];
				d = s - 1;
				while (*s)
					*d++ = *s++;
				*d = ' ';
				if (offset) { offset--; pos++; }
			}
			break;
		case KLEFT:
			if (pos+offset)
			{
				if (pos) pos--;
				else if (offset) offset--;
			}
			break;
		case KRIGHT:
			if (pos+offset < size)
			{
				if (pos<maxpos) pos++;
				else { offset++; }
			}
			break;
		case DEL:
			if (pos+offset < size)
			{
				d = &buf[pos+offset];
				s = d + 1;
				while (*s)
					*d++ = *s++;
				buf[size - 1] = ' ';
			}
			break;

		default:
			if (pos+offset < size)
			{
				if ((c > 31) | (c < 128))
				{
					d = buf + sizeof(buf) -1;
					while(d>(buf+pos+offset))
					{
						*d = d[-1];
						d--;
					}
					*d = (char)c;
				}
				if (pos+offset < size)
				{
					if (pos<maxpos) pos++;
					else { offset++; }
				}
			}
			break;
	}
	buf[size] = 0;
	draw(1); // cursor on
	return 0;
}

// start input line
void EditBox::input(void)
{
	char str[130];
	int a = set_font(FONT0816);

	ininput = this;
	iline = GetY() + 1 + (GetH() / 2 - fonth / 2);
	icol = GetX() + w1 + 8;
	switch (data_type)
	{
		case 0:
			sprintf(str, hex?"%x":"%d", *(int *) ptr);
			break;
		case 1:
			strcpy(str, (const char *) ptr);
			break;
		case 2:
			sprintf(str, "%f", *(double *) ptr);
			if (*str=='0') strcpy(str, str+1);
			break;
	}
	// copy to buffer (and fill it) and find the end of line
	for (int j = 0, i = 0; i < (int)sizeof(buf);)
		if (str[j] == 0)
			buf[i++] = ' ';
		else
			buf[i++] = str[j++];
	buf[size] = 0;
	pos = size-1;
	while (pos >= 0)
	{
		if (buf[pos] > ' ')
			break;
		pos--;
	}
	pos++;
	offset = pos>maxpos?pos-maxpos:0;
	pos -= offset;
	first = 1;
	set_font(a);
}

EditBox::EditBox(int xs,	int ys,	int ws1, int ws2, char *nm, int	key, Window *w,	int *pt, int ink, int paper, ControlCall f, int mn, int mx, int check)
	: Control(xs, ys, ws1+ws2, 21, nm, EDITBOX,	key, BNORMAL, w, f,paper,ink)
{
	int a = set_font(FONT0816);
	w1 = ws1;
	w2 = ws2;
	ptr= pt;
    check_range = check;
	min= mn;
	max=mx;
	data_type=0;
	size = maxpos=(ws2-16)/fontw;
	offset = 0;
	hex = passwd = caps = 0;
	draw();
	set_font(a);
}

EditBox::EditBox(int sz, int xs, int ys, int ws1, int ws2, char *nm, int key, Window *w, char *pt, int ink, int paper, ControlCall f)
: Control(xs, ys, ws1+ws2, 21, nm, EDITBOX,	key, BNORMAL, w, f,paper,ink)
{
	int a = set_font(FONT0816);
	w1 = ws1;
	w2 = ws2;
	ptr= pt;
	data_type=1;
	maxpos=(ws2-16)/fontw;
	size = sz<=0 ? maxpos : ((sz>127) ? 127:sz);
	offset = 0;
	hex = passwd = caps = 0;
	draw();
	set_font(a);
}

EditBox::EditBox(int xs,	int ys,	int ws1, int ws2, char *nm, int	key, Window *w,	double *pt, int	ink, int paper,	ControlCall f, double mn, double mx, int check)
: Control(xs, ys, ws1+ws2, 21, nm, EDITBOX,	key, BNORMAL, w, f,paper,ink)
{
	int a = set_font(FONT0816);
	w1 = ws1;
	w2 = ws2;
	ptr= pt;
	check_range = check;
	mind= mn;
	maxd=mx;
	data_type=2;
	size = maxpos=(ws2-16)/fontw;
	offset = 0;
	hex = passwd = caps = 0;
	draw();
	set_font(a);
}

void EditBox::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	if (!a)	return;
	FlushInput();
	SetStatus(BPUSHED);
	input();
	draw(1);
}

void BaseMenu::MoveOut(void)
{
	draw();
}

void BaseMenu::MoveIn(void)
{
	unsigned a = set_font(FGFONT_MENU);

	owner->clipping_state(0);
	if (type == MENUBUTTON)
	{
		set_fcolor(CScheme->menu_back_active);
		owner->box(x, y-1, w, MENUH-2);
		set_bcolor(CScheme->menu_back_active);
		set_fcolor(CScheme->menu_fore_active);
		owner->text(x+4, y, name);
		Underscore(4,0,CWHITE);
		set_font(a);
		owner->WindowRepaint(x-4, y-2, w+8, MENUH);
	}
	else
	{
		set_fcolor(CScheme->pdmenu_back_active);
		owner->box(x-4, y, w+8, h);
		set_bcolor(CScheme->pdmenu_back_active);
		set_fcolor(CScheme->pdmenu_fore_active);
		owner->text(x+4, y+1, name);
		Underscore(4,1,CScheme->pdmenu_fore_active);
		set_font(a);
		owner->WindowRepaint(x-4, y, w+8, h);
	}
	owner->clipping_state(1);
}

static int is_alted(int c)
{
	for(int i=0;i<26;i++)
		if (alty[i]==c) return i;
	return -1;
}

void Control::Underscore(int xo, int yo, int c)
{
	char *s;
	int pos;
	int size=0;
	if (key==0) return;
	set_fcolor(c);
	if ((pos=is_alted(key)) >= 0)
	{
		set_fcolor(CREDLIGHT);
		if ((s = strchr(name, pos+'a')) != 0)
		{
			size = (int)s - (int) name;
			owner->box(GetX() + _textwidth(name, size) +xo-1, GetY()+13+yo, charwidth(pos+'a')+1,1);
		}
		else if ((s = strchr(name, pos+'A')) != 0)
		{
			size = (int)s - (int) name;
			owner->box(GetX() + _textwidth(name, size) +xo-1, GetY()+13+yo, charwidth(pos+'A')+1,1);
		}
	}
	else if ((s = strchr(name, key)) != 0)
	{
		size = (int)s - (int) name;
		owner->box(GetX() + _textwidth(name, size) +xo-1, GetY()+13+yo, charwidth(key)+1,1);
	}
	else if ((s = strchr(name, key ^ 32)) != 0)
	{
		size = (int)s - (int) name;
		owner->box(GetX() + _textwidth(name, size) +xo-1, GetY()+13+yo, charwidth(key)+1,1);
	}
}

void BaseMenu::draw(void)
{
	unsigned a = set_font(FGFONT_MENU);

	owner->clipping_state(0);
	switch (status & 7)
	{
		case BNORMAL:
			set_fcolor(paper);
			if (type == MENUBUTTON)
			{
				owner->box(GetX(), GetY()-1, GetW(), MENUH-2);
				set_bcolor(GetPaper());
				set_fcolor(GetInk());
				owner->text(GetX()+4, GetY(), GetName());
				Underscore(4,0,GetInk());
				owner->WindowRepaint(GetX()-4, GetY()-2, GetW()+8, MENUH);
			}
			else
			{
				owner->box(GetX()-4, GetY(), GetW()+8, GetH());
				set_bcolor(GetPaper());
				set_fcolor(GetInk());
				owner->text(GetX()+4, GetY()+1, GetName());
				Underscore(4,1,GetInk());
				owner->WindowRepaint(GetX()-4, GetY(), GetW()+8, GetH());
			}
			break;
		case BDISABLED:
			set_bcolor(GetPaper());
			set_fcolor(CScheme->pdmenu_gray);
			owner->text(GetX()+4, GetY(), GetName());
			Underscore(4,0,GetPaper());
			Underscore(4,1,GetPaper());
			owner->WindowRepaint(GetX()-4, GetY(), GetW()+8, GetH());
			break;
	}
	set_font(a);
	owner->clipping_state(1);
}

int Window::GetXM(int a)
{
	if (!(GetStatus() & WMENU))
	{
		::printf("You must use WMENU option for this window\n");
	}
	c_menupos += a;
	return c_menupos - a;
}

int Window::GetYM()
{
	return -(status&WNOTEBOOK ? MENUH+NOTEBOOKH : MENUH);
}

void Window::WindowShape(int sx, int sy, int wx, int wy)
{
	int changepos = (sx!=x || sy!=y), changesize=(wx!=w || wy!=h);
	
	if (changepos && !changesize) // iba zmena polohy
	{
		WindowMove(sx-x, sy-y);
		return;
	}
	else if (!changepos && changesize) // iba zmena velkosti
	{
		WindowResize(wx-w, wy-h);
		return;
	}

	c_menupos = 1;
	if (status&WALIGNED)
	{
		sx &= ~3;
		sy &= ~3;
		wx &= ~3;
		wy &= ~3;
	}

	if (sx>X_width-wx) sx = X_width-wx;
	if (sy>Y_width-wy) sy = Y_width-wy;
	if (sx<0) sx = 0;
	if (sy<0) sy = 0;
	if (wx > X_width) wx = X_width;
 	if (wx < 32)	  wx = 32;
	if (wy > Y_width) wy = Y_width;
	if (wy < (BASEBORDER*2+TITLEH+MENUH+STATUSBARH+NOTEBOOKH))
		wy = BASEBORDER*2+TITLEH+MENUH+STATUSBARH+NOTEBOOKH;

	int nx=sx-x, ny=sy-y, dx=wx-w, dy=wy-h;

	if (CallWindowHook && changepos)  CallWindowHook(x, y, nx, ny, 1); // 1 as move
	if (CallWindowHook && changesize) CallWindowHook(w, h, dx, dy, 2); // 2 as size

	// zrus obrysy povodneho okna
	WindowUpdate(nx, ny, 1); // redraw on screen

	if (status&WNOTEBOOK)
		RelocateTabPage(nx, ny);
	else
		Relocate(nx, ny);
	Resize(dx, dy);
	xpos = 2;
	ypos = 0;
	// nakresli nove okno
	draw();
	SendEvent(this, WINDOWRESIZEEVENT, x,y,w,h);
	SendEvent(this, WINDOWMOVEEVENT, x,y,w,h);
}

void Window::WindowVirtualRoot(int sx, int sy, int wx, int wy)
{
	Window *Root = App::GetRootWindow();
	if (!Root) return;
	int nx=sx-Root->x, ny=sy-Root->y, dx=wx-Root->w, dy=wy-Root->h;

	Root->Relocate(nx, ny);
	Root->Resize(dx, dy);
	Root->clear(0);
	ShowAll();
}

void Window::WindowMove(int nx, int ny)
{
	if (status&WALIGNED && !fulldrag)
	{
		nx &= ~3;
		ny &= ~3;
	}
	if (CallWindowHook) CallWindowHook(x, y, nx, ny, 1); // 1 as move
	XWaitRetrace();
	WindowUpdate(nx,ny);
	if (status&WNOTEBOOK)
		RelocateTabPage(nx, ny);
	else
		Relocate(nx, ny);
	WindowRepaint(0,0,w,h);
	SendEvent(this, WINDOWMOVEEVENT, x,y,w,h);
	rebuild = 1;
}

void Window::WindowResize(int dx, int dy)
{
	if (status&WALIGNED && !fulldrag)
	{
		dx &= ~3;
		dy &= ~3;
	}
	if (CallWindowHook) CallWindowHook(w, h, dx, dy, 2); // 2 as resize
	if (w + dx > X_width 
	    || w + dx < 32 
	    || h + dy > Y_width 
	    || h + dy < (BASEBORDER*2+TITLEH+MENUH+STATUSBARH+NOTEBOOKH))
	{
		return;
	}

	c_menupos = 1;
	XWaitRetrace();
	if (dx<0 || dy<0) WindowUpdate(dx,dy);

	Resize(dx, dy);
	xpos = 2;
	ypos = 0;
	draw();
	SendEvent(this, WINDOWRESIZEEVENT, x,y,w,h);
	rebuild = 1;
}

void Window::WindowCopyFrom(int x, int y, int xs, int ys, int w, int h, Window * p)
{
	WindowPutBitmap(x, y, xs, ys, w, h, p);
}

void Window::WindowPutBitmap(int x, int y, int xs, int ys, int w, int h, DrawBuffer * p)
{
	int opcia = p->GetAlpha() | BITBLIT_WITH_OPERATOR;
	
	if (p->GetType() == BMP_NONE) return;
	int oldp = ppop, oldt = transp_color;
	if (p->GetColorKey() != -1)
	{
		set_ppop(_GTRANSP);
		transp_color = p->GetColorKey();
	}
   	if (w>1 && h>0)
	{
		bitblit(x,y,xs,ys,w,h,p,opcia);
		WindowRepaintUser(x, y, w, h);
	}
	set_ppop(oldp);
	transp_color = oldt;
}

void Window::WindowGetBitmap(int x, int y, int xs, int ys, int w, int h, DrawBuffer * p)
{
	if (p->GetType() == BMP_NONE)
		return;
	x += xoff;
	y += yoff;
	RamToRam(x, y, GetW(), GetH(), xs, ys, p->GetW(), p->GetH(), w, h, image, p->GetArray(), BITBLIT_WITH_CLIPPING);
}

// internal for print error message
static void IErrorText(char *s, int flag)
{
	Snd(1000, 100);
	if (flag)
	{
		printf("FATAL ERROR: %s\n", s);
		getchar();
		exit(flag);
	}
	else
	{
		printf("WARNING: %s\n", s);
		getchar();
	}
}

void IError(char *s, int flag)
{
	int w = (strlen(s) + 8) * 4;
	static Window *errWindow=0;

	if (errWindow)
    	return;
	if (X_width)
	{
		w = w < 75 ? 75 : w;
		Snd(1000,80);
		errWindow = new Window(&errWindow, 0, 0, w * 2, 78, "Error", GuiHwnd(0), CWHITE, CREDLIGHT, ((flag&1) ? WMODAL : 0) | WCENTRED | WCENTRED | WESCAPE);

		if (flag&1) errWindow->WindowText(w - 70, 8, "* * * ERROR * * *", CYELLOW);
		else errWindow->WindowText(w - 78, 8, "* * * WARNING * * *", CYELLOW);

		errWindow->WindowText(32, 28, s, CWHITE);
		errWindow->WindowRect(0,0,2*w,78,CBLACK);

		if (flag&1) errWindow->AddPushButton(w - 40, 48, 80, 21, "Abort", CR, Control::Quit);
		else errWindow->AddPushButton(w - 40, 48, 80, 21, "Ok", CR, Control::Close);
	}
	else IErrorText(s, flag);
	if (verbose) ::printf("IError:: %s\n", s);
//#ifndef FG_THREADED
	if (flag&2) errWindow->ShowModal();
//#endif
}


static FILE *f;

WindowDatabase::WindowDatabase(char *exename)
{
	int magic;
#ifndef __MSDOS__
   sprintf(name, "%s/.windowxy.wdb", App::homedir);
#else
   strcpy(name, "windowxy.wdb");
#endif	
    sprintf(exe, "%s::", exename);
   ::f = fopen(name, "rb");
   if (::f)
	{
		fread(&magic, sizeof(int), 1,::f);
		fread(&xyCounter, sizeof(int), 1,::f);
		if (xyCounter<1 || xyCounter>100 || magic!=0x4357)
		{
			remove(name);
			xyCounter = 0;
			endpointer = &first;
			return; // possibly damaged config
		}
		WindowItem **previous = &first;

		for (int i = 0; i < xyCounter; i++)
		{
			*previous = new WindowItem();
			previous = &((*previous)->next);
		}
		*previous = 0;
		endpointer = previous;
		fclose(::f);
	}
	else
	{
		endpointer = &first;
	}
}

WindowDatabase::~WindowDatabase()
{
	int magic=0x4357;
	if (xyCounter <= 0)	return;
    ::f = fopen(name, "wb");
    if (::f)
	{
		WindowItem *p = first, *pp;
		fwrite(&magic, sizeof(int), 1,::f);
		fwrite(&xyCounter, sizeof(int), 1,::f);

		for (int i = 0; i < xyCounter; i++)
		{
			pp = p->next;
			delete p;
			p = pp;
		}
		fclose(::f);
	}
}

WindowItem *WindowDatabase::findxy(char *s)
{
	WindowItem *p = first;
	char ss[128];
	
	sprintf(ss, "%s%s", exe, s);
	for (int i = xyCounter; i; --i)
	{
		if (!strcmp(p->name, ss))
			return p;
		p = p->next;
	}
	return 0;
}

WindowItem::WindowItem()
{
	char s[80],*p=s;
	int	c;
	fread(&x, sizeof(int),1,::f);
	fread(&y, sizeof(int),1,::f);
	fread(&w, sizeof(int),1,::f);
	fread(&h, sizeof(int),1,::f);
	fread(&icon_x, sizeof(int),1,::f);
	fread(&icon_y, sizeof(int),1,::f);
	fread(&flag, sizeof(int),1,::f);
	while ((c=fgetc(::f)) != 0)	*p++=(char)c;
	*p=0;
	name = new char[strlen(s)+1];
	strcpy(name,s);
}

WindowItem::WindowItem(int xx, int yy, int ww, int hh, char	*s)
{
	x=xx;
	y=yy;
	w=ww;
	h=hh;
	icon_x = -1;
	icon_y = -1;
	flag = 0;
	name = new char[strlen(s)+1];
	strcpy(name,s);
}

WindowItem::~WindowItem()
{
	fwrite(&x,sizeof(int),1,::f);
	fwrite(&y,sizeof(int),1,::f);
	fwrite(&w,sizeof(int),1,::f);
	fwrite(&h,sizeof(int),1,::f);
	fwrite(&icon_x,sizeof(int),1,::f);
	fwrite(&icon_y,sizeof(int),1,::f);
	fwrite(&flag,sizeof(int),1,::f);
	fputs(name,::f);
	fputc(0,::f);
	delete [] name;
}

void WindowDatabase::Add(int &xx, int &yy, int &ww, int &hh, char *s, int fl)
{
	WindowItem *w;

	if (strlen(s) == 0)
		return;

	if ((w = findxy(s))==0)
	{
		char ss[128];
		sprintf(ss, "%s%s",exe, s);
		*endpointer = new WindowItem(xx, yy, ww, hh, ss);
		endpointer = &((*endpointer)->next);
		*endpointer = 0;
		xyCounter++;
	}
	else if (fl&WUSELAST)
	{
		xx = w->x;
		yy = w->y;
		if (fl&WSIZEABLE)
		{
			ww = w->w;
			hh = w->h;
		}
	}
}

void WindowDatabase::DatabaseResize(int xx, int yy, char *s)
{
	WindowItem *w = findxy(s);
	if (w)
	{
		w->w = xx;
		w->h = yy;
	}
}

void WindowDatabase::DatabaseSetIcon(int xx, int yy, char *s)
{
	WindowItem *w = findxy(s);

	if (w)
	{
		w->icon_x = xx;
		w->icon_y = yy;
	}
}

void WindowDatabase::DatabaseGetIcon(int &xx, int &yy, char *s)
{
	WindowItem *w = findxy(s);

	if (w)
	{
		xx = w->icon_x;
		yy = w->icon_y;
	}
}

void WindowDatabase::DatabaseRelocate(int xx, int yy, char *s)
{
	WindowItem *w = findxy(s);

	if (w)
	{
		w->x = xx;
		w->y = yy;
	}
}

// vykresli titulok okna
void Window::WindowDrawTitle(int color)
{
	if (status & WTITLED) // modry/hnedy titulok
	{
		int a = set_font(FGFONT_TITLE), fr = 0, picto=(status&WNOPICTO?0:(status&WMINIMIZE)?PICTOSIZE*2:PICTOSIZE);
		clipping_state(0);
		set_ppop(_GSET);
		if (GetStatus() & WFRAMED) fr = BASEBORDER-1;
		set_fcolor(color);
		set_bcolor(color);

		box(fr, fr, w - (fr + fr)-picto, TITLEH);
		set_fcolor(CWHITE);
		box(fr, fr+TITLEH-1, w - (fr + fr)-picto, 1);
		box(w-((status & WFRAMED) ?4:0), fr, 1, TITLEH);
		if (status&WMINIMIZE && !(status&WNOPICTO))
			box(w-PICTOSIZE*2-4, fr, 1, TITLEH);
		if (status&WMINIMIZE || !(status&WNOPICTO))
			box(w-PICTOSIZE-4, fr, 1, TITLEH);
		text(5 + fr, fr + 1, skipstr(name));
		WindowPiktograms(w - PICTOSIZE*2-3, fr);
		WindowRepaint(fr, 0, w-fr-fr, BASEBORDER+TITLEH);
		clipping_state(1);
		set_font(a);
	}
}

// nakresli okno
void Window::draw(void)
{
    int ff = GetStatus() & WFRAMED;
	int	tt = GetStatus() & WTITLED;
	int	pp = GetStatus() & WNOPICTO;
	int	mm = GetStatus() & WMENU;
	int	mw = GetStatus() & WMENUWINDOWTYPE;
	int	ss = GetStatus() & WSTATUSBAR;
	int	nn = GetStatus() & WNOTEBOOK;
	int vs = BASEBORDER, vrch;

	set_ppop(_GSET);
	clipping_state(0);
	if (tt) vs+=(TITLEH+2);
	if (mm) vs+=MENUH+1;
	if (nn) vs+=NOTEBOOKH+1;
	vrch=vs;
	if (ss) vs+=STATUSBARH;
    if (!ff)  // ak nemas FRAME, potom urob podklad pod titulok
    {
       	vs -= 2;
       	vrch -= BASEBORDER;
       	if (tt) FGmemset(image+w * TITLEH, CBLACK, w); // standard
		else FGmemset(image+w * TITLEH, paper, w); // standard
    }
	if (!ff && !pp && tt)
	{
		FGmemset(image, BORD3, w * TITLEH); // standard
	}
   	else if (!tt) FGmemset(image, paper, w * TITLEH);
	if (ff)
	{
		FGmemset(image, BORD3, w*h); // standard
	    set_fcolor(BORD1);
	    box(0, 0, w, 1);
	    box(0, 0, 1, h);
	    set_fcolor(BORD2);
		box(1, h-1, w, 1);
	    box(w-1, 1, 1, h-1);
	    set_fcolor(paper);
		if (!mw)
		{
		    box(BASEBORDER, vrch, w-BASEBORDER*2, h-BASEBORDER-vs);
		    set_fcolor(CWHITE);
	    	rect(BASEBORDER-1, vrch-1, w-BASEBORDER*2+2, h-BASEBORDER-vs+2);
		    set_fcolor(CBLACK);
			if (nn)
				box(BASEBORDER-1, vrch-2, w-BASEBORDER*2+2,2);
			else
				box(BASEBORDER-1, vrch-1, w-BASEBORDER*2+2,1);
		    box(BASEBORDER-1, vrch-1, 1, h-BASEBORDER-vs+2);
		}
		else 
		{
			box(BASEBORDER-2, vrch-2, w-BASEBORDER*2+4, h-BASEBORDER-vs+4);
		}
	}
    else FGmemset(image+w*TITLEH +w, paper, w*(h-TITLEH-1)); // standard

	if (status & WSIZEABLE && ff)
	{
		set_fcolor(CYELLOW);
		box(w - 13, h - 3, 12, 2);
		box(w - 3,  h - 13, 2, 12);
	}
	if (mm)
	{
		if (ff)
		{
			set_fcolor(CBLACK);
			box(xoff-1, yoff - MENUH-2, w - xoff * 2+2, MENUH+2);
			set_fcolor(CWHITE);
		    box(xoff, vrch-MENUH, 1, MENUH+2);
		}
		else
		{
			set_fcolor(CBLACK);
			box(xoff, yoff - MENUH-1-(tt?1:0), w - xoff * 2, MENUH+1+(tt?1:0));
			set_fcolor(MENUBG);
			if (tt)
				box(xoff,MENUH+1, w-xoff*2, MENUH);
			else
				box(xoff,0, w-xoff*2, MENUH);
		}
		set_fcolor(MENUBG);
		box(xoff, yoff - MENUH-1, w - xoff * 2, MENUH);
		set_fcolor(CBLACK);
		box(xoff, yoff - MENUH-1, 1, MENUH);
	}
	if (ss)
	{
		if (ff)
		{
			set_fcolor(BORD2);
		    set_fcolor(BORD1);
		    set_fcolor(statuscolor);
        }
		else
		{
		    set_fcolor(statuscolor);
		    box(0, h-STATUSBARH, w, STATUSBARH);
		}
	}
	clipping_state(1);
	if (first_draw) SendEvent(this, REPAINTEVENT);

	clipping_state(0);

	// redraw notebooks
	if (status&WNOTEBOOK)
		DrawTabPages();

	BaseGui *pt = Buttony;
	while (pt)
	{
		((Control *) pt)->draw();
		pt = pt->GetNext();
	}
	clipping_state(1);
	first_draw = 1;
//	WindowDrawTitle(CTITLEACTIVE);
	if (current==this)
	{
		WindowStatus(WACTIVE);
		RepaintBlock(0, 0, w, h);
	}
	else
	{
		WindowStatus(WDEACTIVE);
		WindowRepaint(0, 0, w, h);
	}
}

void Window::frame(void)
{
	if (status&WFRAMED)
	{
		clipping_state(0);
		set_fcolor(BORD1);
		box(0, 0, 1, h);
		set_fcolor(BORD2);
		box(1, 1, 2, h-2);
		set_fcolor(0);
		box(3, 3, 1, h-6);

		set_fcolor(BORD1);
		box(w-4, 0, 1, h);
		set_fcolor(BORD2);
		box(w-3, 1, 2, h-2);
		set_fcolor(BORD3);
		box(w-1, 3, 1, h-6);
		clipping_state(1);
	}
}

void Window::WindowStatusBar(int x, char *s, int color)
{
	int y;
	x += xoff+1;
	y  = h-STATUSBARH+4-(status&WFRAMED?BASEBORDER:0);
	clipping_state(0);
// hack!
	wwrk-=4;
	WindowText(x, y, s, color, statuscolor);
	clipping_state(1);
}

FGRect Window::WindowFillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, int col)
{
	FGRect r;
	if (col != -1) set_fcolor(col);
	r = ftriangle(x1,y1,x2,y2,x3,y3);
	if (r.w) WindowRepaintUser(r.x, r.y, r.w, r.h);
	return r;
}

FGRect Window::WindowFillPolygon(int n, int p[][2], int col)
{
	FGRect r;
	if (col!=-1) set_fcolor(col);
	r = fpolygon(n, p);
	if (r.w) WindowRepaintUser(r.x, r.y, r.w, r.h);
	return r;
}

void Window::WindowDrawPolygon(int n, int p[][2], int col)
{
	if (col!=-1) set_fcolor(col);
	polygon(n, p);
	WindowRepaintUser(0,0,wwrk,hwrk);
}

void Window::WindowSpline(int points[8])
{
    #define NPTS   64

    int xpts[NPTS], ypts[NPTS];
    int i;

	calc_spline(points, NPTS, xpts, ypts);

	for (i=1; i<NPTS; i++) 
	{
    	WindowLine(xpts[i-1], ypts[i-1], xpts[i], ypts[i]);
        if (ppop == _GXOR) WindowPixel(xpts[i], ypts[i]);
	}
}

void SendEvent(Window *w, int event, int key, int x, int y)
{
	GuiEvent e(event,key,x,y);
	w->SendToWindow(&e);
}

void SendEvent(Window *w, int event, int x, int y, int ww, int hh)
{
	GuiEvent e(event,0,x,y,ww,hh);
	w->SendToWindow(&e);
}

void App::Timer(void)
{
	static unsigned int last=0,now=0;
	now = FGClock();
	if (last>now)
	{
		last = now;
		return;
	}
	if (now >= (last+ticks) || last>now)
	{
		last = now;
		__fg_fps_total = __fg_fps;
		__fg_fps = 0; // zero FPS counter
		if (OnEverySecond) OnEverySecond(now/1000);
	}
//printf("IntTimer = %ld [%ld]\n", IntCounter, now); //IntCounter=0;
}

static void InitCRC(void)
{
  int I, J;
  unsigned int C;

  for (I=0;I<256;I++)
  {
    for (C=I,J=0;J<8;J++)
      C=(C & 1) ? (C>>1)^0xEDB88320L : (C>>1);
    CRCTab[I]=C;
  }
}

unsigned int CalculateCRC(unsigned StartCRC, void *Addr, unsigned Size)
{
  unsigned I;
  if (!StartCRC) InitCRC();
  for (I=0; I<Size; I++)
    StartCRC = CRCTab[(unsigned char)StartCRC ^ ((unsigned char*)Addr)[I]] ^ (StartCRC >> 8);
  return(StartCRC);
}

void Window::WindowSetWorkRect(int x,int y,int w,int h)
{
	if (set_clip_rect(x,y,w,h)==0) return; // error!
	xpos = 2;
	ypos = 0;
}

//
// Configuration file handling	   (for fastgl)
//

int Config::find(char * s)
{
	for (int i=0;i<entries;i++)
	{
		if (!strcmp(pole[i].name,s)) return i;
	}
	return -1;
}

void Config::CreateEmpty(void)
{
	char s[100] = "; ";
	strcat(s,name);
	WriteComment(s);
	WriteComment(";<.......name of item..........> t <...body of item... (LF)");
	WriteComment("; where 'name' is string up to 32 chars length");
	WriteComment("; where 't' is one of 's'(tring), 'i(nt)' or 'd'(ouble)");
	WriteComment("; Warning! don't use TABs! & STRING MUST be enclosed in double \"\"");
	WriteComment("|..............................| T |.............................");
}

Config::Config(char * s, int write)
{
	char nn[128];
#if defined(__MSDOS__) || defined (_WIN32)
	sprintf(nn, "%s.rc", s);
#else
	if (App::homedir) sprintf(nn, "%s/.%s.rc", App::homedir, s);
	else sprintf(nn, ".%s.rc", s);
#endif
	strncpy(name, nn, 63);
	change = 0;
	entries = 0;
	writable = write;
	if (writable && verbose) ::printf("Opening config file '%s'\n", name);
	f = fopen(name,"r");
	if (f)
	{
		ReadConfig();
		fclose(f);
	}
	else
	{
		CreateEmpty();
	}
}

Config::~Config()
{
	if (writable) Sync();
}

void  Config::WriteComment(char * s)
{
	int i;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((i=find(s))==-1) pole[entries++].ModifyNode(s);
	else  pole[i].ModifyNode(s);
	change = 1;
}

void  Config::WriteInt(char * s, int i)
{
	int ii;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

void  Config::WriteString(char * s, char * i)
{
	int ii;
	if (strlen(i)>127) return;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

void  Config::WriteHex(char * s, char * i, int c)
{
	int ii;
	if (c>128) return;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i,c);
	else  pole[ii].ModifyNode(s,i,c);
	change = 1;
}

void  Config::WriteDouble(char * s, double i)
{
	int ii;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

static int ChToI(char xCH)
{
	if ((xCH >= '0') & (xCH <= '9'))
		return ((int) xCH) - '0';
	if ((xCH >= 'a') & (xCH <= 'f'))
		return ((int) xCH) - 'a' + 10;
	if ((xCH >= 'A') & (xCH <= 'F'))
		return ((int) xCH) - 'A' + 10;
	return -1;
}

static int hexdata(char *s)
{
	int a = ChToI(s[0]);
	int b = ChToI(s[1]);
	if (a!=-1 && b!=-1) return a*16 + b;
	return 256;
}

void Config::ReadConfig(void) 
{
	char s[360], *p, n[33], *t, hex[128];
	int typ,i;
	double d;
	
	for(;entries<MAX_CONF_ITEMS;)
	{
		*s = 0;
		fgets(s,359,f);
		if (!*s) break;
		*(s+strlen(s)-1) = 0;
		p = s;
		t = n;
		if (!isalpha(s[0]) && *s != '_') 
		{
			pole[entries].ModifyNode(s);
		}
		else
		{
			while(isalnum(*p) || *p=='_') (*t++) = (*p++);
			*t = 0;
			switch(typ = s[33])
			{
				case Node::_DOUBLE:
					d = atof(s+35);
					pole[entries].ModifyNode(n,d);
					break;
				case Node::_INT:
					i = atoi(s+35);
					pole[entries].ModifyNode(n,i);
					break;
				case Node::_STRING:
					t = s+36;
					while(*t!='"') t++;
					*t = 0;
					pole[entries].ModifyNode(n,s+36);
					break;
				case Node::_HEXA:
					t = s+36;
					i = 0;
					while((typ=hexdata(t))<256 && i<128)
					{
						hex[i++] = typ;
						t+=2;
					}
					pole[entries].ModifyNode(n,hex,i);
					break;
			}
		}
		entries++;
	}
}

void Config::Sync(void)
{
	char s[360], hex[260],pom[8];
	int j;
	
	if (change==0) return;
	if (verbose) ::printf("Writing config file '%s'\n", name);
	errno = 0;
	f = fopen(name,"w");
	if (f)
	{
		for(int i=0;i<entries;i++)
		{
			switch(pole[i].type)
			{
				case Node::_COMMENT:
					fputs(pole[i].s,f);
					break;
				case Node::_STRING:
					sprintf(s, "%-32s %c \"%s\"", pole[i].name, pole[i].type, pole[i].s);
					fputs(s,f);
					break;
				case Node::_INT:
					sprintf(s, "%-32s %c %d", pole[i].name, pole[i].type, pole[i].i);
					fputs(s,f);
					break;
				case Node::_DOUBLE:
					sprintf(s, "%-32s %c %f", pole[i].name, pole[i].type, pole[i].d);
					fputs(s,f);
					break;
				case Node::_HEXA:
					*hex = 0;
					for(j=0;j<pole[i].len;j++)
					{
						sprintf(pom, "%02x", pole[i].s[j]&255);
						strcat(hex,pom);
					}
					sprintf(s, "%-32s %c \"%s\"", pole[i].name, pole[i].type, hex);
					fputs(s,f);
					break;
			}
			fputc('\n',f);
		}
		fclose(f);
	}
	if (errno) perror("Config::Sync() ");
}

BaseGui::BaseGui(int xs, int ys, int ws, int hs, char *nm, Type typ, FGPixel i, FGPixel p, long flag)
	: DrawBuffer(ws,hs,nm,typ,i,p)
{
	set_ppop(_GSET);
	status = flag;
	next = (BaseGui	*)0;
	if (typ==WINDOW	&& flag&WUSELAST && entryPoint) entryPoint->Add(xs, ys, ws, hs, name,flag);
	if (typ	!= ROOTWINDOW)
		PositionCheck(xs, ys, ws, hs);
	if (typ==WINDOW	&& flag&WCENTRED)
	{
		xs = (X_width-ws)/2;
		ys = (Y_width-hs)/2;
	}
	x =	xs;
	y =	ys;
	Resize(ws-w, hs-h);
	if (!MOVEABLE) status |= WUNMOVED;
	if (status & WMODAL) status &= ~WMINIMIZE;
}

void BaseGui::Rechain(BaseGui *l, BaseGui *n)  // place to object chain
{
	if (!n && !l) // no more windows ...
	{
	}
	else if	(n==(BaseGui *)0)	// last in queque
	{
		l->next	= (BaseGui *)0;
	}
	else if	(l==(BaseGui *)0)	// first in queqe
	{
		n->last	= (BaseGui *)0;
	}
	else if	(n && l)
	{
		l->next	= n;
		n->last	= l;
	}
}

BaseGui::~BaseGui()
{
	if (type == ROOTWINDOW) return;
	Rechain(last,next);
}

void Window::SendToWindow(GuiEvent *e)
{
	if (image==0)
	{
		IError("Internal Error 1 (SendToWindow)",	1);
		return;
	}

	assert(this);
	e->wnd = this;

	switch(e->GetType())
	{
		case REPAINTEVENT:
			OnPaint();
			break;
		case GETFOCUSEVENT:
			OnFocus();
			break;
		case LOSTFOCUSEVENT:
			OnLostFocus();
			break;
//		case INITEVENT:
//			OnInit();
//			break;
		case KEYEVENT:
			OnKeyPress(e->GetKey());
			break;
		case MOVEEVENT:
			OnMouseMove(e->GetX(), e->GetY());
			break;
		case ACCELEVENT:
			OnAccelerator(e->accel);
			break;
		case WINDOWMOVEEVENT:
			OnMove(e->GetX(), e->GetY());
			break;
		case WINDOWRESIZEEVENT:
			OnResize(e->GetW(), e->GetH());
			break;
		case BUTTONHOLDEVENT:
			break;
		case STARTDRAGLEFTEVENT:
			OnStartDrag(1, e->GetX(), e->GetY());
			break;
		case STARTDRAGRIGHTEVENT:
			OnStartDrag(0, e->GetX(), e->GetY());
			break;
		case DRAGLEFTEVENT:
			OnEndDrag(1, e->GetX(), e->GetY(), e->GetW(), e->GetH());
			break;
		case DRAGRIGHTEVENT:
			OnEndDrag(0, e->GetX(), e->GetY(), e->GetW(), e->GetH());
			break;
		case CLICKLEFTEVENT:
			OnClick(e->GetX(), e->GetY());
			break;
		case CLICKRIGHTEVENT:
			OnContextPopup(e->GetX(), e->GetY());
			break;
//		case TERMINATEEVENT:
//			OnClose();
//			break;
		case ICONIZEEVENT:
			OnIconize();
			break;
		case NOTIFYEVENT: // not implemented
//			OnCloseQuery(e->GetKey());
			break;
//		default:
//			::printf("Unhandled message %d in WndHandler!", e->GetType());
	}
	if (handler)
		handler(e);
}

void MenuWindow::MenuWindowHandler(GuiEvent *p)
{
	if (Proc) Proc(p);
	if (CurrentMenu==0) return;
	switch(p->GetType())
	{
		case KEYEVENT:
			if (p->GetKey() !=	ESC) break;
			delete CurrentMenu;
			break;
		case ACCELEVENT:
			p->accel->RunHandler();
			if (!(p->guiType == LABEL || p->guiType == PUSHBUTTON || p->guiType ==	MENUBUTTON || p->guiType == MENUBUTTON2))  break;
		case LOSTFOCUSEVENT:
		case TERMINATEEVENT:
			if (CurrentMenu == p->wnd) delete CurrentMenu; // do not delete others
			break;
	}
}

void Window::default_clipping(void)
{
	if (status & WFRAMED)
		yoff += BASEBORDER;
	if (status & WTITLED)
		yoff += (TITLEH+2);
	if (status & WMENU)
		yoff += (MENUH+1);
	if (status & WNOTEBOOK)
		yoff += (NOTEBOOKH+1);
	if (status & WFRAMED)
		xoff += BASEBORDER;

	// set clipping window
	wwrk = w - xoff*2;
	hwrk = h - yoff;
	if (status & WFRAMED)
		hwrk -= BASEBORDER;
	if (status & WSTATUSBAR)
		hwrk -= STATUSBARH;
}

Window::Window(Window **wp, int xs, int ys, int ws, int hs, char *nm, GuiHwnd hnd, int i, int p, int flag):
  BaseGui(xs, ys, ws, hs, nm, flag & WMENUWINDOWTYPE ? MENUWINDOW : WINDOW, i, p, flag|WEXIST)
{
	Window *lst;

	// fix minimal size - 5.feb.2003
	if (w<8) w = 8;
	if (h<18) h = 18;

	init();
	rebuild = 1;
	if (entryPoint)
		entryPoint->DatabaseGetIcon(icon_x, icon_y, name);
	if (current) FlushInput();
	if (current) if (current->GetStatus()&WMENUWINDOWTYPE && GetCounter()>1)
	{
		GuiEvent e(TERMINATEEVENT);
		current->SendToWindow(&e);
	}

	Resize(0, 0);

	if (wp) *wp = this;			// save pointer to itself
	itself = wp;

	handler = hnd;
	if (nWnd)			// first window
	{
		lst = current;
	}
	else
	{
		lst = (Window *) 0;		// get addr of previous
		first = this;
	}
	current = this;
	nWnd++;
	SetLast(lst);
	statuscolor = CScheme->statusbar;
	if (lst)
	{
		GuiEvent e(LOSTFOCUSEVENT);
		lst->SendToWindow(&e);
		lst->WindowStatus(WDEACTIVE);
		lst->SetNext(this);
	}
	first_draw = 0;
	if (!(status&WICONWINDOWTYPE))
		draw();
	GuiEvent e(INITEVENT);
	default_clipping();
	set_colors(i,p);
	SendToWindow(&e);
	if ((lst && flag & WLASTFOCUS) || (lst && lst->GetStatus() & WMODAL) && nWnd)
		lst->WindowFocus();
	set_colors(i,p);
	idm = oldidm = 0;
	idw = oldidw = 0;
}

// ROOT wnd
Window::Window(Window **wp, FGPixel color):
  BaseGui(0, 0, X_width, Y_width, "Root", ROOTWINDOW, get_fcolor(), get_bcolor(), WEXIST|WUNMOVED)
{
	init();
	rebuild = 1;
	assert(!current);		// must be first
	Resize(0, 0);
	clear(color);

	if (wp) *wp = this;			// save pointer to itself
	itself = wp;
	assert(!nWnd);	
	first = this;
	current = this;
	nWnd++;
	SetLast(0);
	DoRootWindow = 1;
}

void Window::RemoveControls(void)
{
	BaseGui *butony = Buttony, *butony2;
	while (butony)
	{
		butony2 = butony->GetNext();
		delete butony;
		butony = butony2;
		nControls--;
	}
	Buttony = 0;
	GuiEvent::AutoRepeatEnd();
	c_menupos = 1;
	idm = oldidm = 0;
	idw = oldidw = 0;
}

void Window::EnableControls(void)
{
	Control *butony = (Control *)Buttony, *butony2;
	while (butony)
	{
		butony2 = (Control *)butony->GetNext();
		butony->Enable();
		butony = butony2;
	}
	idm = oldidm = 0;
	idw = oldidw = 0;
}

void Window::DisableControls(void)
{
	Control *butony = (Control *)Buttony, *butony2;
	while (butony)
	{
		butony2 = (Control *)butony->GetNext();
		butony->Disable();
		butony = butony2;
	}
	idm = oldidm = 0;
	idw = oldidw = 0;
}

Window::~Window()
{
	BaseGui *posledny, *novy, *l, *n;

	if (iconized && iconized!=(Window *)-1)
		delete iconized;

	if (type!= ROOTWINDOW)
	{
		FlushInput();
		status &= ~WEXIST;
		GuiEvent e(TERMINATEEVENT);
		SendToWindow(&e);
    
		if (itself) *itself = 0;

		if (status&WNOTEBOOK)
			DeleteTabPages();
		else
			RemoveControls();
			
		if (!shut_down) WindowUpdate(0,0,1);		// delete all
	}
	posledny = GetLast();
	status = 0;			// no anythings
	if (nWnd<=1/* || type== ROOTWINDOW*/)			// last in system
	{
		first = current = 0;
		nWnd  = 0;
		return;
	}
	rebuild = 1;
	novy = this==current ? posledny:current;
	if (!novy)
	{
	    ::printf("Internal error - window list mismatch!\a\n");
	    exit(-1);
	}
	novy->SetNext(0);
	nWnd--;
	l = GetLast();
	n = GetNext();
	if (n == (Window *) 0)		// last in queqe
	{
		current = (Window *) l;
	}
	else if (l == (Window *) 0)	// first in queqe
	{
		first = (Window *) n;
	}
	idm = oldidm = 0;
	idw = oldidw = 0;
	((Window *) novy)->WindowStatus(WACTIVE);
}

void ButtonGroup::AddToGroup(Control *p, int activ)
{
	int i;
	if (count>=size)
	{
		IError("ButtonGroup - out of range!", 0);
		return; // maximum numbers reached
	}
	for(i=0;i<count;i++)
	{
		assert(p != array[i]);
	}
	i = p->GetType();
	if (i!=BaseGui::POINTBUTTON && i!=BaseGui::CHECKBUTTON && i!=BaseGui::PUSHBUTTON) return; // bad type
	if (type==0) 
	{
	    if (i==BaseGui::CHECKBUTTON || i==BaseGui::POINTBUTTON) type=1;
	    else type=2;
	}
	else if (type==1 && i==BaseGui::PUSHBUTTON) IError("ButtonGroup mismatch",0);
	else if (type==2 && (i==BaseGui::POINTBUTTON || i==BaseGui::CHECKBUTTON)) IError("ButtonGroup mismatch",0);
	array[count++] = p;
	p->RegisterToGroup(this); // say to button for group members
	if (activ && !curr) 
	{
	    curr = p;
	    if (type==1) p->SetTrigger(1);
	    else ((PushButton *)p)->Push();
	}
}

void Control::RefreshGroup(void) 
{ 
	if (grp) 
		grp->RefreshGroup(this); 
	idm = oldidm = 0;
}

void ButtonGroup::RefreshGroup(Control *c)
{
	PushButton *pb1=(PushButton *)c;
	switch(type)
	{
		case 1:
			if (curr) curr->SetTrigger(0);
			if (curr == c) // to same
			{
				curr->SetTrigger(1);
			}
			break;
		case 2:
			if (curr) ((PushButton *)curr)->Release();
			pb1->Push();
			break;
	}
	curr = c;
}

void Slider::MoveIn(void)
{
	set_ppop(_GSET);
	set_fcolor(CYELLOW);
	owner->clipping_state(0);
	owner->rect(GetX(),GetY(),GetW(),GetH());
	owner->clipping_state(1);
	owner->WindowRepaint(GetX(), GetY(), GetW(), GetH());
}

void Slider::MoveOut(void)
{
	set_ppop(_GSET);
	owner->clipping_state(0);
	frame(GetX(),GetY(),GetW(),GetH());
	owner->clipping_state(1);
	owner->WindowRepaint(GetX(), GetY(), GetW(), GetH());
}

void Slider::frame(int x, int y, int w, int h, int f) 
{
	set_fcolor(f?CWHITE:CBLACK);
	owner->vline(x, y, h);
	owner->hline(x, y, w);
	set_fcolor(CGRAYED);
	owner->vline(x + w - 2, y + 1, h - 2);
	owner->hline(x + 1, y + h - 2, w - 2);
	set_fcolor(f?CBLACK:CWHITE);
	owner->vline(x + w - 1, y + 0, h - 0);
	owner->hline(x + 0, y + h - 1, w - 1);
}

void Slider::ClickDown(void)
{
	if (smer==0) modify((GetMouseX()-owner->GetX())-GetX()-10);
	else modify((GetMouseY()-owner->GetY())-GetY()-10);
}

void Slider::modify(int value)
{
	if (value>= 0)
	{
	    *val = maxv;
	}
	else if (value < (maxv-minv)/steps+8)
	{
	    *val = minv;
	}
	if (value>= 0 && value < (maxv-minv)/steps)
	{
	    *val = minv+value*steps;
	}
	draw();
}

void SlideBarV::draw(void)
{
	static char str[40];
	unsigned dl=0, a = set_font(FONT0808);
	set_ppop(_GSET);
	owner->clipping_state(0);
	switch (status & 7)
	{
		case BNORMAL:
			set_fcolor(CScheme->slider);
			owner->box(GetX() + 1, GetY() + 1, GetW() - 3, GetH() - 3);
			frame(GetX(),GetY(),GetW(), dl=(maxv-minv)/steps+16+6);
			frame(GetX(),GetY()+dl,GetW(),GetH()-dl);
			frame(GetX()+2, GetY()+3+((*val-minv)/steps),w-4,w,1);
			set_colors(CBLACK, CScheme->slider);
			str[0] = char((*val%1000)/100+'0');
			str[1] = 0;
			str[2] = char((*val%100)/10+'0');
			str[3] = 0;
			str[4] = char(*val%10+'0');
			str[5] = 0;
			str[6] = char((*val%10000)/1000+'0');
			str[7] = 0;
			owner->text(GetX()+4,GetY()+dl+3,str+6);
			owner->text(GetX()+4,GetY()+dl+11,str);
			owner->text(GetX()+4,GetY()+dl+19,str+2);
			owner->text(GetX()+4,GetY()+dl+27,str+4);
			break;

		case BDISABLED:
			set_fcolor(CGRAY2);
			owner->box(GetX() + 1, GetY() + 1, GetW() - 3, GetH() - 3);
			frame(GetX(),GetY(),GetW(), dl=(maxv-minv)/steps+16+6);
			frame(GetX(),GetY()+dl,GetW(),GetH()-dl);
			frame(GetX()+2, GetY()+3+((*val-minv)/steps),w-4,w,1);
			set_colors(CScheme->slider_disable, CScheme->slider);
			str[0] = char((*val%1000)/100+'0');
			str[1] = 0;
			str[2] = char((*val%100)/10+'0');
			str[3] = 0;
			str[4] = char(*val%10+'0');
			str[5] = 0;
			str[6] = char((*val%10000)/1000+'0');
			str[7] = 0;
			owner->text(GetX()+4,GetY()+dl+3,str+6);
			owner->text(GetX()+4,GetY()+dl+11,str);
			owner->text(GetX()+4,GetY()+dl+19,str+2);
			owner->text(GetX()+4,GetY()+dl+27,str+4);
			break;
		case BUNVISIBLE:
			set_fcolor(GetPaper());
			owner->box(GetX(), GetY(), GetW(), GetH());
			break;
	}
	set_font(a);
	owner->clipping_state(1);
	owner->WindowRepaint(GetX(), GetY(), GetW(), GetH());
	if (fnc) fnc(this);
}

void SlideBarH::draw(void)
{
	static char str[40];
	unsigned dl=0, a = set_font(FONT0808);

	set_ppop(_GSET);
	owner->clipping_state(0);
	switch (status & 7)
	{
		case BNORMAL:
			set_fcolor(CScheme->slider);
			owner->box(GetX() + 1, GetY() + 1, GetW() - 3, GetH() - 3);
			frame(GetX(),GetY(),dl=(maxv-minv)/steps+16+6,GetH());
			frame(GetX()+dl,GetY(),GetW()-dl,GetH());
			frame(GetX()+3+((*val-minv)/steps),GetY()+2,h,h-4,1);
			set_colors(CBLACK, CScheme->slider);
			sprintf(str, "%3d", *val);
			owner->text(GetX()+dl+2,GetY()+4,str);
			break;

		case BDISABLED:
			set_fcolor(CGRAY2);
			owner->box(GetX() + 1, GetY() + 1, GetW() - 3, GetH() - 3);
			frame(GetX(),GetY(),dl=(maxv-minv)/steps+16+6,GetH());
			frame(GetX()+dl,GetY(),GetW()-dl,GetH());
			frame(GetX()+3+((*val-minv)/steps),GetY()+2,16,12,1);
			set_colors(CScheme->slider_disable, CGRAY2);
			sprintf(str, "%3d", *val);
			owner->text(GetX()+dl+2,GetY()+4,str);
			break;

		case BUNVISIBLE:
			set_fcolor(GetPaper());
			owner->box(GetX(), GetY(), GetW(), GetH());
			break;
	}
	set_font(a);
	owner->clipping_state(1);
	owner->WindowRepaint(GetX(), GetY(), GetW(), GetH());
	if (fnc) fnc(this);
}

void SlideBar::draw(void)
{
	unsigned dl, a = set_font(FONT0808);
	set_ppop(_GSET);
	owner->clipping_state(0);
	switch (status & 7)
	{
		case BNORMAL:
			set_fcolor(CScheme->slider);
			goto bdisa;
		case BDISABLED:
			set_fcolor(CGRAY2);
bdisa:		if (smer==0)
			{
				owner->box(GetX(), GetY(), GetW(), GetH());
				frame(GetX(),GetY(),dl=(maxv-minv)/steps+16+6,GetH());
				frame(GetX()+3+((*val-minv)/steps),GetY()+2,16,GetH()-4,1);
			}
			else
			{
				owner->box(GetX(), GetY(), GetW(), GetH());
				frame(GetX(),GetY(),GetW(), dl=(maxv-minv)/steps+16+6);
				frame(GetX()+2, GetY()+3+((*val-minv)/steps),GetW()-4,16,1);
			}
			break;

		case BUNVISIBLE:
			set_fcolor(GetPaper());
			owner->box(GetX(), GetY(), GetW(), GetH());
			break;
	}
	set_font(a);
	owner->clipping_state(1);
	owner->WindowRepaint(GetX(), GetY(), GetW(), GetH());
	if (fnc) fnc(this);
}

int Config::ReadInt(char * s)
{
	int	n =	find(s);
	if (n>=0) return pole[n].i;
	return 0;
}

char const * Config::ReadString(char * s)
{
	int	n =	find(s);
	if (n>=0) return (char	const *)pole[n].s;
	return "";
}

double Config::ReadDouble(char * s)
{
	int	n =	find(s);
	if (n>=0) return pole[n].d;
	return 0.0;
}
	
int	   Config::ReadInt(char	* s, int &p)
{
	int	n =	find(s);
	if (n>=0) return (p = pole[n].i), 1;
	return 0;
}

int	   Config::ReadString(char * s,	char *p, int max)
{
	int	n =	find(s);
	if (max>128) max = 128;
	if (n>=0)
	{
	 	strncpy(p,pole[n].s,max);
		return 1;
	}
	return 0;
}

int	   Config::ReadString(char * s,	char **p)
{
	int	n =	find(s);
	if (n>=0)
	{
	 	*p = pole[n].s;
		return 1;
	}
	return 0;
}

int	   Config::ReadDouble(char * s,	double &p)
{
	int	n =	find(s);
	if (n>=0) return (p = pole[n].d), 1;
	return 0;
}

int	   Config::ReadHex(char * s, char *p, int max)
{
	int	n =	find(s);
	if (max>128) max = 128;
	if (n>=0)
	{
	 	memcpy(p, pole[n].s, max);
		return 1;
	}
	return 0;
}

void MenuWindow::Separator(void)
{
	int i;
	WindowBox(0, i=GetYM(8)+3, GetWW(), 1, CBLACK);
	WindowBox(0, i+1, GetWW(), 1, CWHITE);
}

PushButton *Window::AddPushButton(int xs, int ys, int ws, int hs, char *nm,	int	key, ControlCall f, int flag)
{
	return new PushButton(xs, ys, ws, hs, nm, key, flag,this,f);
}

PushButton *Window::AddPushButton(int xs, int ys, int key, Bitmap *bm, ControlCall f, int flag)
{
	if (bm->GetType()==0) return	0;
	return new PushButton(xs, ys, key, flag,this,bm, f);
}

SlideBarH *Window::AddSlideBarH(int xs, int ys, int min, int max, int step, int *val, ControlCall f)
{
	return new SlideBarH(xs, ys, min, max, step, val, this,f);
}

SlideBarV *Window::AddSlideBarV(int xs, int ys, int min, int max, int step, int *val, ControlCall f)
{
	return new SlideBarV(xs, ys, min, max, step, val, this,f);
}

SlideBar *Window::AddSlideBar(int xs, int ys, int w, int h, int *val, ControlCall f)
{
	int a = w>h;
	return new SlideBar(xs, ys, a?w+16+6:w, a?h:h+16+6, a?w:h, val, this,f);
}

TwoStateButton	*Window::AddTwoStateButton(int	xs,	int	ys,	int w, int h, char *nm, int key, int * variable, ControlCall f)
{
	return new TwoStateButton(xs, ys, w, h, nm, key, this, variable, f);
}

CheckButton	*Window::AddCheckButton(int	xs,	int	ys,	char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new CheckButton(1,xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

PointButton	*Window::AddPointButton(int	xs,	int	ys,	char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new PointButton(1,xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

CheckButton	*Window::AddCheckButtonMask(int m, int	xs,	int	ys,	char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new CheckButton(m, xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

PointButton	*Window::AddPointButtonMask(int m, int	xs,	int	ys,	char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new PointButton(m, xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

EditBox	*Window::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f, int min, int max)
{
	return new EditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, min,	max, 1);
}

EditBox	*Window::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f)
{
	return new EditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, 0, 0, 0); // no checking
}

EditBox	*Window::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f)
{
	return new EditBox(0, xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

EditBox	*Window::AddEditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f)
{
	return new EditBox(sz, xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

EditBox	*Window::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, double *p, ControlCall f, double min, double max)
{
	return new EditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, min,	max, 1);
}

EditBox	*Window::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, double *p, ControlCall f)
{
	return new EditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, 0, 0, 0); // no checking
}

BaseMenu *Window::AddBaseMenu(char *nm,	int	key, ControlCall f)
{
   	int a = set_font(FGFONT_MENU);
	BaseMenu *menu = new BaseMenu(GetXM(textwidth(nm)+8), GetYM(), textwidth(nm)+8,	fonth,	nm,	key, this, CScheme->menu_back, CScheme->menu_fore, f, MENUBUTTON);
	set_font(a);
	return menu;
}

BaseMenu *MenuWindow::AddMenu(char *nm,	int	key, ControlCall f)
{
   	int a = set_font(FGFONT_MENU);
	BaseMenu *menu = new BaseMenu(GetXM(), GetYM(fonth+4), GetWW()-8, fonth+4, nm, key, this, GetPaper(), GetInk(), f, MENUBUTTON2);
	set_font(a);
	return menu;
}

Label *MenuWindow::AddLabel(char *nm, int key, ControlCall f, int i, int p)
{
	return new Label(GetXM(), GetYM(18),nm,key,this,f,i==-1?this->ink:i, p==-1?this->paper:p);
}

CheckButton	*MenuWindow::AddCheckButton(char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new CheckButton(1, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

PointButton	*MenuWindow::AddPointButton(char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new PointButton(1, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

CheckButton	*MenuWindow::AddCheckButtonMask(int m, char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new CheckButton(m, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

PointButton	*MenuWindow::AddPointButtonMask(int m, char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new PointButton(m, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

EditBox	*MenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f, int min, int max)
{
	return new EditBox(GetXM(),	GetYM(24), ws1,	ws2, nm, key, this, p, GetInk(),GetPaper(), f, min, max, 1);
}

EditBox	*MenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f)
{
	return new EditBox(GetXM(),	GetYM(24), ws1,	ws2, nm, key, this, p, GetInk(),GetPaper(),f, 0, 0, 0);
}

EditBox	*MenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f)
{
	return new EditBox(0, GetXM(),	GetYM(24), ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

EditBox	*MenuWindow::AddEditBox(int sz, int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f)
{
	return new EditBox(sz, GetXM(),	GetYM(24), ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

EditBox	*MenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p, ControlCall f, double min, double max)
{
	return new EditBox(GetXM(),	GetYM(23), ws1,	ws2, nm, key, this,	p, GetInk(), GetPaper(),	f, min,	max, 1);
}

EditBox	*MenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,ControlCall f)
{
	return new EditBox(GetXM(),	GetYM(23), ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, 0, 0, 0);
}

/*
ListBox * Window::AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int *cnt, int *pos, void (*drawone)(int,int,int,int,void *), void *data)
{
	return new ListBox(xs, ys, wpol, hpol, w, h, cnt, pos, drawone, this, data);
}

ListBox * Window::AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int cnt, void (*drawone)(int,int,int,int,void *), void *data)
{
	return new ListBox(xs, ys, wpol, hpol, w, h, cnt, drawone, this, data);
}
*/

FGListBox * Window::AddListBox(int xs, int ys, int w, int h, int dropdown, FGListBoxCallBack drawone)
{
	return new FGListBox(xs, ys, w, h, dropdown, this, drawone);
}

FGListBox * Window::AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int *cnt, int *pos, FGListBoxCallBack drawone, void *data)
{
	FGListBox *l = new FGListBox(xs, ys, w, h, hpol, this, drawone);
	l->SetSize(*cnt);
	l->SetIndex(*pos);
	return l;
}

FGListBox * Window::AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int cnt, FGListBoxCallBack drawone, void *data)
{
	FGListBox *l = new FGListBox(xs, ys, w, h, hpol, this, drawone);
	l->SetSize(cnt);
	l->SetIndex(0);
	return l;
}

void ProgressBar::draw(void)
{
	int tmp;
	char s[16];
 	int fnt=set_font(2);

	tmp = (int)((w/(float)steps)*value);
	if (tmp!=sirka) // changed
	{
		set_ppop(_GSET);
		set_fcolor(CWHITE);
		owner->clipping_state(0);
		owner->rect(x-2,y-2,w+4,h+4);
		owner->rect(x,y,w+1,h+1);
		set_fcolor(CBLACK);
		owner->rect(x-2,y-2,w+3,h+3);
		owner->rect(x-1,y-1,w+1,h+1);
		set_fcolor(paper);
		owner->box(x,y,tmp,h);
		set_fcolor(ink);
		owner->box(x+tmp,y,w-tmp,h);
		sprintf(s,"%d%%", (int)((100./steps)*value)+1);
		if (tmp>=textwidth(s))
		{
			set_colors(ink,paper);
			owner->text(x+tmp/2-textwidth(s)/2,y+h/2-8,s);
		}
		owner->WindowRepaint(x-2,y-2,w+4,h+4);
		owner->clipping_state(1);
	}
	sirka = tmp;
	set_font(fnt);
}

ProgressBar::ProgressBar(Window *win, int xx, int yy, int ww, int hh, int s)
	: Control(xx, yy, ww, hh, 0, PROGRESSBAR, 0, BNORMAL, win, 0)
{
	if (s<=0)
	{
		 s=100;
	     IError("ProgressBar: (step <= 0)", 0);
	}
	steps = s;
	value = 0;
	sirka = -1;
	ink = CWHITE;
	paper = CBLUELIGHT;
	draw();
}

ProgressBar *Window::AddProgressBar(int xs, int ys, int ws, int hs, int size)
{
    return new ProgressBar(this, xs,ys,ws,hs,size);
}

void Label::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	MoveOut();
	if (a) 
	{
		// if error, set cControl to 0
		SetActive();
	}
}

void Label::draw(void)
{
	int tmp = set_font(font_size);
	w = textwidth(name);
	h = fonth;
	if (is_transparent)
		set_ppop(_GTRANSP);
	else
		set_ppop(_GSET);
	set_colors(ink+high_light,paper);
	owner->clipping_state(0);
	owner->text(x,y,name);
	owner->clipping_state(1);
	owner->WindowRepaint(x,y,w,h);
	set_font(tmp);
	set_ppop(_GSET);
}

Label *Window::AddLabel(int x, int y, char *s, int key, ControlCall f, int i, int p)
{
	return new Label(x,y,s,key,this,f,i==-1?this->ink:i, p==-1?this->paper:p);
}
void Control::RunHandler(void)
{
	if (this)
	{
		cControl = this;
		if (call_handler) 
		    if (call_handler->fnc) 
			call_handler->fnc(call_handler);
		call_handler = 0; // no more call it
	}
}

void BaseGui::Resize(int dx, int dy)
{
	DrawBuffer::Resize(dx, dy);
	if (type == WINDOW && entryPoint) entryPoint->DatabaseResize(w,h,name); 
}

void Window::Relocate(int xx, int yy)
{ 
	x += xx;
	y += yy; 
	if (entryPoint)
		entryPoint->DatabaseRelocate(x,y,name);
	Control *p = (Control *) GetCurrentControls();
	while (p)
	{
		p->Relocate(xx, yy);
		p = (Control *) p->GetNext();
	}
}

static void Yes_No(CallBack t)
{
	GuiEvent e(NOTIFYEVENT, t->GetLocalId());
	delete _Note_Ptr;
	close_adept->SendToWindow(&e);
	close_adept = 0;
}

int Window::ShowNotify(void)
{
	if (status&WNOTIFY && !close_adept && changed)
	{
		close_adept = this;
		_Note_Ptr = new Window(&_Note_Ptr, 320, 334, 308, 148, "Note!", 0, CYELLOW, CREDLIGHT, 0xa7);
		_Note_Ptr->AddPushButton(168, 72, 64, 21, "No", 'n', Yes_No);
		_Note_Ptr->AddPushButton(56, 72, 64, 21, "Yes", 'y', Yes_No);
		_Note_Ptr->WindowText(32, 24, "Text has been changed, save ?");
		return 1;
	}
	else if (status&WNOTIFY && !close_adept)
 // not changed -> close immediatelly
	{
		GuiEvent e(NOTIFYEVENT, 0);
		SendToWindow(&e);
		return 1;
	}
	return 0;
}

int	Window::TestNullHandler(void)
{
	if (current->ShowNotify()) return 1;
	if (current->status&WNOPICTO)
	{
		IError("Can`t close this window",0);
		return 1;
	}
	return 0;
}

static void ShowEfect(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2)
{
	int s=25, xs=(x2-x1)/s, ys=(y2-y1)/s, ws=(w2-w1)/s, hs=(h2-h1)/s;
	for(;s;s--)
	{
		draw_box(x1,y1,w1,h1,0,_GNOT);
		draw_box(x1+1,y1+1,w1,h1,0,_GNOT);
		draw_box(x1+1,y1+1,w1,h1,0,_GNOT);
		draw_box(x1,y1,w1,h1,0,_GNOT);
		x1 += xs;
		y1 += ys;
		w1 += ws;
		h1 += hs;
	}
}

int Window::iconize(int xs, int ys)
{
	int rc;
	if (!(status&WMINIMIZE)) return 0; // bad window
	if (iconized==0)
	{
		if (fulldrag && icon) ShowEfect(x,y,w,h,xs,ys,icon->GetW(),icon->GetH());
		WindowStatus(WHIDE);
		if (status&WNOICON) iconized = (Window *)-1;
		else
		{
			iconized = new IconWindow(this, xs, ys, icon);
			int f = set_font(0);
			iconized->WindowText(0,0,name);
			set_font(f);
		}
		icon_x = xs;
		icon_y = ys;
		rc = 1;
	}
	else if (iconized != (Window *)-1)
	{
		icon_x = iconized->x;
		icon_y = iconized->y;
		if (fulldrag && icon) ShowEfect(icon_x, icon_y, icon->GetW(),icon->GetH(),x,y,w,h);
		delete iconized;
		iconized=0;
		WindowStatus(WVISIBLE);
		WindowFocus();
		rc = 2;
	}
	else // hide and no icon
	{
		if (fulldrag && icon) ShowEfect(icon_x, icon_y, icon->GetW(),icon->GetH(),x,y,w,h);
		iconized=0;
		WindowStatus(WVISIBLE);
		WindowFocus();
		rc = 2;
	}
	if (entryPoint) entryPoint->DatabaseSetIcon(icon_x, icon_y, name);
	GuiEvent e(ICONIZEEVENT, IsIconized());
	SendToWindow(&e);
	return rc;
}

int Window::WindowIconize(void)
{
	if (icon_x==-1 && icon_y==-1)
		return iconize(GetMouseX()-16, GetMouseY()-16);
	else
		return iconize(icon_x, icon_y);
}

int Window::WindowIconize(int xs, int ys)
{
	return iconize(xs, ys);
}

void Window::init(void)
{
	c_menupos = 1;
	itself = 0;
	handler = 0;
	Buttony = 0;
	tab_page = 0;
	nControls = 0;
	CallWindowHook = 0;
	changed = 0;
	iconized = 0;
	icon = 0;
	ypos = 0;
	xpos = 1;
	icon_x = -1;
	icon_y = -1;
	status |= WCLIPPED;
	status &= ~(WHIDEN|WLOCKED);
	curr_ctrl = -1;
	current_tab_page = "";
}

void Window::WindowPanel(int x, int y, int a, int b, char *s, int ink1, int ink2)
{
	set_fcolor(ink1);
	hline(x, y, a);
	vline(x, y, b);
	set_fcolor(ink2);
	hline(x, y + b - 1, a);
	vline(x	+ a - 1, y, b);
	WindowRepaintUser(x, y, a, b);
	if (s) WindowText(x+8,y-8,s);
}

void SetDebug(int l)
{
    __fgl_log_level = l;
}

void Window::RepaintBlock(int x, int y, int w, int h)
{
	RamToVideo(image,x, y, this->GetX()+x, this->GetY()+y,
		w, h, GetW(), GetH(),
		BITBLIT_COPY);
}

void Window::PixelW(int x, int y)
{
	cinstance->putpixel(x,y);
}


void Window::HlineW(int x, int y, int w)
{
	cinstance->hline(x,y,w);
}

void Control::Close(CallBack cb) 
{
	delete cb->owner;
}

void Control::Quit(CallBack)
{ 
	App::AppDone();
}

void App::SendToApp(GuiEvent	*x)
{
	if (quequeIndex<MAX_EVENT_QUEQUE)
	{
		memcpy(queque+quequeIndex++,x,sizeof(GuiEvent));
	}
	else IError("Message queque is full!",0);
}

EditBox *Window::FindNextEditBox(void)
{
	BaseGui *c = ininput;
	while(c->GetNext())
	{
		c = c->GetNext();
		if (c->GetType()==EDITBOX && (c->GetStatus()&7) != BDISABLED)
			return (EditBox *)c;
	}
	c = Buttony;
	while(c && c!=ininput)
	{
		if (c->GetType()==EDITBOX && (c->GetStatus()&7) != BDISABLED)
			return (EditBox *)c;
		c = c->GetNext();
	}
	return 0;
}

void Window::HideAll(void)
{
	App::GetRootWindow()->set_fcolor(cApp->background);

	if (!(cApp->flags&APP_ROOTWINDOW)) clear_frame_buffer(cApp->background);
	else App::GetRootWindow()->WindowUpdateBlock(0,0,X_width,Y_width);
}

void Window::ShowAll(void)
{
	BaseGui *p = (BaseGui *)first;
	Window *pw;
	
	while (p)
	{
		pw = (Window *)p;
		pw->WindowUpdateBlock(0,0, pw->GetW(), pw->GetH());
		p = p->GetNext();
	}
}

void TwoStateButton::draw(void)
{
	int a = set_font(FGFONT_BUTTON);
	int dl= textwidth(name);	// maximalny zobrazitelny pocet znakov
	int xx = w / 2 - dl / 2, yy = h / 2 - fonth / 2;
	set_ppop(_GSET);
	frame();
	owner->clipping_state(0);
	if (status & WTRIGGER)
	{
		if (variable) *variable |= mask;
			set_fcolor(CScheme->button_bord1);
			owner->box(x,y,w,h);
			set_fcolor(CScheme->button_fore_pushed);
			owner->vline(x + w - 2, y + 1, h - 2);
			owner->hline(x + 1, y + h - 2, w - 2);
			owner->rect(x, y, w, h);
			set_fcolor(CGRAY3);
			owner->box(x + 1, y + 1, w - 3, h - 3);
			set_bcolor(CGRAY3);
			set_fcolor(CScheme->button_fore);
			owner->text(x + xx, y + yy, name);
			Underscore(xx , yy, CDARK);
	}
	else
	{
		if (variable) *variable &= ~mask;
				set_fcolor(CScheme->button_back);
				owner->box(x + 1, y + 1, w - 3, h - 3);
				set_bcolor(CScheme->button_back);
				set_fcolor(CScheme->button_fore);
				owner->text(x + xx, y  + yy, name);
				Underscore(xx , yy, CDARK);
	}
	owner->clipping_state(1);
	set_font(a);
	owner->WindowRepaint(x, y, w, h);
}

// last in list
BaseGui	*Window::GetLastButton(void)
{
	if (Buttony == 0) return 0;
	BaseGui *p = Buttony;
	while(p->GetNext())
	{
		p = p->GetNext();
	}
	return p;
}

//
// Add current controls to the named tabpage
//
void Window::AddTabPage(const char *name)
{
	if ((status & WNOTEBOOK) == 0)
	{
		IError("You must specify WNOTEBOOK window flag before AddTabPage()", 0);
		return;
	}
	FGTabPage *tp = new FGTabPage(name, Buttony, yoff-NOTEBOOKH-2);
	if (tab_page == 0) // first use
	{
		tab_page = new TabPageList;	// create an empty list
	}
	Buttony = 0;	// reset for next TabPage
	// insert TabPage to the list
	TabPageList * p = (TabPageList *)tab_page;
	p->push_back(*tp);
	// redraw the whole window
	draw();
}

//
// Set current named tabpage
//
FGTabPage * Window::SetTabPage(const char *name)
{
	FGTabPage *result=0;
	
	if ((status & WNOTEBOOK) == 0 || tab_page == 0)
	{
		IError("You must specify WNOTEBOOK window flag before SetTabPage()", 0);
		return 0;
	}
	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0) return 0;
	TabPageListPtr iter = l->begin();

	while(iter != l->end())
	{
		iter->SetFlag(0);
		if (strcmp(iter->GetString(), name) == 0)
		{
			Buttony = iter->GetControls();
			// redraw the whole window
			iter->SetFlag(1);
			current_tab_page = iter->GetString();
			GuiEvent::AutoRepeatEnd();
			c_menupos = 1;
		    	idm = oldidm = 0;
			idw = oldidw = 0;
			result = &*iter;
		}
		++iter;
	}
	draw();
	return result;
}

//
// Delete named tabpage
//
FGTabPage * Window::DeleteTabPage(const char *name)
{
	if ((status & WNOTEBOOK) == 0 || tab_page == 0)
	{
		IError("You must specify WNOTEBOOK window flag before DeleteTabPage()", 0);
		return 0;
	}
	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0) return 0;
	TabPageListPtr iter = l->begin();

	// make current
	FGTabPage *curr = SetTabPage(name);
	if (curr == 0)
		return 0;
	// delete all current controls
	RemoveControls();
	// make a first current
	if (l->size())
		curr = SetTabPage(iter->GetString());
	else
	{
		// redraw the whole window
		draw();
		// the list is empty
		return 0;
	}
	// !!! return current !!!
	return curr;
}

void Window::DeleteTabPages(void)
{
	if ((status & WNOTEBOOK) == 0 || tab_page == 0)
	{
		return;
	}
	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0) return;
	TabPageListPtr iter = l->begin();
	if (l->size()) while(iter != l->end())
	{
		Buttony = iter->GetControls();
		RemoveControls();
		++iter;
	}
	delete l;
	tab_page = 0;
}

//
// Redraw bookmarks only
//
void Window::DrawTabPages(void)
{
	if (tab_page == 0)
		return;
	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0)
		return;
	TabPageListPtr iter = l->begin();
	if (l == 0) return;
	int _w = 0;
	int f = set_font(FONTSYSLIGHT);
	
	if (l->size()) while(iter != l->end())
	{
		int _y = iter->GetY();
		int _x = _w+4;
		int sirka = iter->GetWidth()-2;
		
		set_fcolor(paper);
		box(_x+1, _y+1, sirka-2, NOTEBOOKH-1);
		set_colors(0, paper);
		{
			rect(_x, _y+1, sirka, NOTEBOOKH);
			rect(_x+1, _y, sirka-2, NOTEBOOKH+1);
			box(_x, _y+NOTEBOOKH, sirka+2, 2);
		}

		if (iter->GetFlag() == 0)
		{
			set_fcolor(0);
			rect(_x, _y+1, sirka, NOTEBOOKH+1);
			set_fcolor(ink);
		}
		else
		{
			set_fcolor(paper);
			box(_x+2, _y+2, sirka-4, NOTEBOOKH);
			set_fcolor(CScheme->notebook_active);
		}
		text(_x+NOTEGAP-2, _y+4,iter->GetString());
		_w += iter->GetWidth();
		++iter;
	}
	set_font(f);
}

//
// Test bookmarks for click, and make one current if needed
//
int Window::ClickTabPage(int xx, int yy)
{
	int _w = 0;
	if ((status & WNOTEBOOK) == 0 || tab_page == 0)
	{
		return 0;
	}
	xx -= x;
	yy -= y;

	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0) return 0;
	TabPageListPtr iter = l->begin();

	// here we go for markers, return TRUE if gets one
	if (l->size()) while(iter != l->end())
	{
		int _y = iter->GetY();
		int _x = _w+4;
		int sirka = iter->GetWidth();

		if (_x<=xx && _x+sirka>xx)
			if (_y<=yy && _y+NOTEBOOKH+2>yy)
				if (iter->GetFlag() == 0)
				{
					SetTabPage(iter->GetString());
					return 1;
				}
				else return 0;
		_w += sirka;
		++iter;
	}
	return 0;
}

void Window::SetNextTabPage(void)
{
	if ((status & WNOTEBOOK) == 0 || tab_page == 0)
	{
		return ;
	}
	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0) return;
	TabPageListPtr iter = l->begin();
	// here we go for markers, return TRUE if gets one
	if (l->size()) while(iter != l->end())
	{
		if (iter->GetFlag() == 1)
		{
			++iter;
			if (iter != l->end())
				SetTabPage(iter->GetString());
			else
				SetTabPage(l->begin()->GetString());
			return;
		}
		++iter;
	}
}

// relokuj vsetky zalozky
void Window::RelocateTabPage(int xx, int yy)
{ 
	Control *p;
	x += xx;
	y += yy; 
	if (entryPoint)
		entryPoint->DatabaseRelocate(x,y,name);
	if ((status & WNOTEBOOK) == 0)
	{
		return ;
	}

	TabPageList *l = (TabPageList *)tab_page;
	if (l == 0) return;
	TabPageListPtr iter = l->begin();

	// here we go for markers, return TRUE if gets one
	if (l->size()) while(iter != l->end())
	{
		p = (Control *)iter->GetControls();
		while (p)
		{
			p->Relocate(xx, yy);
			p = (Control *) p->GetNext();
		}
		++iter;
	}
}

//
// Tun until window close or mrXXX happen
//
int Window::ShowModal(void)
{
	int st = status;
	status |= WMODAL;
	int rc = cApp->RunModal(this);
	status = st;
	return rc;
}

FontProperty::FontProperty(char *n, int sz)
{
#ifdef FG_TTF
	_font = TTF_OpenFont(n, sz);
	if (_font == 0) IError("TrueType font open error!",0);
	algorithm = 0;
#else
	IError("Using TTF with compiled-in support!", 0);
#endif
}

FontProperty::~FontProperty()
{
#ifdef FG_TTF
	TTF_CloseFont((TTF_Font *)_font);
#endif
}

void FontProperty::Bold(int value)
{
#ifdef FG_TTF
	int tmp = TTF_GetFontStyle((TTF_Font *)_font);
	if (value)
		TTF_SetFontStyle((TTF_Font *)_font, tmp | TTF_STYLE_BOLD);
	else
		TTF_SetFontStyle((TTF_Font *)_font, tmp & ~TTF_STYLE_BOLD);
#endif
}

void FontProperty::Italic(int value)
{
#ifdef FG_TTF
	int tmp = TTF_GetFontStyle((TTF_Font *)_font);
	if (value)
		TTF_SetFontStyle((TTF_Font *)_font, tmp | TTF_STYLE_ITALIC);
	else
		TTF_SetFontStyle((TTF_Font *)_font, tmp & ~TTF_STYLE_ITALIC);
#endif
}

void FontProperty::Underline(int value)
{
#ifdef FG_TTF
	int tmp = TTF_GetFontStyle((TTF_Font *)_font);
	if (value)
		TTF_SetFontStyle((TTF_Font *)_font, tmp | TTF_STYLE_UNDERLINE);
	else
		TTF_SetFontStyle((TTF_Font *)_font, tmp & ~TTF_STYLE_UNDERLINE);
#endif
}

void Window::WindowText(int x, int y, char *s, FontProperty * f, int color, int bk)
{
#ifdef FG_TTF
	DrawBuffer *text;
	switch(f->algorithm)
	{
		default:
		case 0:
			text = TTF_RenderText_Solid((TTF_Font *)f->_font, s, color, bk);
			break;
#ifndef BPP8
		case 1:
			text = TTF_RenderText_Shaded((TTF_Font *)f->_font, s, color, bk);
			break;
		case 2:
//			text = TTF_RenderText_Blended((TTF_Font *)f->_font, s, color, bk);
			break;
#endif
	}
	bitblit(x,y,0,0,text->GetW(), text->GetH(), text);
	WindowRepaintUser(x, y, text->GetW(), text->GetH());
	delete text;
#endif
}

void Window::WindowTextUTF8(int x, int y, char *s, FontProperty * f, int color, int bk)
{
#ifdef FG_TTF
	DrawBuffer *text;
	switch(f->algorithm)
	{
		default:
		case 0:
			text = TTF_RenderUTF8_Solid((TTF_Font *)f->_font, s, color, bk);
			break;
#ifndef BPP8
		case 1:
			text = TTF_RenderUTF8_Shaded((TTF_Font *)f->_font, s, color, bk);
			break;
		case 2:
//			text = TTF_RenderUTF8_Blended((TTF_Font *)f->_font, s, color, bk);
			break;
#endif
	}
	bitblit(x,y,0,0,text->GetW(), text->GetH(), text);
	WindowRepaintUser(x, y, text->GetW(), text->GetH());
	delete text;
#endif
}

void Window::WindowTextUnicode(int x, int y, const unsigned short *s, FontProperty * f,int color, int bk)
{
#ifdef FG_TTF
	DrawBuffer *text;
	switch(f->algorithm)
	{
		default:
		case 0:
			text = TTF_RenderUNICODE_Solid((TTF_Font *)f->_font, s, color, bk);
			break;
#ifndef BPP8
		case 1:
			text = TTF_RenderUNICODE_Shaded((TTF_Font *)f->_font, s, color, bk);
			break;
		case 2:
//			text = TTF_RenderUNICODE_Blended((TTF_Font *)f->_font, s, color, bk);
			break;
#endif
	}
	bitblit(x,y,0,0,text->GetW(), text->GetH(), text);
	WindowRepaintUser(x, y, text->GetW(), text->GetH());
	delete text;
#endif
}


