/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2003  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	nezmar@atlas.sk

	fastgl.h   - declarations of many objects for FastGL
*/

#ifndef	__WINDOW_H
#define	__WINDOW_H

/* HJH modification: protect compiler options for structure alignment and enum
 * size if the compiler is Borland C++ */
#include "base.h"
#pragma interface
#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#if	1
#define	FGFONT_TITLE	FONTSYS
#define	FGFONT_MENU		FONTSYSLIGHT
#define	FGFONT_BUTTON	FONTSYSLIGHT
#else
#define	FGFONT_TITLE	FONT0816
#define	FGFONT_MENU		FONT0816
#define	FGFONT_BUTTON	FONT0816
#endif

#define	IM		 	CScheme->window_fore // CBLACK standard menu fore & background colors
#define	PM			CScheme->window_back

#define	BNORMAL		7
#define	BPUSHED		3
#define	BDISABLED	5
#define	BUNVISIBLE	6

#define	WFRAMED		0x0001
#define	WTITLED		0x0002
#define	WMODAL		0x0004
#define	WNOTIFY		0x0008
#define	WHIDEN		0x0010
#define	WUNMOVED	0x0020
#define	WMENU		0x0040
#define	WNOPICTO	0x0080
#define	WBITMAP		0x0100
#define	WCLICKABLE	0x0200
#define	WDIRTY		0x0400
#define	WTEMPHIDEN	0x0800
#define	WLOCKED		0x1000
#define	WTRIGGER	0x2000
#define	WLASTFOCUS	0x4000
#define	WSIZEABLE	0x8000
#define	WUSELAST	0x00010000
#define	WMINIMIZE	0x00020000
#define	WSTATUSBAR	0x00040000
#define	WEXIST		0x00080000
#define	WGLCONTEXT	0x00100000
#define	WNOTEBOOK	0x00200000
//#define	WLINEMENU	0x00400000
#define	WCENTRED	0x00800000
#define	WESCAPE		0x01000000
#define	WALIGNED	0x02000000
#define	WFASTMOVE	0x04000000
#define	WCLIPPED	0x08000000
#define	WNOICON		0x10000000

#define	WMENUWINDOWTYPE	   0x80000000
#define	WICONWINDOWTYPE	   0x40000000

#define	WSTANDARD	(WCLICKABLE|WTITLED|WFRAMED)

#define	APP_WINDOWDATABASE		1
#define	APP_CFG					2
#define	APP_ENABLEALTX			4
#define	APP_MAGNIFIER			8
#define	APP_ROOTWINDOW			16
#define APP_ALL					0x1F
#define	APP_LOCALDIR			0x80000000

#define		mrNone		0
#define		mrOk		-1
#define		mrCancel	-2
#define		mrYes		-1
#define		mrNo		-2
#define		mrAll		-3
#define		mrIgnore	-4
#define		mrClose		-5
#define		mrQuit		-6

class BaseGui;
class Control;
class Window;
class Bitmap;
class EditBox;
class ButtonGroup;
class FGListBox;
class PushButton;
class PointButton;
class CheckButton;
class BaseMenu;
class PushButton;
class Slider;
class FGTabPage;

typedef void (*FGListBoxCallBack)(int flag, Window *wnd, int x, int y, void *data, int index);

// some global fnc

extern "C" void	IError(char	*s,	int	flag);
extern "C" int GetMouseX(void);
extern "C" int GetMouseY(void);

void SetRepeatDelay(int	c1,	int	c2);
void SaveScreen(void);
void FlushInput(void);
void SetWindowMoveStyle(int);
extern long IntCounter; // counter of userevent pass

typedef struct
{
	int		window_back, window_fore;
	int		active_title;
	int		inactive_title;
	int		wnd_bord1;
	int		wnd_bord2;
	int		wnd_bord3;
	int		statusbar;
	int		menu_back;
	int		menu_fore;
	int		menu_back_active;
	int		menu_fore_active;
	int		button_fore;
	int		button_back;
	int		button_fore_pushed;
	int		button_back_pushed;
	int		button_bord1;
	int		button_bord2;
	int		button_bord3;
	int		edit_back;
	int		edit_fore;
	int		edit_bord1;
	int		edit_bord2;
	int		slider;
	int		menuwindow_back;
	int		menuwindow_fore;
	int		pdmenu_back_active;
	int		pdmenu_fore_active;
	int		pdmenu_gray;
	int		edit_disable;
	int		notebook_active;
	int		slider_disable;
} ColorScheme;

// triangle
// subbitmap

extern ColorScheme * CScheme;

//
// for TrueType fonts
//
class FontProperty
{
		friend class Window;
		void *_font;
		int   algorithm;
	public:
		FontProperty(char *n, int sz);
		~FontProperty();
		void Bold(int value);
		void Underline(int value);
		void Italic(int value);
		void Solid(void) { algorithm = 0; }
		void Shaded(void) { algorithm = 1; }
		void Blended(void) { algorithm = 2; } // not implemented yet
		int  Valid(void) { return !!_font; }
};

class Node
{
	enum itemtype
	{
	_DOUBLE='d',
	_INT='i',
	_STRING='s',
	_COMMENT='c',
	_HEXA='h'
	};

	friend class Config;
	int type;
	char name[36];
	int	i;
	char  s[132];
	double d;
	int len;
	void ModifyNode(char * comment)	// Modify comment
	{
		type = _COMMENT;
		*name =	0;
		strncpy(s, comment, 128);
	}
	void ModifyNode(char * meno, char *	val) // Modify string
	{
		type = _STRING;
		strncpy(name, meno, 32);
		strncpy(s, val, 128);
	}
	void ModifyNode(char * meno, int val) // Modify integer
	{
		type = _INT;
		strncpy(name, meno, 32);
		i	 = val;
	}
	void ModifyNode(char * meno, double	val) // Modify double
	{
		type = _DOUBLE;
		strncpy(name, meno, 32);
		d	 = val;
	}
	void ModifyNode(char * meno, char *	val, int kolko) // Modify string
	{
		type = _HEXA;
		strncpy(name, meno, 32);
		len = kolko>128? 128 : kolko;
		memcpy(s, val, len);
	}
};

typedef unsigned int (*TimerCall)(unsigned int);

#define TIMER_RESOLUTION	10	/* Experimentally determined */
class FGTimer
{
	protected:
		friend void 				HandleAlarm(int sig);
		int						state;
		static void				StartTicks(void);
	public:
		FGTimer(unsigned int val, TimerCall cb)
		{
			state = SetTimer(val, cb);
		}
		FGTimer(TimerCall cb)
		{
			state = SetTimer(0, cb);
		}
		~FGTimer()
		{
			TimerQuit();
		}
		static int			SetTimer(unsigned int val, TimerCall cb);
		static unsigned int GetTicks(void);
		static void			Delay(unsigned int);
		static int			TimerInit(void);
		static void			TimerQuit(void);
};

class Config
{
		enum { MAX_CONF_ITEMS=500 };
	protected:
		char 	name[64];
		FILE 	*f;
		int		entries;
		int		change;
		Node 	pole[MAX_CONF_ITEMS];
		int		writable;

		int	find(char *	s);
		void CreateEmpty(void);
		void ReadConfig(void);
		void WriteComment(char * s);
	public:
		
		void	Sync(void);
		int		ReadInt(char	* s);
		char	const * ReadString(char * s);
		double	ReadDouble(char * s);
		int		ReadInt(char	* s, int &p);
		int		ReadString(char * s,	char *p, int max);
		int		ReadString(char * s,	char **p);
		int		ReadDouble(char * s,	double &p);
		int		ReadHex(char * s, char *, int);

		void	WriteInt(char * s, int);
		void	WriteString(char	* s, char *);
		void	WriteDouble(char	* s, double);
		void	WriteHex(char * s, char *, int);

		Config(char	*,int wrt=1);
		~Config();
};

//
//	data structure to event-driven ...
//

class GuiEvent
{
	protected:
		static	char *eventName[];
		int		type;
		int		key;
		int		buttons, x,	y, w, h;
		friend class BaseGui;
		friend class App;
		void AutoRepeatStart(Control *);
		void AutoRepeatDo(void);
	public:
		static void AutoRepeatEnd(void);
		Window	*wnd;	// address of destination
		int		guiType;
		Control *accel;
		GuiEvent(int a=NOEVENT,	int	b=0, int c=0, int d=0, int e=0, int f=0)
		{
			type= a;
			key	= b;
			x	= c;
			y	= d;
			w 	= e;
			h	= f;
			accel = 0;
		}
		// current
		int		GetKey(void)  const { return key;	}
		char	*GetName(void)  const { if (type<0 || type>=LASTEVENT) return eventName[LASTEVENT]; return eventName[type]; }
		int		GetType(void)  const { return	type; }
		int		GetButtons(void)  const {	return buttons;	}
		int		GetX(void)  const { return	x; }
		int		GetY(void)  const { return	y; }
		int		GetW(void)  const { return	w; }
		int		GetH(void)  const { return	h; }
		GuiEvent * GetUserEvent(int);
		void	TranslateUserEvent(void);
		
		static void	InitUserEvent(void);
		static void	CloseUserEvent(void);
};

typedef	void (*Widget)(void);
typedef	void (*ControlCall)(Control *);
typedef	void (*GuiHwnd)(GuiEvent *);
typedef	int	(*MainHwnd)(GuiEvent *);
typedef	Control * CallBack;

//
//	base objects for all graphics objects
//
class BaseGui : public DrawBuffer
{
	protected:
		enum bs	{ EBOX_LOW=1, EBOX_UPR ,BDOWN, BUP, WHIDE, WVISIBLE, WDEACTIVE, WACTIVE };
		BaseGui	*last, * next;
		int		 x,y;
		BaseGui(const BaseGui &)
		{
			::printf("No COPY Constructor defined!\n");
			exit(-1);
		}
		void		PositionCheck(int &x, int &y, int &w, int &h);
		BaseGui	* 	Over(BaseGui *od, BaseGui *po);
		BaseGui	* 	OverSprite(BaseGui *od,	int	x, int y, int w, int h);
		static	  	BaseGui	* FindId(BaseGui *,	int	idw);
		BaseGui	* 	FindName(BaseGui *,	char *);
		static	  	BaseGui	* Find(BaseGui *p,GuiEvent *e);
		BaseGui	* 	FindEvent(BaseGui *p,GuiEvent *e);
		void		Rechain(BaseGui *l, BaseGui *n);  // place to object chain
	public:
		int		  	IsVisible(void)  const	{ return !(status&(WHIDEN|WTEMPHIDEN));	}
		virtual		~BaseGui();
		BaseGui	* 	GetNext(void)  const {	return next; }
		BaseGui	* 	GetLast(void)  const {	return last; }
		virtual		void MoveIn(void){}
		virtual		void MoveOut(void){}
		void		SetLast(BaseGui *l) { last	= l; }
		void		SetNext(BaseGui *l) { next	= l; }
		int		  	GetX(void)  const { return	x; }
		int		  	GetY(void)  const { return	y; }
		void	  	Resize(int dx, int dy);
		BaseGui(int	xs,	int	ys,	int	ws,	int	hs,	char *nm, Type typ, FGPixel i, FGPixel p, long flag);

		virtual void OnFocus(void) { }
		virtual void OnLostFocus(void) { }
};

#ifndef __GNUC__
#pragma	pack(1)
#endif

class Bitmap : public DrawBuffer
{
protected:
	void init(void);
	struct RGB3
	{
		char bBlue;				 /* Blue component of the color definition */
		char bGreen;			 /* Green component of the color definition*/
		char bRed;				 /* Red component of the color definition  */
	};
	struct RGB4
	{
		char bBlue;				 /* Blue component of the color definition */
		char bGreen;			 /* Green component of the color definition*/
		char bRed;				 /* Red component of the color definition  */
		char fcOptions;			 /* Reserved, must be zero                 */
	};
	struct bitmap_file_header
	{
		unsigned short	usType;
		unsigned short	cbSizeLow, cbSizeHigh;
		short			reserve1, reserve2;
	};
	struct bitmap_info
	{
		unsigned short	imageOffset;
		short			reserve;
	};
	struct bitmap_info_header1
	{
	  unsigned long	 size;
	  unsigned short cx;			   /* Bit-map width in pels                  */
	  unsigned short cy;			   /* Bit-map height in pels                 */
	  unsigned short cPlanes;		   /* Number of bit planes                   */
	  unsigned short cBitCount;		   /* Number of bits per pel within a plane  */
	  unsigned long	 ulCompression;	   /* Compression scheme used to store the bitmap */
	  unsigned long	 cbImage;		   /* Length of bit-map storage data in chars*/
	  unsigned long	 cxResolution;	   /* x resolution of target device          */
	  unsigned long	 cyResolution;	   /* y resolution of target device          */
	  unsigned long	 cclrUsed;		   /* Number of color indices used           */
	  unsigned long	 cclrImportant;	   /* Number of important color indices      */
	};
	struct bitmap_info_header2
	{
	  unsigned long	 size;
	  unsigned long	 cx;			   /* Bit-map width in pels                  */
	  unsigned long	 cy;			   /* Bit-map height in pels                 */
	  unsigned short cPlanes;		   /* Number of bit planes                   */
	  unsigned short cBitCount;		   /* Number of bits per pel within a plane  */
	  unsigned long	 ulCompression;	   /* Compression scheme used to store the bitmap */
	  unsigned long	 cbImage;		   /* Length of bit-map storage data in chars*/
	  unsigned long	 cxResolution;	   /* x resolution of target device          */
	  unsigned long	 cyResolution;	   /* y resolution of target device          */
	  unsigned long	 cclrUsed;		   /* Number of color indices used           */
	  unsigned long	 cclrImportant;	   /* Number of important color indices      */
	};

		bitmap_file_header	bmfh;
		bitmap_info			bmi;
		union {
			bitmap_info_header1	bmih1;
			bitmap_info_header2	bmih2;
		} bmih;
		union {
			RGB3 rgb3[256];
			RGB4 rgb4[256];
		} rgb;

		long		*cUsedTable;
		char		*original_name;
		void		turn_bitmap(int cnt, FGPixel *from, FGPixel *to);
		void		DoPalette(int);
		int			pcx(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			gif(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			jpg(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			png(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			tga(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
	public:
		char *GetName(void) { return name; }
		char *GetOriginalName(void) { return original_name; }
		void SetOriginalName(char *s) { if (original_name) free(original_name); original_name = strdup(s); }
		Bitmap(char	*name);
		Bitmap(void	*memimage);
		Bitmap(Window *);
		Bitmap(int,int,FGPixel=CBLACK);
		virtual ~Bitmap();
		int			BitmapSave(char *);
};
#ifndef __GNUC__
#pragma	pack()
#endif

//
//	Base object fo controls (some buttons, menu & input boxes ..
//

class ButtonGroup
{
	protected:
		short	count;
		short	size;
		short	type; // 1-check,point; 2-push; 0 undefined
		Control * 	curr;
		Control **	array;
	public:
		ButtonGroup(int s=32)
		{
			size = s;
			count = type = 0;
			curr = 0;
			array = new Control *[size];
		}
		~ButtonGroup() { delete [] array; }
		void AddToGroup(Control *p, int activ = 0);
		void RefreshGroup(Control *c);
		void Clear(void) { type = count = 0; curr = 0; }
};

extern Control * call_handler;

class Control :	public BaseGui
{
	protected:
		ControlCall	fnc;			// handler
		Window		*owner;			// owner of this
		ButtonGroup 	*grp;		// if is group members
		int		key;			// hotkey
		int		mask;
		Bitmap		*icon;
		short	xr,	yr;			 // x,y real coordinates
#define IOSTATE_NORMAL		0x00
#define IOSTATE_TWOSTATE	0x02
		short	local_id, iomode;

		Control(int	xs,	int	ys,	int	ws,	int	hs,	char *nm, Type typ, int key, int flag, Window *, ControlCall f=0, FGPixel p=CGRAY1, FGPixel i=CDARK);
		Control(int	xs,	int	ys,	Type typ, int key, int flag,	Window *, Bitmap *,	ControlCall f=0);
		virtual		~Control();

		void		Underscore(int xo, int yo, int c);
		static		Control *	cControl; // current activated control
		void 		frame(void);
	public:
		void		RunHandler(void);
		ControlCall	GetHandler(void)  const { return fnc; }
		void		RegisterToGroup(ButtonGroup *g) { grp = g;}
		void		RemoveFromGroup(void) { grp = 0; }
		virtual		void SetName(char *s) {	BaseGui::SetName(s); draw(); }
		void		SetKey(int k) {	if (k>='a' && k<='z') k	= toupper(k); key=k; draw(); }
		int			GetKey(void) const { return key; }
		void		SetHandler(ControlCall w) {	fnc=w; }
		virtual		void draw(void)=0;
		void		Enable(void);
		void		Disable(void);
		void		Relocate(int xx, int yy) { xr += xx; yr	+= yy; }
		virtual		void ClickUp(int a);  // if TRUE, so call handler
		virtual		void ClickDown(void) {}
		void		SetTrigger(int a) {	if (a) status |= WTRIGGER;	else	status &= (~WTRIGGER);	draw();}
		int	   		GetTrigger(void) const { return status&WTRIGGER?1:0;	}
		static 		Control * ButtonFind(GuiEvent *e,int	n);
		int	   		GetXr(void) const { return xr; }
		int	   		GetYr(void) const { return yr; }
		Window		*GetOwner(void) const { return owner;	}
		void		RefreshGroup(void);
		int			GetLocalId() const { return local_id; }
		void		SetInk(FGPixel i) { ink = i; draw(); }
		void		SetPaper(FGPixel p) { paper = p; draw(); }
		static		void Close(CallBack cb);
		static		void Quit(CallBack cb);

		virtual		void OnActivate(void) { RunSignal(); }
		virtual		void Ioctl(void *, int io) { iomode = io; }
		void		SetActive(void)
		{ 
			call_handler = cControl = this; 
			OnActivate();
		}
};

class Label : public Control {
	protected:
		virtual	void draw(void);
		void ClickUp(int a);
		unsigned short int is_transparent:1, high_light:1, font_size:8;
	public:
		Label(int xs, int ys, char	*nm, int key, Window *w, ControlCall f, int i, int p)
			: Control(xs, ys, nm?strlen(nm)*fontw:16, fonth, nm, LABEL, key, BNORMAL, w,	f, p, i)
		{
			font_size = fontn;
			is_transparent = high_light = 0;
		 	draw();
		}
		void SetTransparent(void) { is_transparent = 1; draw(); }
		void MoveIn(void)
		{
			high_light = 1;
			draw();
		}
		void MoveOut(void)
		{
			high_light = 0;
			draw();
		}
};

class PushButton : public Control {
	protected:
			virtual	void draw(void);
	public:
		PushButton(int xs, int ys, int ws, int hs, char	*nm, int key, int flag,	Window *w, ControlCall f=0)
			: Control(xs, ys, ws, hs, nm, PUSHBUTTON, key, flag, w,	f)
			{
			 	draw();
			}
		PushButton(int xs, int ys, int key,	int	flag, Window *w, Bitmap	*bm, ControlCall f=0)
			: Control(xs, ys,PUSHBUTTON, key, flag,	w, bm, f)
			{
			 	draw();
			}
		void ClickDown(void);
		void ClickUp(int);
		void MoveIn(void);
		void MoveOut(void);
		void Push(void);
		void Release(void);
};

class CheckButton :	public Control {
	protected:
			virtual	void draw(void);
			int		* variable;
	public:
		CheckButton(int m, int	xs,	int	ys,	char *nm, int key, int flag, Window	*w,	int	fg,	int	bg,	ControlCall f=0, int *var=0)
			: Control(xs, ys, strlen(nm)?strlen(nm)*8+24:14, 14, nm, CHECKBUTTON,	key, flag, w, f,bg,fg)
			{
				variable = var;
				mask = m;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		void ChangeItem(int *v) { variable = v; SetTrigger(*v&mask); draw(); }
};

class PointButton :	public Control {
	protected:
			virtual	void draw(void);
			int		* variable;
	public:
		PointButton(int m, int	xs,	int	ys,	char *nm, int key, int flag, Window	*w,	int	fg,	int	bg,	ControlCall f=0, int *var=0)
			: Control(xs, ys, strlen(nm)?strlen(nm)*8+24:16, 14, nm, POINTBUTTON,	key, flag, w, f,bg,fg)
			{
				variable = var;
				mask = m;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		void ChangeItem(int *v) { variable = v; SetTrigger(*v&mask); draw(); }
};

class TwoStateButton : public PointButton
{
	protected:
		virtual void draw(void);
	public:
		TwoStateButton(int xs, int ys, int ws, int hs, char *nm, int key, Window *wnd, int *var=0, ControlCall f=0) :
			PointButton(1, xs,ys,"", key, BNORMAL, wnd, 0, 15, f, var)
		{
			w = ws;
			h = hs;
			SetName(nm); // ma v sebe draw
		}
};

extern EditBox *ininput;

class EditBox :	public Control {
		friend class FGUpDown;
	protected:
		virtual	void draw(void);
		void	draw(int);
		int		w1,w2,min,max,data_type;
		unsigned first:1,passwd:1,check_range:1,hex:1,caps:2;
		double	mind,maxd;
		void  *	ptr;
		static	char	buf[128], clip[128];
		static	int is_clip;
		int		pos, maxpos, iline, icol, offset, size;
		void UpDown(double step)
		{
			if (data_type==0)
			{
				 *(int	*)ptr += int(step);
			}
			else if	(data_type==2)
			{
				 *(double *)ptr += step;
			}
			TestRange();
		}
		void TestRange(void)
		{
			if (check_range==0) return;
			if (data_type==0)
			{
				int	dato = *(int *)ptr;
				if (dato<min) *(int	*)ptr =	min;
				else if	(dato>max) *(int *)ptr = max;
			}
			else if	(data_type==2)
			{
				double dato	= *(double *)ptr;
				if (dato<mind) *(double	*)ptr =	mind;
				else if	(dato>maxd)	*(double *)ptr = maxd;
			}
		}
	public:
		virtual ~EditBox()
		{
			if (ininput	== this)
				ininput	= 0;
		}
		void MoveIn(void);
		void MoveOut(void);
		void ClickUp(int);
		int	inputproc(unsigned);
		void input(void);
		void PasswdMode(int m) { passwd = m; draw(); }
		void HexMode(int m) { hex = m; draw(); }
		void CapsMode(int m);
		void ChangeItem(int	*p)
		{
			assert(data_type==0);
			ptr	= p;
			draw();
		}
		void ChangeItem(char *p)
		{
			assert(data_type==1);
			ptr= p;
			draw();
		}
		void ChangeItem(double *p)
		{
			assert(data_type==2);
			ptr	= p;
			draw();
		}
		void SetSize(int ns)
		{
			pos = offset = 0;
			assert(ns>0 && ns<128);
			size = ns;
			draw();
		}
		void Disable(void);
		EditBox(int xs,	int ys,	int ws1, int ws2, char *nm, int	key, Window *w,	int *pt, int ink, int paper, ControlCall f, int mn, int mx, int check);
		EditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, Window	*w,	char *pt, int ink, int paper, ControlCall f);
		EditBox(int xs,	int ys,	int ws1, int ws2, char *nm, int	key, Window *w,	double *pt, int	ink, int paper,	ControlCall f, double mn, double mx, int check);
};

class ProgressBar : public Control
{
	protected:
		int		steps, value;
		int		sirka;
	public:
		void	draw(void);
		void	setProgress(int a) { if (a>=0 && a<=steps) value = a; draw(); }
		int		progress(void) const { return value; }
		ProgressBar(Window *w, int xx, int yy, int ww, int hh, int s);
};

class BaseMenu : public	Control	{
			virtual	void draw(void);
	public:
		void MoveIn(void);
		void MoveOut(void);
		virtual	void ClickUp(int a);  // if TRUE, so call handler
		BaseMenu(int x,	int	y, int ww, int hh, char	*nm, int key, Window *w, FGPixel bg, FGPixel fg, ControlCall f, Type type)
			: Control(x,y,ww, hh, nm, type, key, BNORMAL, w, f,bg,fg)
			{
				draw();
			}
};

class Slider  : public Control
{
	protected:
		virtual void draw(void) {}
		int		minv,maxv,*val,steps,smer; // 0 je vodorovny, 1 zvysly
		void 	frame(int,int,int,int,int f=0);
		virtual void MoveIn(void);
		virtual void MoveOut(void);
		virtual void ClickDown(void);
	public:
		void	modify(int);
		void    redraw(void)
		{
			ControlCall tmp = fnc;
			fnc = 0;
			draw();
			fnc = tmp;
		}
		Slider(Window *win, int xs, int ys, int ws, int hs, int minimal, int maximal, int s, int *v, ControlCall f) :
			Control(xs, ys, ws, hs, 0, SLIDEBAR, 0, BNORMAL, win, f)
		{
			if (minimal>maximal)
			{
				minv = maximal;
				maxv = minimal;
			}
			else
			{
				minv = minimal;
				maxv = maximal;
			}
			if (*v<minv) *v=minv;
			if (*v>maxv) *v=maxv;
			val  = v;
			steps= s;
			smer = h>w;
		}
};

class SlideBarH : public Slider 
{
// 3 - 16 - val - 3 - 38
	public:
		SlideBarH(int x, int y, int min, int max, int step, int *val, Window *win, ControlCall f=0):
			Slider(win,x,y,(max-min)/step+16+12+38,16,min,max,step,val,f)
		{
			draw();
		}
		virtual void draw(void);
};

class SlideBarV : public Slider
{
// 3 - 16 - val - 3 - 38
	public:
		SlideBarV(int x, int y, int min, int max, int step, int *val, Window *win, ControlCall f=0):
			Slider(win,x,y,16,(max-min)/step+16+12+38,min,max,step,val,f)
		{
			draw();
		}
		virtual void draw(void);
};

class SlideBar : public Slider
{
// 3 - 16 - val - 3
	public:
		SlideBar(int x, int y, int w, int h, int max, int *val, Window *win, ControlCall f=0):
			Slider(win,x,y,w,h,0,max,1,val,f)
		{
			draw();
		}
		virtual void draw(void);
};

class Window : public BaseGui
{
		friend	class Slider;
		friend	class FGListBox;
		friend	volatile void DestroyWindow(void);
		friend	void   GuiEvent::TranslateUserEvent(void);
		friend	int	isgmkey(int	k);
		friend	class Control;
	protected:
		int		icon_x, icon_y;
		int		curr_ctrl;			// default button
		short	c_menupos;
		short	first_draw:1, changed:1, nControls:8; // pocet buttonov priradenych oknu
		short	xpos, ypos;
		Window		**itself;
		void 	(*handler)(GuiEvent	*);

		BaseGui	*Buttony;
		void	*tab_page;	// FGList<FGTabPage>

		static	int	nWnd;
		Window	*iconized;
		DrawBuffer *icon;
		FGPixel	statuscolor;
		const   char * current_tab_page;
		static	int	allControls;
		static	int	rebuild, optimize;
		static	Window *current, *first;  // last in list of all windows
		static	int	line_x1, line_y1, line_x2, line_y2;

		BaseGui	*GetLastButton(void); // last in list
		int		OverWnd(void);
		void	RepaintBlock(int x,	int	y, int w, int h);
		void	init(void);
		void	default_clipping(void);
		long	WindowStatus(long status);
		void	WindowHide(void);
		void	WindowShow(void);
		void	WindowRecopy(void);
		int		OdkryteOkno(Window *pokial, int x, int y, int w, int h);
		int		Rozporcuj(int X1, int Y1, int X2, int Y2, int *in, int *out);
		int		GetXM(int);
		int		GetYM(void);
		void	(*CallWindowHook)(int x, int y, int& nx, int& ny, int);
		void 	WindowUpdate(int dx, int dy, int all=0);
		void	intersect(int, int, int, int);
		static  int	_over(int, int, int, int, int xx, int yy, int ww, int hh);
		virtual int iconize(int xs, int ys);
		virtual void WindowPiktograms(int x, int y);
		virtual void WindowDrawTitle(int color); // old virtual
		static	void PixelW(int x, int y);
		static	void HlineW(int x, int y, int w);
		void	  	Relocate(int xx, int yy);
	public:
		void frame(void);
		static	void HideAll(void);
		static	void ShowAll(void);
		virtual	void draw(void);
		static	void nextwnd(void);
		EditBox *FindNextEditBox(void);
		void	SetDefaultControl(int _id)
		{
			curr_ctrl = _id;
		}
		void	SetChange(void) { changed = 1; }
		void	SetIconPosition(int x, int y) { icon_x = x; icon_y = y; }
		void	ResetChange(void) { changed = 0; }
		int		ShowNotify(void);
		void	InstallWindowHook(void(*hook)(int, int, int&, int&, int))
		{
			CallWindowHook = hook;
		}
		int		IsIconized(void) const { return (int)iconized; }
		static	int	TestNullHandler(void);
		void	wputc(int);
		void	move(int x=0,	int	y=0) { wputc(-1); xpos=x; ypos=y; }
		static	Window * GetCurrent(void)	{ return current; }
		BaseGui	* GetCurrentControls(void) { return	Buttony; }
		void	WindowUpdateBlock(int x, int y, int w, int h);
		void	WindowRepaint(int	xr,	int	yr,	int	w, int h);
		void	WindowRepaintUser(int	xr,	int	yr,	int	w, int h);

		Window(Window **, int xs, int ys, int ws, int hs, char *nm,	GuiHwnd=0, int i=IM, int p=PM, int flag=WTITLED|WFRAMED|WCLICKABLE);
		Window(Window **, FGPixel);	// root window
		virtual ~Window();

		PushButton	*AddPushButton(int xs, int ys, int ws, int hs, char *nm,	int	key=0, ControlCall f=0, int flag=BNORMAL);
		PushButton	*AddPushButton(int xs, int ys, int key, Bitmap *bm, ControlCall f=0, int flag=BNORMAL);
		SlideBarH	*AddSlideBarH(int xs, int ys, int min, int max, int step, int *val, ControlCall f=0);
		SlideBarV	*AddSlideBarV(int xs, int ys, int min, int max, int step, int *val, ControlCall f=0);
		SlideBar	*AddSlideBar(int xs, int ys, int width, int height, int *val, ControlCall f=0);
		CheckButton	*AddCheckButton(int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		PointButton	*AddPointButton(int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		CheckButton	*AddCheckButtonMask(int mask, int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		PointButton	*AddPointButtonMask(int mask, int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		TwoStateButton	*AddTwoStateButton(int	xs,	int	ys,	int w, int h, char *nm, int key=0, int * variable=0, ControlCall=0);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f, int min, int max);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f=0);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox		*AddEditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, double *p, ControlCall f, double min, double max);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, double *p, ControlCall f=0);
		BaseMenu	*AddBaseMenu(char *nm,	int	key=0, ControlCall f=0);
		ProgressBar	*AddProgressBar(int xs, int ys, int ws, int hs, int size);
		FGListBox 	*AddListBox(int xs, int ys, int w, int h, int dropdown, FGListBoxCallBack drawone);

		// absolete - for compatibility only!!
		FGListBox  	*AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int *cnt, int *pos, FGListBoxCallBack drawone, void *data=0);
		FGListBox 	*AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int cnt, FGListBoxCallBack drawone, void *data);
		Label		*AddLabel(int x, int y, char *s, int key=0, ControlCall f=0, int i=-1, int p=-1);
		void 		WindowMove(int,int);
		void 		WindowFocus(void);
		void 		WindowText(int	x, int y, char *s);
		void 		WindowText(int	x, int y, char *s, int color);
		void 		WindowText(int	x, int y, char *s, int color, int bk);
		void 		WindowText(int x, int y, char *s, FontProperty * f, int color, int bk);
		void 		WindowTextUTF8(int x, int y, char *s, FontProperty * f, int color, int bk);
		void 		WindowTextUnicode(int x, int y, const unsigned short *s, FontProperty * f, int color, int bk);
		void 		WindowPanel(int, int, int, int, char *s=0, int ink=CWHITED, int ink2=CDARK);
		void 		WindowBox(int x, int y, int a,	int	b);
		void 		WindowBox(int x, int y, int a,	int	b, int c);
		void 		WindowPixel(int x,	int	y, int c);
		void 		WindowPixel(int x,	int	y);
		void 		WindowRect(int	x, int y, int a, int b);
		void 		WindowRect(int	x, int y, int a, int b,	int	c);
		void 		WindowPatternRect(int	x, int y, int a, int b, Pattern *);
		void 		WindowPatternRect(int	x, int y, int a, int b,	int	c, Pattern *);
		void 		WindowLine(int	x, int y, int a, int b);
		void 		WindowLine(int	x, int y, int a, int b,	int	c);
		void 		WindowPatternLine(int	x, int y, int a, int b, Pattern *p=&PatternDot);
		void 		WindowPatternLine(int	x, int y, int a, int b,	int	c, Pattern *p=&PatternDot);
		void 		WindowDrawCircle(int x, int y,	int	r, int c);
		void 		WindowDrawCircle(int x, int y,	int	r);
		void 		WindowFillCircle(int x, int y,	int	r, int c);
		void 		WindowFillCircle(int x, int y,	int	r);
		void 		WindowDrawEllipse(int x, int y,	int	rx, int ry, int c);
		void 		WindowDrawEllipse(int x, int y,	int	rx, int ry);
		void 		WindowFillEllipse(int x, int y,	int	rx, int ry, int c);
		void 		WindowFillEllipse(int x, int y,	int	rx, int ry);
		void 		WindowDrawArc(int x, int y,	double ang1, double ang2, int r, int c);
		void 		WindowDrawArc(int x, int y,	double ang1, double ang2, int r);
		void 		WindowScrollDown(int p1, int p2, int p3, int p4 ,int p5);
		void 		WindowScrollUp(int	p1,	int	p2,	int	p3,	int	p4 ,int	p5);
		void		WindowScrollLeft( int p1 , int p2 , int p3, int p4, int p5);
		void		WindowScrollRight( int p1 , int p2 , int p3, int p4, int p5);
		void 		WindowStatusBar(int, char *s, int c=CBLACK);
		int	 		TitleFind(GuiEvent	*e);
		static 		Window *	WindowFind(GuiEvent	*e);
		int	 		printf(const char * , ... );
		int	 		printf(int,int,const char *, ... );
		void 		SendToWindow(GuiEvent *p);
		void 		WindowLock(void) {	status|=WLOCKED; }
		void 		WindowUnLock(void);
		void 		WindowResize(int, int);
		void 		WindowShape(int, int, int, int);
		static 		void WindowVirtualRoot(int, int, int, int);
		void 		puts(char *s) { while(*s) wputc(*s++);	}
		void 		WindowCopyFrom(int x,	int	y, int xs, int ys,	int	w, int h, Window *p);
		void 		WindowPutBitmap(int x,	int	y, int xs, int ys,	int	w, int h, DrawBuffer *p);
		void 		WindowGetBitmap(int x,	int	y,	int	xs,	int	ys,	int	w, int h, DrawBuffer *p);
		void 		SetName(char *s)
		{
			BaseGui::SetName(s);
			if (this==current) WindowStatus(WACTIVE);
			else WindowStatus(WDEACTIVE);
		}
		int 		WindowIconize(void);
		int 		WindowIconize(int, int);
		int 		ShowModal(void);
		void 		WindowAttachIcon(DrawBuffer *ico) { icon = ico; }
		FGRect 		WindowFillPolygon(int,	int[][2], int col=-1);
		FGRect 		WindowFillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, int col=-1);
		void 		WindowDrawPolygon(int,	int[][2], int col=-1);
		void 		WindowSpline(int points[8]);
		void 		WindowSetWorkRect(int,int,int,int);
		void 		RemoveControls(void);
		void		DisableControls(void);
		void		EnableControls(void);
		void		SetInk(FGPixel i) { ink = i; }
		void		SetPaper(FGPixel p) { paper = p; }

		void		AddTabPage(const char *name);
		FGTabPage * DeleteTabPage(const char *name);
		void		DeleteTabPages(void);
		void		SetNextTabPage(void);
		const char *	GetCurrentTabPage(void)
		{
			return current_tab_page;
		}
		FGTabPage * SetTabPage(const char *name);
		void		DrawTabPages(void);
		int			ClickTabPage(int, int);
		void		RelocateTabPage(int xx, int yy);

		virtual void OnKeyPress(int) { }
		virtual void OnMouseMove(int , int) { }
		virtual void OnClick(int, int) { }
		virtual void OnContextPopup(int, int) { }
		virtual void OnPaint(void) { }
		virtual void OnStartDrag(int , int, int) {}
		virtual void OnEndDrag(int ,int, int, int, int) {}
		virtual void OnMove(int , int) { }
		virtual void OnResize(int , int) { }
		virtual void OnIconize(void) { }
		virtual void OnAccelerator(Control *) { }
//		virtual void OnCloseQuery(int&) { }
};

class App
{
		enum { MAX_EVENT_QUEQUE=32	}; // numbers of events to app per one user event
	protected:
		static	GuiEvent queque[MAX_EVENT_QUEQUE];
		static	void (*DelayProc)(void);
		static	int	quequeIndex;
		static	void (*OnEverySecond)(int);
		static  Window *Root;
		static	MainHwnd appHandler;
		static  unsigned long ticks;

	public:
		static	int		flags;
		static	int		video;
		static	int		Argc;
		static	int		background;
		static	char	**Argv;
		static	char	* name;
		static	char	*homedir;
		static  char	currdir[128];
		static	int 	ttf_support;

		static void	GetDragVector(int &a, int &b, int &c, int &d);
		static void	SetDragShape(void(*a)(int,int,int,int)=0);
		static	void SetDelayProc(void (*fnc)(void))
		{
			DelayProc =	fnc;
		}
		static	void CallDelayProc(void)
		{
			if (DelayProc) DelayProc();
			else delay(10);
		}

		App(int, int &,	char **	&, int b, int f=APP_ENABLEALTX);
		virtual ~App();
		
		static	void SendToApp(GuiEvent	*x);
		void	Run(MainHwnd hwnd=0);
		void	FGYield(void);
		static	void AppDone(void)
		{
			GuiEvent e(QUITEVENT);
			SendToApp(&e);
		}
		static	void Timer(void);
		// NOTE! resolution is cca. 20ms
		static void	SetTimerProc(void ((*p)(int)), int t=1000)
		{
			OnEverySecond =	p;
			ticks = t;
		}
		static Window * GetRootWindow(void)
		{
			return Root;
		}
		void BroadcastMessage(GuiEvent *event);
		int RunModal(Window *which);

		// Object-oriented API via overloaded methods
		virtual void OnKeyPress(int) { }
		virtual void OnMouseMove(int , int) { }
		virtual void OnClick(int, int) { }
		virtual void OnContextPopup(int, int) { }
		virtual void OnStartDrag(int , int, int) {}
		virtual void OnEndDrag(int ,int, int, int, int) {}
		virtual void OnCursorOut(int)
		{
		}
		FGMouseCursor * CursorLoad(FGMouseCursor *cur);
};

extern App *cApp; // current app
extern Config *cCfg;

class MenuWindow : public Window
{
	protected:
		friend class Window;
		static GuiHwnd Proc;
		static MenuWindow *	CurrentMenu;
		static void	MenuWindowHandler(GuiEvent *p);
		int		offset;
		void _initmw(GuiHwnd proc)
		{
			offset = 0;
			Proc = proc;
			CurrentMenu	= this;
		}
		virtual ~MenuWindow()
		{
			CurrentMenu	= 0;
		}
	public:
		MenuWindow(int x, int y, int w,	int	h, char	*name, GuiHwnd proc=0):
			Window(0,x,y,w,h,name,MenuWindowHandler,CBLACK,CGRAY2,WFRAMED|WCLICKABLE|WNOPICTO|WMENUWINDOWTYPE)
		{
			paper = CScheme->menuwindow_back;
			ink   = CScheme->menuwindow_fore;
			draw();
			_initmw(proc);
		}
		MenuWindow(int w, int h, char	*name, GuiHwnd proc=0):
			Window(0,GetMouseX(),GetMouseY(),w,h,name,MenuWindowHandler,CBLACK,CGRAY2,WFRAMED|WCLICKABLE|WNOPICTO|WMENUWINDOWTYPE)
		{
			paper = CScheme->menuwindow_back;
			ink   = CScheme->menuwindow_fore;
			draw();
			_initmw(proc);
		}
		int	GetXM(void)	const
		{
			return 4;
		}
		int	GetYM(int a)
		{
			return (offset+=a,offset-a);
		}
		BaseMenu	*AddMenu(char *nm,	int	key=0, ControlCall f=0);
		Label		*AddLabel(char *s, int key=0, ControlCall f=0, int i=-1, int p=-1);
		CheckButton	*AddCheckButton(char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		PointButton	*AddPointButton(char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		CheckButton	*AddCheckButtonMask(int, char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		PointButton	*AddPointButtonMask(int, char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p,ControlCall f, int min, int max);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f=0);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox	*AddEditBox(int sz, int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,	ControlCall f, double min, double max);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,	ControlCall f=0);
		void Separator(void);
};

void SendEvent(Window *w, int event, int key=0,	int	x=0, int y=0);
void SendEvent(Window *w, int event, int x, int y, int, int);
unsigned int CalculateCRC(unsigned StartCRC, void *Addr, unsigned Size);
void SetDebug(int l=1);

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#endif
