CFLAGS  = -g -c -O2 -fno-exceptions -march=pentium  # -mcpu=i586
CC	= gcc
HDRS    = base.h _fastgl.h config.h

LIBS	= -ldl -lm -lstdc++ # -liberty
INCPATH	= -I/usr/X11R6/include -I/usr/include/freetype2
LIBPATH	= -L/usr/X11R6/lib

OBJS	=   engine.o fastgl.o base.o resource.o widgets.o user.o linkeyb.o dll.o linux.o editor.o bitmap.o timer.o drivers.o detect.o fgstring.o listbox.o
SRCS	=   fastgl.cpp base.cpp resource.cpp widgets.cpp user.cpp linkeyb.cpp linux.cpp dll.cpp editor.cpp engine.asm bitmap.cpp timer.cpp drivers.cpp detect.c fgstring.cpp listbox.cpp
BASE	=   /usr
INSTDIR =   $(BASE)/include/fastgl


ifdef X11SUPPORT
OPTIONS0	= -DX11_DRIVER
endif

ifdef DGASUPPORT
OPTIONS1	= -DDGA_DRIVER
endif

ifdef PNGSUPPORT
OPTIONS2	= -DFG_PNG
endif

ifdef JPGSUPPORT
OPTIONS3	= -DFG_JPEG
endif

ifdef TTFSUPPORT
OPTIONS4	= -DFG_TTF
OBJS2		= fgttf.o
SRCS2		= fgttf.cpp
endif

CFLAG	= $(CFLAGS) $(COLORDEPTH) $(OPTIONS0) $(OPTIONS1) $(OPTIONS2) $(OPTIONS3) $(OPTIONS4) $(INCPATH)

default: $(STARGET) $(DTARGET).3 file2asm
	-rm -Rf /usr/local/include/fastgl
	-rm -f /usr/local/lib/$(DTARGET).3
	-rm -f /usr/local/lib/$(STARGET)
	install -d -g root -o root -m 755 $(INSTDIR)
	install -d -g root -o root -m 755 $(BASE)/lib
	install -d -g root -o root -m 755 $(BASE)/bin
	install -g root -o root -m 0755 file2asm $(BASE)/bin
	install -g root -o root -m 0644 fastgl.h $(INSTDIR)
	install -g root -o root -m 0644 base.h $(INSTDIR)
	install -g root -o root -m 0644 widgets.h $(INSTDIR)
	install -g root -o root -m 0644 fglist.h $(INSTDIR)
	install -g root -o root -m 0644 listbox.h $(INSTDIR)
	install -g root -o root -m 0644 fgstring.h $(INSTDIR)
	install -g root -o root -m 0644 fglmesa.h $(INSTDIR)
	install -g root -o root -m 0644 callback.h $(INSTDIR)
	install -g root -o root -m 0644 $(STARGET) $(BASE)/lib
	install -g root -o root -m 0755 $(DTARGET).3 /usr/lib
	ldconfig

clean:
	rm -f *.a	
	rm -f *.o
	rm -f *.obj
	rm -f *.lib
	rm -f *.err
	rm -f *.so.?
	rm -f file2asm core *.exe

$(STARGET): $(OBJS) $(OBJS2)
	ar rcs $(STARGET) $(OBJS) $(OBJS2) 

$(DTARGET).3: $(OBJS) $(OBJS2)
	$(CC) -g -fnostdinc -shared -Wl,-soname,$(DTARGET) -o $(DTARGET).3 $(OBJS) $(OBJS2) $(LIBS) $(LIBPATH) $(X11SUPPORT) $(DGASUPPORT) $(JPGSUPPORT) $(PNGSUPPORT) $(TTFSUPPORT)

file2asm: file2asm.cpp
	$(CC) -s -O -o file2asm file2asm.cpp
	cp file2asm $(BASE)/bin/file2asm
		
fastgl.o:  $(HDRS)     fastgl.cpp fastgl.h
	$(CC)  $(CFLAG)  fastgl.cpp 

bitmap.o:  $(HDRS)     bitmap.cpp fastgl.h
	$(CC)  $(CFLAG)  bitmap.cpp

widgets.o:  $(HDRS)     widgets.cpp widgets.h fastgl.h
	$(CC)  $(CFLAG)  widgets.cpp

editor.o:  $(HDRS)     editor.cpp widgets.h fastgl.h
	$(CC)  $(CFLAG)  editor.cpp

linux.o:  $(HDRS)     linux.cpp linux.h
	$(CC)  $(CFLAG)  linux.cpp

drivers.o:  $(HDRS)     drivers.cpp linux.h tdfx.h
	$(CC)  $(CFLAG)  drivers.cpp

detect.o:  $(HDRS)     detect.c 
	$(CC)  $(CFLAG)  detect.c

base.o:  $(HDRS)   base.cpp
	$(CC)  $(CFLAG)  base.cpp 

dll.o:  $(HDRS)   dll.cpp dll.h
	$(CC)  $(CFLAG)  dll.cpp 

user.o:  $(HDRS)   user.cpp fastgl.h
	$(CC)  $(CFLAG)  user.cpp

timer.o:  $(HDRS)   timer.cpp fastgl.h
	$(CC)  $(CFLAG)  timer.cpp

resource.o:  $(HDRS)   resource.cpp
	$(CC)  $(CFLAG)  resource.cpp

linkeyb.o:  linkeyb.h  linkeyb.cpp
	$(CC)  $(CFLAG)  linkeyb.cpp

fgstring.o:  fgstring.h fgstring.cpp
	$(CC)  $(CFLAG)  fgstring.cpp

listbox.o:  listbox.h  listbox.cpp
	$(CC)  $(CFLAG)  listbox.cpp

fgttf.o:  fgttf.h  fgttf.cpp
	$(CC)  $(CFLAG)  fgttf.cpp

engine.o:engine.asm
	nasm -f elf -dLINUX engine.asm

