#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"

#include "drivers.h"
#include "msdos.h"

#ifdef __WATCOMC__
int	__dpmi_free_dos_memory(int sel)							/* DPMI 0.9 AX=0101 */
{
    union REGS  r;

    r.w.ax = 0x101;                 /* DPMI free DOS memory             */
    r.w.dx = sel;                   /* DX := selector from 0x100        */
    int386(0x31, &r, &r);
	if (r.w.cflag	& 1)
		return -1;				// error
	return 0;
}

/****************************************************************************
*
* Function:     DPMI_allocRealSeg
* Parameters:   size    - Size of memory block to allocate
*               sel     - Place to return protected mode selector
*               r_seg   - Place to return real mode segment
*
* Description:  Allocates a block of real mode memory using DPMI services.
*               This routine returns both a protected mode selector and
*               real mode segment for accessing the memory block.
*
****************************************************************************/
int	__dpmi_allocate_dos_memory(int _paragraphs,	int	*_ret_selector_or_max)
{
	union REGS reg;

	reg.w.ax = 0x100;
	reg.w.bx = _paragraphs;
	int386(0x31, &reg, &reg);
	if (reg.w.cflag	& 1)
		return -1;				// error

	*_ret_selector_or_max =	reg.w.dx;
	return reg.w.ax;
}

int	__dpmi_free_ldt_descriptor(unsigned sel)
{
	union REGS reg;

	reg.w.ax = 0x1;
	reg.w.bx = sel;
	int386(0x31, &reg, &reg);
	if (reg.w.cflag	& 1)
		return -1;				// error
	return 0;
}

/****************************************************************************
*
* Function:     DPMI_allocSelector
* Returns:      Newly allocated protected mode selector
*
* Description:  Allocates a new protected mode selector using DPMI
*               services. This selector has a base address and limit of 0.
*
****************************************************************************/
int	__dpmi_allocate_ldt_descriptors(int _count)						/* DPMI 0.9 AX=0000 */
{
    int         sel;
    union REGS  r;

    r.w.ax = 0;                     /* DPMI allocate selector           */
    r.w.cx = _count;                     /* Allocate a single selector       */
    int386(0x31, &r, &r);
    if (r.x.cflag)
		return -1;
    sel = r.w.ax;

    r.w.ax = 9;                     /* DPMI set access rights           */
    r.w.bx = sel;
    r.w.cx = 0x8092;                /* 32 bit page granular             */
    int386(0x31, &r, &r);
    return sel;
}

/****************************************************************************
*
* Function:     DPMI_mapPhysicalToLinear
* Parameters:   physAddr    - Physical memory address to map
*               limit       - Length-1 of physical memory region to map
* Returns:      Starting linear address for mapped memory
*
* Description:  Maps a section of physical memory into the linear address
*               space of a process using DPMI calls. Note that this linear
*               address cannot be used directly, but must be used as the
*               base address for a selector.
*
****************************************************************************/
int	__dpmi_physical_address_mapping(__dpmi_meminfo *info)					/* DPMI 0.9 AX=0800 */
{
    union REGS  r;

    r.w.ax = 0x800;                 /* DPMI map physical to linear      */
    r.w.bx = info->address >> 16;
    r.w.cx = info->address & 0xFFFF;
    r.w.si = info->size >> 16;
    r.w.di = info->size & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
        return -1;
    info->address = (long)((r.w.bx << 16) + r.w.cx);
    return info->address;
}

int	__dpmi_free_physical_address_mapping(__dpmi_meminfo *info)				/* DPMI 0.9 AX=0801 */
{
    union REGS  r;

    r.w.ax = 0x801;                 /* DPMI unmap physical to linear      */
    r.w.bx = info->address >> 16;
    r.w.cx = info->address & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
        return -1;
	return 0;
}

/****************************************************************************
*
* Function:     DPMI_setSelectorBase
* Parameters:   sel     - Selector to change base address for
*               linAddr - Linear address used for new base address
*
* Description:  Sets the base address for the specified selector.
*
****************************************************************************/
int	__dpmi_set_segment_base_address(int sel, unsigned long linAddr)			/* DPMI 0.9 AX=0007 */
{
    union REGS  r;

    r.w.ax = 7;                     /* DPMI set selector base address   */
    r.w.bx = sel;
    r.w.cx = linAddr >> 16;
    r.w.dx = linAddr & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
		return -1;
	return 0;
}

/****************************************************************************
*
* Function:     DPMI_setSelectorLimit
* Parameters:   sel     - Selector to change limit for
*               limit   - Limit-1 for the selector
*
* Description:  Sets the memory limit for the specified selector.
*
****************************************************************************/
int	__dpmi_set_segment_limit(int sel, unsigned long limit)				/* DPMI 0.9 AX=0008 */
{
    union REGS  r;

    r.w.ax = 8;                     /* DPMI set selector limit          */
    r.w.bx = sel;
    r.w.cx = limit >> 16;
    r.w.dx = limit & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
		return -1;
	return 0;
}

unsigned qnx_segment_overlay_flags(void)
{
	int val;

	val = (int) seg_overlay();
	return val;
}
#endif // watcom c++

int FGVesaDriver::set_page(int page)
{
	if (h*pitch*(page+1) > (int)linear_size)
		return 0;
	dpmir.x.ax = 0x4f07; // set screen start address
	dpmir.x.bx = 0x0080; // set display start during vertical retrace
	dpmir.x.cx = 0;
	dpmir.x.dx = h*page;
	__dpmi_int(16, &dpmir);

	if (dpmir.x.ax != 0x004F)
    	return 0;	// not supported by BIOS
	return 1;
}

int FGVesaDriver::set_mode(int ww, int hh)
{
	char s[128];
   	req_w = ww;
    req_h = hh;
	
	FGDriver::set_mode(w,h);
   	mode_num = find(req_w, req_h);
    if (mode_num==-1)
	{
       	printf("Mode %dx%d not found!\n", ww, hh);
        return 0;
    }
	if (verbose)
		printf("\tmode \"%dx%dx%d\" as VESA:0x%x passed ... ", modelist[mode_num].w, modelist[mode_num].h, bpp*8, modelist[mode_num].mode);
	dpmir.x.cx = modelist[mode_num].mode;
	dpmir.x.ax = 0x4f01;
	dpmir.x.di = 0;
	__dpmi_int(16, &dpmir);
	islinear = (islinear && !!(modelist[mode_num].flag&FG_LINEAR));
	if ((peek(_es_sel, 6) &	0xffff)	< 64)
	{
		printf("Sorry, your card not support 64 KB Window !");
		return 0;
	}
	basesegment = peek(_es_sel, 8)	& 0xffff;
	granularity	= peek(_es_sel,	4) & 255;
	linear_base	= (void *)peek(_es_sel,	0x28);
	videobase = 0;	// always NULL as segment start
#ifndef __QNX__
	if (islinear && linear_base)
	{
		map_linear((unsigned)linear_base, linear_size, &VideoSelector, &VideoLinear);
		if (!(VideoSelector != 0xFFFFFFFF && VideoLinear))
		{
			islinear = 0;
		}
		else
		{
			::VideoSelector = VideoSelector;
		}
	}
#else
	// not supported for WATCOM C++ ;-(
	islinear = 0;
#endif
	int value = modelist[mode_num].mode | (islinear?0x4000:0x0000);
	if (_setvideomode(value) ==	0) return 0; //set properly mode
	if (verbose) printf("O.K.\n");
	if (islinear==0 && bpp>1)
	{
		printf("Linear mode is required for this BPP!");
		return 1;
	}
	sprintf(s, "%s, VESA version %d.%02d", drv_name[driver], vesaversion>>8, vesaversion&255);
	subname = strdup(s);
/*	dpmir.x.dx = value;
	dpmir.x.ax = 0x4f0B; // get pixel clock
	dpmir.x.bx = 0x0001; // get pixel clock
	dpmir.d.ecx= 85;     // Hz
	__dpmi_int(16, &dpmir);
	synch = dpmir.d.ecx;
*/	synch = 0;
	virt_w = req_w;
	virt_h = req_h;
	return postinit();
}

extern "C" void __crt0_load_environment_file(char *app_name);

int	FGVesaDriver::available(void)
{
	// nasty HACK
	if (getenv("DJDIR") == 0)
	{
		printf("\nIf your application crash here, set DJGPP enviroment variable\nto the correct value please!\a\n\n\n");
	}
	detect();
	memset(&dpmir, 0, sizeof(dpmir));
	dpmir.x.ax = 0x4f00;
	dpmir.x.di = 0;
	dpmir.x.es = _es = __dpmi_allocate_dos_memory(512 / 16, &_es_sel);
	poke(_es_sel, 0, '2' * 0xffffff	+ 'E' *	65536 +	'B'	* 256 +	'V');
	__dpmi_int(16, &dpmir);
	if (dpmir.x.ax != 0x004F)
    {
		printf("Sorry, your card not support VESA standard!");
		return 0;
    }
	vesaversion= (peek(_es_sel,	4) & 0xffff);
	linear_size = (peek(_es_sel,	0x12) &	0xffff)	* 0x10000;
	if (peek(_es_sel, 10) & 1) // CAPABLE
	{
		dpmir.x.ax = 0x4f08; // set DAC width to 8 bit
		dpmir.x.bx = 0x0800;
		__dpmi_int(16, &dpmir);
		if (dpmir.x.ax == 0x004F && (dpmir.x.bx&0xff00 && CardInfo.vendor!=0x121a)==0x0800)
		palette_8 = 1;
	}
	return 1;
}

void sched_yield(void)
{
	static int idle_supp=0;
	__dpmi_regs	dpmir;
	dpmir.x.ax = 0x1680;
	if (idle_supp) __dpmi_int(0x2f, &dpmir);
	if (dpmir.x.ax&255) idle_supp = 0;
}

