// ---------------------------------------------------------------------------
//
//		VESA frontend driver
//
// ---------------------------------------------------------------------------

#ifdef __DJGPP__
#include <dpmi.h>
#include <go32.h>
#include <sys/movedata.h>
#include <sys/segments.h>
#endif

#ifdef __WATCOMC__
	typedef struct
	{
		unsigned long handle;	/* 0, 2 */
		unsigned long size;		/* or count *//* 4, 6 */
		unsigned long address;	/* 8, 10 */
	}
	__dpmi_meminfo;

	typedef union
	{
		struct
		{
			unsigned long edi;
			unsigned long esi;
			unsigned long ebp;
			unsigned long res;
			unsigned long ebx;
			unsigned long edx;
			unsigned long ecx;
			unsigned long eax;
		}
		d;
		struct
		{
			unsigned short di, di_hi;
			unsigned short si, si_hi;
			unsigned short bp, bp_hi;
			unsigned short res, res_hi;
			unsigned short bx, bx_hi;
			unsigned short dx, dx_hi;
			unsigned short cx, cx_hi;
			unsigned short ax, ax_hi;
			unsigned short flags;
			unsigned short es;
			unsigned short ds;
			unsigned short fs;
			unsigned short gs;
			unsigned short ip;
			unsigned short cs;
			unsigned short sp;
			unsigned short ss;
		}
		x;
		struct
		{
			unsigned char edi[4];
			unsigned char esi[4];
			unsigned char ebp[4];
			unsigned char res[4];
			unsigned char bl, bh, ebx_b2, ebx_b3;
			unsigned char dl, dh, edx_b2, edx_b3;
			unsigned char cl, ch, ecx_b2, ecx_b3;
			unsigned char al, ah, eax_b2, eax_b3;
		}
		h;
	}
	__dpmi_regs;

extern "C" int	__dpmi_free_dos_memory(int sel);							/* DPMI 0.9 AX=0101 */
extern "C" int	__dpmi_allocate_dos_memory(int _paragraphs,	int	*_ret_selector_or_max);
extern "C" int	__dpmi_free_ldt_descriptor(unsigned sel);
extern "C" int	__dpmi_allocate_ldt_descriptors(int _count);						/* DPMI 0.9 AX=0000 */
extern "C" int	__dpmi_physical_address_mapping(__dpmi_meminfo *info);					/* DPMI 0.9 AX=0800 */
extern "C" int	__dpmi_free_physical_address_mapping(__dpmi_meminfo *);				/* DPMI 0.9 AX=0801 */
extern "C" int	__dpmi_set_segment_base_address(int sel, unsigned long linAddr);			/* DPMI 0.9 AX=0007 */
extern "C" int	__dpmi_set_segment_limit(int sel, unsigned long limit);				/* DPMI 0.9 AX=0008 */
extern "C" unsigned seg_overlay(void);
extern "C" unsigned qnx_segment_overlay_flags(void);

#pragma aux seg_overlay "_*";
#pragma aux qnx_segment_overlay_flags "*_";
#endif // watcom c++

class FGVesaDriver : public FGDriver
{
		// selector for VRAM
		unsigned	VideoSelector;
		// linear address of mapped FRAMEBUFFER (dos only)
		unsigned	VideoLinear;
		__dpmi_regs dpmir;
		int			vesaversion;
		unsigned	_es;
 		int			_es_sel; // segment & selector
		unsigned	basesegment; // z VESA info

		// ------------------------------------------------------------
		/* memory access helper functions */
		int map_linear(unsigned long	adr, unsigned long len, unsigned *sel, unsigned *la	)
		{
		  if(*sel==0xFFFFFFFF)
		  {
			*sel = __dpmi_allocate_ldt_descriptors(1);
			if (*sel==0xFFFFFFFF)	return 0;
		  }
		  if(!*la)
		  {
			__dpmi_meminfo meminfo;
			meminfo.address	= adr;
			meminfo.size	= len;
			if(__dpmi_physical_address_mapping(&meminfo)==-1)
			   return 0;
			if(__dpmi_set_segment_base_address(*sel,meminfo.address)==-1)
			   return 0;
			if(__dpmi_set_segment_limit(*sel, len-1)==-1)
			   return 0;
			*la=meminfo.address;
		  }
		  return 1;
		}

		void	free_linear(unsigned *sel, unsigned	*la)
		{
		  if(*la)
		  {
			  __dpmi_meminfo meminfo;
			  meminfo.address =	*(unsigned *)la;
			  __dpmi_free_physical_address_mapping(&meminfo);
			  *la =	0;
		  }
		  if(*sel != 0xFFFFFFFF)
		  {
			  __dpmi_free_ldt_descriptor(*sel);
			  *sel = 0xFFFFFFFF;
		  }
		}

	protected:
	public:
		virtual int	available(void);
		FGVesaDriver(void) : FGDriver("VESA")
        {
			VideoSelector = 0xFFFFFFFF;
			VideoLinear = 0;
            driver = ::driver;
			_es = _es_sel = 0;
        }
		~FGVesaDriver()
		{
			if (VideoLinear)  free_linear(&VideoSelector, &VideoLinear);
			if (VideoSelector!=0xFFFFFFFF)
				__dpmi_free_ldt_descriptor(VideoSelector);
			VideoSelector = 0xFFFFFFFF;
			if (_es_sel) __dpmi_free_dos_memory(_es_sel);
		}
		virtual int text_mode(void)
		{
			_setvideomode(3);
			return TRUE;
		}
		virtual int	link(void)
		{
        	return 1;
        }
		virtual void get_all_modes(vmode *p)
		{
			for (int m=0x100; m<0x180; m++)
			{
				dpmir.x.cx = m;
				dpmir.x.ax = 0x4f01;
				dpmir.x.di = 0;
				__dpmi_int(16, &dpmir);
				if (dpmir.x.ax != 0x004F)
					continue;				// not valid mode

				if ((peek(_es_sel, 0)&0x19) != 0x19) continue;
				if ((peek(_es_sel, 25) &0xff) != bpp*8) continue;
				p[total_modes].flag = (peek(_es_sel, 0) & 128)?FG_LINEAR:0;
				p[total_modes].refresh = 0;
				p[total_modes].w = (peek(_es_sel, 18) & 0xffff);
				p[total_modes].h = (peek(_es_sel, 20) & 0xffff);
				p[total_modes].mode = m;
				total_modes++;
			}
		}
		virtual int set_mode(int ww, int hh);
		virtual int set_mouse(void)
		{
			__dpmi_regs	dpmir;

			dpmir.x.ax = 0;
			__dpmi_int(0x33, &dpmir);
			if (dpmir.x.ax != 0)
			{
				ismouse	= 1;			/* present, but not initted */
				dpmir.x.ax = 15;		// citlivost
				dpmir.x.cx = 8;
				dpmir.x.dx = 16;
				__dpmi_int(0x33, &dpmir);
				dpmir.x.ax = 4;
				dpmir.x.cx = short(X_width / 2);
				dpmir.x.dx = short(Y_width + 2);
				__dpmi_int(0x33, &dpmir);
				dpmir.x.ax = 7;
				dpmir.x.dx = 0;
				dpmir.x.cx = short(X_width-1);
				__dpmi_int(0x33, &dpmir);
				dpmir.x.ax = 8;
				dpmir.x.dx = 0;
				dpmir.x.cx = short(Y_width-1);
				__dpmi_int(0x33, &dpmir);
				ismouse = 1;
				mouse_string = "DOS_INT33";
			}
			return TRUE;
		}
		virtual void change_mouse(void)
		{
			if (ismouse)
			{
				dpmir.x.ax = 4;
				dpmir.x.cx = short(X_width / 2);
				dpmir.x.dx = short(Y_width + 2);
				__dpmi_int(0x33, &dpmir);
				dpmir.x.ax = 7;
				dpmir.x.dx = 0;
				dpmir.x.cx = short(X_width-1);
				__dpmi_int(0x33, &dpmir);
				dpmir.x.ax = 8;
				dpmir.x.dx = 0;
				dpmir.x.cx = short(Y_width-1);
				__dpmi_int(0x33, &dpmir);
			}
		}
		virtual void reset_mouse(void)
		{
			mouseflag = -1;
			ismouse = 0;
		}
		virtual int set_keyboard(void)
		{
			return TRUE;
		}
		virtual void reset_keyboard(void)
		{
			iskeyboard = 0;
		}
		virtual int get_event(int& type, int& key, int& x, int& y, int& buttons)
		{
			static int xx, yy, bb;
			int k;
			__dpmi_regs	dpmir;
			
			dpmir.h.ah = 0x11;		// sched_yield
			__dpmi_int(0x16, &dpmir);
			if (!(dpmir.x.flags	& 0x40))
			{
				dpmir.h.ah = 0x10;	// test for key buffer

				__dpmi_int(0x16, &dpmir);
				switch (dpmir.h.al)
				{
					case 0xe0:
						k =	dpmir.h.ah + 0x200;
						break;
					case 0x00:
						k =	dpmir.h.ah + 0x100;
						break;
					default:
						k =	dpmir.h.al;
						break;
				}
				type = KEYEVENT;
				key	= k;
				x = y = buttons = 0;
				return TRUE;
			}
			if (ismouse)
			{
				dpmir.x.ax = 3;		// test for key buffer
				__dpmi_int(0x33, &dpmir);
				if (dpmir.x.bx != bb ||	dpmir.x.cx != xx ||	dpmir.x.dx != yy)
				{
					bb = buttons = dpmir.x.bx;
					xx = x = dpmir.x.cx;
					yy = y = dpmir.x.dx;
					type = MOUSEEVENT;
					key	= 0;
					return TRUE;
				}
			}
			return FALSE;
		}
		virtual int set_page(int);
};
