/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 2000  Rosimildo da SIlva

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    rdasilva@connecttel.com
    This driver was based on the SVGA code for Linux.
    rtems_fgdriver.cc - Rtems support routines
*/

#define __USE_GNU
#include "config.h"
#include "base.h"
#include "_fastgl.h"
#include "fastgl.h"

#include <rtems/mw_uid.h>
#include <rtems/kd.h>

#include "drivers.h"
#include "rtems_fgdriver.h"

int kbd_fd=-1;
void (*ptr_vga_setpage)(int);
static const char *Q_NAME       = "MWQ";
#define            Q_MAX_MSGS      128

extern int rtems_get_key( unsigned char scancode );




FGRtemsDriver::~FGRtemsDriver()
{
}

void FGRtemsDriver::mouseopen(void)
{
	if (msefd != -1)
	{
		return;
	}

   int rc;
   /* if this is the first time around, create the message queue */
   rc = uid_open_queue( Q_NAME, O_CREAT | O_RDWR, Q_MAX_MSGS );
   if( rc )
   {
      printf("Could not open UID queue, rc=%d\n", rc );
      exit(-1);
   }

   /* 
    * open the mouse -- if error check the device table to make 
    * sure that device driver is loaded properly.
    */
   msefd = open( "/dev/mouse", O_NONBLOCK );
   if( msefd < 0 )
   {
      printf("Could not initialize mouse device.\n");
      exit(-1);
   }
   uid_register_device( msefd, Q_NAME );
	ismouse = 1;
	mousex = (w-1) / 2;
	mousey = (h-1) / 2;
}

void FGRtemsDriver::mouseclose(void)
{
	if (msefd == -1)
	{
		return;
	}
	close( msefd );
	msefd = -1;
}

int FGRtemsDriver::checkmouse(int& type, int& key, int& x, int& y, int& buttons)
{
	return 0;
}

void FGRtemsDriver::linux_postinit(void)
{
}

int FGRtemsDriver::set_mouse(void)
{
	mouseopen();
	return ismouse;
}

void FGRtemsDriver::reset_mouse(void)
{
   mouseclose();
	ismouse = 0;
}

int FGRtemsDriver::set_keyboard(void)
{
   printf( "RtemsDriver:set_keyboard()\n" );

   /* kbd it is already opened */
   kbd_fd = fileno( stdin );
   uid_register_device( kbd_fd, Q_NAME );
   if( ( kbd_fd=GII_keyboard_init(kbd_fd)) < 0 )
   {
      printf("Could not initialize keyboard\n");
		ioctl(kbd_fd, KDSKBMODE, K_XLATE);
		exit(-1);
   }
	iskeyboard = 1;
   printf( "RtemsDriver:set_keyboard() completed.\n" );
	return TRUE;
}

void FGRtemsDriver::reset_keyboard(void)
{
   printf( "RtemsDriver:reset_keyboard()\n" );
	if (iskeyboard) GII_close(); // close keyboard
	ioctl(kbd_fd, KDSKBMODE, K_XLATE);
	iskeyboard = 0;
   printf( "RtemsDriver:reset_keyboard() completed.\n" );
}

#define  MIN( n1, n2 )     ((n1) > (n2)) ? ( n2 ) : ( n1 )
int FGRtemsDriver::get_event(int& type, int& key, int& x, int& y, int& buttons)
{
  struct MW_UID_MESSAGE m;
  int rc;
  /* let's make sure that the type is invalid */
  m.type = MV_UID_INVALID;

  key = 0;
  /* get a message from the queue, wait up to one miliseconds */
  rc = uid_read_message( &m, 0 );

  /* return if timed-out or something went wrong */
  if( rc < 0 )
  {
     if( errno != ETIMEDOUT )
        printf( " rc= %d, errno=%d\n", rc, errno );
     return 0;
  }
  /* let's pass the event up to microwindows */
  switch( m.type )
  {
    /* Mouse or Touch Screen event */
    case MV_UID_REL_POS:
    case MV_UID_ABS_POS:
    {
       int b = m.m.pos.btns;
       x = MIN( X_max, mousex + m.m.pos.x );
       y = MIN( Y_max, mousey + m.m.pos.y );
       buttons = 0;
       key  = 0;
       type = MOUSEEVENT;
       if( b & MV_BUTTON_RIGHT ) buttons |= 2;
       if( b & MV_BUTTON_LEFT ) buttons |= 1;
       mousex = x;
       mousey = y;
       but = buttons;

       return 1;
     }
       break;

    /* KBD event */
    case MV_UID_KBD:
        key = rtems_get_key( (unsigned char )m.m.kbd.code );
        if( key )
        {
           type = KEYEVENT;
           x = y = buttons = 0;
           return 1;
        }
        break;

    /* micro-windows does nothing with those.. */
    case MV_UID_TIMER:
    case MV_UID_INVALID:
    default:
       ;
  }
  return 0;
}


void FGSVGADriver::get_all_modes(vmode *p)
{
	vga_modeinfo *modeinfo;

	for (int m=1; m<=ptr_vga_lastmodenumber(); m++)
	{
		if (ptr_vga_hasmode(m)==0) continue;
		modeinfo = ptr_vga_getmodeinfo(m);
		if (modeinfo==0) continue;
		if (modeinfo->bytesperpixel != bpp) continue; // not at the end, not good mode
#if FASTGL_BPP==15
		if (modeinfo->colors != 32768) continue; // not at the end, not good mode
#endif
		p->flag = (modeinfo->flags & CAPABLE_LINEAR)?FG_LINEAR:0;
		p->w = modeinfo->width;
		p->h = modeinfo->height;
		p->mode = m;
		total_modes++;
		p++;
	}
}

int FGSVGADriver::link(void)
{
   printf( "RtemsDriver:link()\n" );
	detect();
   printf( "RtemsDriver:link() after detect()\n" );
	subname = (char *)drv_name[driver];

   // RTEMS is statically linked...
	ptr_vga_init = vga_init;
	ptr_vga_setmode = vga_setmode;
	ptr_vga_hasmode = vga_hasmode;
	ptr_vga_setpage = vga_setpage;

   ::ptr_vga_setpage = vga_setpage;

	ptr_vga_getmodeinfo = vga_getmodeinfo;
	ptr_vga_getgraphmem = vga_getgraphmem;
	ptr_vga_setlinearaddressing = vga_setlinearaddressing;
	ptr_vga_getmousetype = vga_getmousetype;
	ptr_waitretrace = vga_waitretrace;
	ptr_vga_lastmodenumber = vga_lastmodenumber;

   printf( "RtemsDriver:link() vga_init() called.\n" );
   ptr_vga_init();

	linear_base = (void *)__svgalib_linear_mem_base;
	linear_size = __svgalib_linear_mem_size;
 	mmio_base = (void *)__svgalib_mmio_base;
	mmio_size = __svgalib_mmio_size;

   printf( "RtemsDriver: linear: ( base=%X, size=%X ), mmio: ( base=%X, size=%X )\n", 
           	(int)linear_base, (int)linear_size, (int)mmio_base, (int)mmio_size );

	linux_postinit();
   printf( "RtemsDriver:link() completed\n" );
	return TRUE;
}

int	FGSVGADriver::set_mode(int ww, int hh)
{
  	req_w = ww;
   req_h = hh;
	FGDriver::set_mode( w, h );
  	mode_num = find( req_w, req_h );
   if( mode_num==-1 )
   {
     	printf("Mode %dx%d not found!\n", ww, hh);
      return 0;
   }
	int rc = ptr_vga_hasmode( modelist[ mode_num ].mode );
  	if (rc)
	{
		if (verbose) printf("\tmode \"%dx%dx%d\" as %d passed ... ", modelist[mode_num].w, modelist[mode_num].h, bpp*8, modelist[mode_num].mode);
    	rc = ptr_vga_setmode(modelist[mode_num].mode);
         if (rc==-1) return 0;
		if (verbose) printf("O.K.\n");
		int sz=0;

      /* let's make sure that the card supports the linear mode */
    	if( ( modelist[ mode_num ].flag & FG_LINEAR ) == 0 )
      {
			islinear=0;
			linear_flag=0;
      }
		if (islinear) sz = ptr_vga_setlinearaddressing();
		if (sz>0)
		{
			islinear=1;
		}
		set_bank_proc = &set_vga_banking;
		videobase = (FGPixel *)ptr_vga_getgraphmem();
		visible = 1;
		synch = 0;
		virt_w = req_w;
		virt_h = req_h;
		return postinit();
	}
	else printf("\tmode \"%dx%dx%d\" as %d not available!\n", modelist[mode_num].w, modelist[mode_num].h, bpp*8, modelist[mode_num].mode);
      return 0;
}


void Snd(int a, int b)
{
	if (kbd_fd>=0) ioctl(kbd_fd, KDMKTONE, (b<<16)+(a&0xffff));
}

void delay(unsigned int ms)
{
  FGTimer::Delay(ms);
}


