/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 2000  Rosimildo da SIlva

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    rdasilva@connecttel.com
    This driver was based on the SVGA code for Linux.
    rtems_fgdriver.cc - Rtems support routines
*/


#ifndef __RTEMS_FGDRIVER_H_
#define __LINUX_FGDRIVER_H_


// ---------------------------------------------------------------------------
//
//		Rtems base driver
//
// ---------------------------------------------------------------------------

extern "C" 
{
#include <vga.h>
extern unsigned long int __svgalib_banked_mem_base, __svgalib_banked_mem_size;
extern unsigned long int __svgalib_mmio_base, __svgalib_mmio_size;
extern unsigned long int __svgalib_linear_mem_base, __svgalib_linear_mem_size;
};

class FGRtemsDriver : public FGDriver
{
	protected:
		int		tty_fd, msefd;
		int		tty_no;
		int		visible;		/* Is VT visible? */
		unsigned	vgamouse:1, gpmmouse:1, 
				svgalib:1, switching:1, 
				dgalib:1, dgamouse:1;
		int		n_arguments;
		pid_t	old_pid, new_pid;
		int		devtype;
		int		oldx, oldy;
		int		mousex, mousey, but;
		FGPixel *sbak;		/* addr of backing store */
		int		uid, gid;

		// jump table to svgalib
		int  (*ptr_vga_init)(void);
		int  (*ptr_vga_setmode)(int);
		int  (*ptr_vga_hasmode)(int);
		vga_modeinfo * (*ptr_vga_getmodeinfo)(int);
		unsigned char * (*ptr_vga_getgraphmem)(void);
		int  (*ptr_vga_getmousetype)(void);
		int  (*ptr_vga_setlinearaddressing)(void);
		void (*ptr_waitretrace)(void);
		int  (*ptr_vga_lastmodenumber)(void);
		void mouseopen(void);
		void mouseclose(void);
		int  checkmouse(int& type, int& key, int& x, int& y, int& buttons);
	public:
		// return TRUE if user==root && console is virtual && svgalib.so is available
		virtual int	available(void)
		{
			return TRUE;
		}
		void linux_postinit(void);
		FGRtemsDriver(char *n) : FGDriver( n )
		{
			n_arguments = 1;
			tty_fd = msefd = -1;
			devtype = 0;
			but = 0;
			svgalib = 0;
			switching = 0;
			tty_no = -1;
			sbak = 0;
		}
		~FGRtemsDriver();
		virtual void change_mouse(void) { }

		virtual int set_mouse(void);
		virtual void reset_mouse(void);
		virtual int set_keyboard(void);
		virtual void reset_keyboard(void);
		virtual int get_event(int& type, int& key, int& x, int& y, int& buttons);
};

// ---------------------------------------------------------------------------
//
//		SVGAlib frontend driver
//
// ---------------------------------------------------------------------------

extern "C" unsigned *set_bank_proc, set_vga_banking;

class FGSVGADriver : public FGRtemsDriver
{
		typedef void (*vv)(void);
		typedef int (*iv)(void);
		typedef void (*vi)(int);
		typedef void (*vii)(int,int);
		typedef int (*ii)(int);

	public:
		FGSVGADriver() : FGRtemsDriver( "SVGAlib" )
		{
			svgalib = 1;
		}
		~FGSVGADriver() { }
		virtual int	link(void);
		virtual void get_all_modes(vmode *p);
		virtual int set_mode(int ww, int hh);
		virtual int text_mode(void)
		{
			isgraph = 0;
			if (ptr_vga_setmode) ptr_vga_setmode( 0 );
			return 1;
      }
	virtual int EnableBuffering(void)
	{
		buffering = 0;
		if (linear_size<w*h*bpp*2U)
			return buffering;
		buffer[0] = videobase;
		buffer[1] = videobase+(w*h*bpp);
		buffering = 1;	// double-buffering
		bufnum = 2;
		current_buffer = 0;
		if (linear_size>=w*h*bpp*3U)
		{
			bufnum = 3; // triple-buffering
			buffer[1] = videobase+(w*h*bpp*2);
		}
		return buffering; // default is disabled
	}
		virtual int set_page(int page)
		{
			if ((page+1)*w*pitch > linear_size) return 0;
			vga_setdisplaystart(h*pitch*page);
			vga_waitretrace();
			return 1;
		}
};

#endif /* __RTEMS_FGDRIVER_H_  */
