/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2003  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    nezmar@atlas.sk
    Get & translate user events

*/

/* TODO :
	Zakladna schema je vygenerovat nejaku udalost, tu posla oknu,
	pripadne i aplikacii via handler ak nejaky existuje.
	Nova schema by mala dodrzat staru a pridat pretazenie
	urcitych definovanych sprav. Tieto by mali rozne
	parametre. Pretazenie by zaroven odstavilo posielanie
	konkretnej spravy via handler. Podobne by sa mal spravat
	i aplikacny handler.
*/

#include <signal.h>
#ifdef __QNX__
//#include <sys/kernel.h>	
//#include <sys/name.h>
#include <sys/sched.h>
#endif

#if defined( __linux__) || defined( __rtems__) || defined(__sun__)
#include <sched.h>
#ifdef __rtems__
#include <rtems/kd.h>
#else
#include <sys/kd.h>
#endif
#include <sys/ioctl.h>
#endif
	
#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"
#include "drivers.h"

int mouseflag = -1;
static int mx, my, mb, oldx, oldy, oldb, drag, action;
int	fulldrag = 1;
static int startXdrag =	0;
static int startYdrag =	0;
static int endXdrag	= 0;
static int endYdrag	= 0;

#ifndef	__QNX__
#define		REPEAT_DELAY1			22
#define		REPEAT_DELAY2			4
#else
#define		REPEAT_DELAY1			5000
#define		REPEAT_DELAY2			3200
#endif

#define		DRAG_TRESHOLD		3

static int repeat_status=0, repeat_delay1=REPEAT_DELAY1, repeat_delay2=REPEAT_DELAY2;
static GuiEvent	 hold_event(BUTTONHOLDEVENT);
static void (*DragShape)(int , int ,int	, int );

EditBox	*ininput = 0;
Control	*idm = 0, *oldidm =	0;
Window	*idw = 0, *oldidw =	0;
long IntCounter;
static int ctrl_break=0;
char *GuiEvent::eventName[]= {
	  "NOEVENT", "KEYEVENT", "MOUSEEVENT", "QUITEVENT", "ACCELEVENT",
	  "TERMINATEEVENT", "MOVEEVENT", "CLICKRIGHTEVENT",
	  "CLICKLEFTEVENT",	"DRAGLEFTEVENT", "DRAGRIGHTEVENT",
	  "GETFOCUSEVENT", "LOSTFOCUSEVENT", "INITEVENT",
	  "BUTTONHOLDEVENT", "WINDOWRESIZEEVENT","WINDOWMOVEEVENT",
	  "STARTDRAGLEFTEVENT", "STARTDRAGRIGHTEVENT", "REPAINTEVENT",
	  "NOTIFYEVENT", "ICONIZEEVENT", "CURSOROUTEVENT", "DBLCLICKRIGHTEVENT",
	  "DBLCLICKLEFTEVENT", "CURSORCHANGEEVENT",
  // must by at the end
	  "NONSENSE"};

static FGPixel __p4[8] = {0, 0, 0, 0, 1, 1, 1, 1};
Pattern __patt(__p4, 8, 1);

//
// Save Screen to BMP file
//

void SaveScreen(void)
{
	Bitmap *a =	new Bitmap(X_width, Y_width);
	char s[20];
	time_t num = time(0);
	get_block(0,0,X_width,Y_width,a->GetArray());
	sprintf(s,"%x",(unsigned int)num);
	if (verbose) printf("Bitmap '%s' saved\n", s);
	a->BitmapSave(s);
	delete a;
}

void SetRepeatDelay(int	c1,	int	c2)
{
	if (c1>0 &&	c1<2000) 
		repeat_delay1 = c1;
	if (c2>0 &&	c2<500)	 
		repeat_delay2 = c2;
}

void SetWindowMoveStyle(int a)
{
	fulldrag = a;
}

int	GetMouseX(void)
{
	return mx;
}

//
// Get real mouse Y-position
//

int	GetMouseY(void)
{
	return my;
}

void GuiEvent::InitUserEvent(void)
{
}

void GuiEvent::CloseUserEvent(void)
{
#if !defined(__MSDOS__) && !defined(__rtems__) && !defined(_WIN32)
	sync();
#endif
	mouseflag = -1;
}

//
// return 0 if none, or keycode
//
int get_key(void)
{
	GuiEvent e;
	mouseflag = 0;
	e.GetUserEvent(1);
	if (e.GetType()==KEYEVENT)
		return e.GetKey();
	else return 0;
}

// start autorepeat routine
void GuiEvent::AutoRepeatStart(Control *id)
{
	hold_event.key=id->GetId();
	hold_event.accel=id;
	hold_event.guiType=id->GetType(); // od buttonu typ ...
	repeat_status =	1;
	AutoRepeatDo();
	repeat_status =	REPEAT_DELAY1;
}

// end autorepeat routine
void GuiEvent::AutoRepeatEnd(void)
{
 	repeat_status =	0;
}

// process -||-
void GuiEvent::AutoRepeatDo(void)
{
 	if (--repeat_status	<= 0)
	{
		Window::GetCurrent()->SendToWindow(&hold_event);
		repeat_status =	REPEAT_DELAY2;
	}
}

void App::GetDragVector(int &a, int &b, int &c, int &d)
{
	a =	startXdrag;
	b =	startYdrag;
	c =	endXdrag;
	d =	endYdrag;
}

void App::SetDragShape(void (*fnc)(int, int, int, int))
{
	DragShape = fnc;
}

void CloseUserInput(void)
{
	GuiEvent::CloseUserEvent();
}

#if defined(__MSDOS__) || defined(_WIN32)
void sched_yield(void);
#endif

GuiEvent *GuiEvent::GetUserEvent(int waits)
{
	static int xx =	X_width	/ 2, yy	= Y_width / 2;
	int	xx2, yy2, rc;
	GuiEvent *val =	0;
	type = 0;
	xx2	= xx;
	yy2	= yy;
	
	if (mouseflag == 1)
		vector_mouse_cursor(xx, yy, TRUE);
	for (; waits; waits--)
	{
		rc = __fg_driver->get_event(type,key,x,y,buttons);
#ifdef FG_MOUSEKEYS
		if (type == KEYEVENT)
		{
			static int last=0, accel=2, cbut=0;
			int doit=0;

			if (last != key)	// reset accel
				accel = 2;
				
			switch(key)
			{
				case FG_MOUSEKEYS_UP:
					type = MOUSEEVENT;
					y = yy-accel;
					x = xx;
					doit = 1;
					break;
				case FG_MOUSEKEYS_DOWN:
					type = MOUSEEVENT;
					y = yy+accel;
					x = xx;
					doit = 1;
					break;
				case FG_MOUSEKEYS_LEFT:
					type = MOUSEEVENT;
					x = xx-accel;
					y = yy;
					doit = 1;
					break;
				case FG_MOUSEKEYS_RIGHT:
					type = MOUSEEVENT;
					x = xx+accel;
					y = yy;
					doit = 1;
					break;
				case FG_MOUSEKEYS_BUTTON1:
					type = MOUSEEVENT;
					x = xx;
					y = yy;
					cbut ^= 1;
					doit = 1;
					break;
				case FG_MOUSEKEYS_BUTTON2:
					type = MOUSEEVENT;
					x = xx;
					y = yy;
					cbut ^= 2;
					doit = 1;
					break;
			}
			if (doit)
			{
				if (accel<16)		// max accel diff
					accel += 2;
				// round up co-ordinates
				if (x<0)
					x = 0;
				if (x>GetXRes()-1)
					x = (GetXRes()-1);
				if (y<0)
					y = 0;
				if (y>(GetYRes()-1))
					y = (GetYRes()-1);
				buttons = cbut;
			}
			last = key;
		}
#endif
		if (rc==0)
		{
			sched_yield();
		}
		else
		{
			if (type==KEYEVENT && key==PRTRSC)
			{
				SaveScreen();
				continue;
			} // PrtScr
#if defined(__linux__) || defined(__rtems__) || defined(_WIN32)
			else if (type==KEYEVENT && key==0x0003 && ctrl_break==0)
			{
				raise(SIGINT);
				exit(2);
			}
#endif // LINUX
			val= this;
			xx = x;
			yy = y;
			break;
		}
		IntCounter ++;
		if (waits != 1)
		{
			App::CallDelayProc();			// call delay handler
			App::Timer();                   // call timer handler
			if (repeat_status>0)
				AutoRepeatDo();
		}
	}
//	printf("typ %d, key %x, x %d, y %d b %d\n", type,key,x,y,buttons);
	if (mouseflag == 1)
		vector_mouse_cursor(xx2, yy2,FALSE);
	if (type ==	KEYEVENT)
		mouseflag =	0;
	else
		mouseflag =	1;
	return val;
}

int RemoveMousePointer(void)
{
	if (mouseflag == 1)
		vector_mouse_cursor(mx, my, FALSE);
	mouseflag =	0;
	return mouseflag;
}

void UpdateMousePointer(void)
{
	if (mouseflag == 1)
		vector_mouse_cursor(mx, my, TRUE);
}

void ShowMousePointer(void)
{
	mouseflag =	1;
	vector_mouse_cursor(mx, my, TRUE);
}

static void	blue_rect(int x, int y,	int	w, int h)
{
	int	a =	set_ppop(_GNOT);
	draw_pattern_box(x,y, w-1, h-1, &__patt);
	set_ppop(a);
}

static void	blue_rect2(int x, int y, int w, int h)
{
	if (!DragShape) blue_rect(x,y,w,h);
	else DragShape(x,y,w,h);
}

int isgmkey(int	k)
{
	Window *c =	Window::GetCurrent();

	switch (k)
	{
		case CTRL_TAB:			// ctrl+TAB
	 		if (c) Window::nextwnd();
			return 1;
		case ALT_X:			// alt+X
			if (cApp->flags	& APP_ENABLEALTX)
			{
				App::AppDone();
				return 1;
			}
			else return	0;
		case TAB:
			if (c) if (c->GetStatus() & WNOTEBOOK)
			{
				c->SetNextTabPage();
			}
			break;
		case ESC:
			if (c) if (!(c->GetStatus() & WESCAPE))
				break;
		case ALT_F04:			// alt+F4
			if (c)
				if (!Window::TestNullHandler())
					delete c;
			return 1;
		case ALT_F02:			// alt+F12
			if (c) 
			{
				if (c->Window::WindowIconize()) return 1;
			}
			return 0;
	}
	return 0;
}

void FlushInput(void)
{
	if (ininput)				   // if input, force ESC
	{
		ininput->inputproc(CR);
	}
}

void DoubleClick(int& event)
{
	static int last=0, tmout=0;
	int now = FGClock();
	if (event == last && (now-tmout) <= DBLCLICKTIME)
		event += (DBLCLICKLEFTEVENT-CLICKLEFTEVENT);
	else
	{
		last = event;
		tmout = now;
	}
}

void GuiEvent::TranslateUserEvent(void)
{
	static int mousefirst, resize =	0, tmp_event,
		resizex, resizey,
		dragx, dragy,
		clickx,	clicky,	clickw,	clickh,
		vyrez=0;
	static Control *idcb;
	static Control *idc2;
	Window *n =	0, *curr	= Window::GetCurrent(), *ecurr=0;
	static Window *drag_okno;
	int	magn = cApp->flags&APP_MAGNIFIER;
    EditBox *ebox=0;
	guiType = 0;
	
	GuiEvent app(NOEVENT); // events to  application

	call_handler = 0; // reset Control handler
	if (curr)
		switch (type)
		{
			case KEYEVENT:
				if (!ininput)	//ai nieje input
				{
					if (isgmkey(key))
						return ;
					idc2 = Control::ButtonFind(this, 7);
					if (curr->IsVisible() && idc2)	//tak zisti buttony
					{
						idc2->ClickDown();
						type = ACCELEVENT;	//bude to sprava
						guiType	= idc2->GetType(); // od buttonu typ ...
						accel = idc2;
						key	= idc2->GetId();
						delay(100);
						Puk();
						idc2->ClickUp(TRUE);
						curr = Window::GetCurrent();
					}
					else // plain KEY - pass to AppProc
					{
						app.type = KEYEVENT;
						app.key  = key;
					}
				}
				else
				{
 					if (key==TAB)
					{
						ebox = curr->FindNextEditBox();
						ecurr = curr;
						ininput->inputproc(CR);
					}
					else ininput->inputproc(key);
					type = NOEVENT;
				}
				break;
			case MOUSEEVENT:
				if (mousefirst)
				{
					mousefirst = 0;
					mx = oldx =	x;
					my = oldy =	y;
					mb = oldb =	buttons;
				}
				mx = x;
				my = y;
				mb = buttons;
				tmp_event = type;
				type = NOEVENT;
				// second phase for windowresize
				if (resize && ((mb & 1)	== 1))
				{
					if (fulldrag && !(curr->GetStatus()&WFASTMOVE))
					{
						curr->WindowResize(mx -	oldx, my - oldy);
					}
					else
					{
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
						resizex	+= (mx - oldx);
						resizey	+= (my - oldy);
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
					}
					resize++;
					break;
				}
				// last phase for windowresize
				else if	(resize	&& (!mb))
				{
					if (!fulldrag || curr->GetStatus()&WFASTMOVE)
					{
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
						if (resizex	+ resizey)
						{
							curr->WindowResize(resizex,	resizey);
						}
					}
					resize = 0;
				}
				// second phase for windowmove
				else if	(drag && ((mb &	1) == 1))
				{
					if (fulldrag && !(curr->GetStatus()&WFASTMOVE))
					{
						curr->WindowMove(mx	- oldx,	my - oldy);
					}
					else
					{
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
						dragx += (mx - oldx);
						dragy += (my - oldy);
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
					}
					drag++;
				}
				// last phase for windowmove
				else if	(drag && (!mb))
				{
					if (!fulldrag || curr->GetStatus()&WFASTMOVE)
					{
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
						if (dragx +	dragy)
							curr->WindowMove(dragx,	dragy);
					}
					drag = 0;
				}
				// release mousebutton on control
				else if	(action	&& (!mb))
				{
					action = 0;
					idcb->ClickUp(TRUE);
					AutoRepeatEnd();
					curr = Window::GetCurrent();
					if (curr->IsVisible() && ((idc2->ButtonFind(this, 3)) == idcb))
					{
						type = ACCELEVENT;	//bude to sprava
						guiType	= idcb->GetType();
						accel = idcb;
					}
					else
						type = NOEVENT;
					key	= idcb->GetId();
					break;
				}
				// push mousebutton on control
				else if	(action	&& ((mb	& 1) ==	1))
				{
					int tt = idcb->GetType();
					if (curr->IsVisible() && ((idc2	= Control::ButtonFind(this,3)) != idcb) && tt!=Control::SLIDEBAR)
					{
						action = 0;
						AutoRepeatEnd();
						if (tt!=Control::POINTBUTTON && tt!=Control::CHECKBUTTON) 
						    idcb->ClickUp(FALSE);	// nevolat handler !!!!
						break;
					}
					else
					{
						idcb->ClickDown();
					}
				}
				else if	(vyrez && (!mb)) // vyrez end
				{
					if (drag_okno) // right click to wnd
					{
						startXdrag = clickx	- curr->GetX() - curr->xoff;
						startYdrag = clicky	- curr->GetY() - curr->yoff;
						endXdrag = clickw;
						endYdrag = clickh;
					}
					else
					{
						startXdrag = clickx;
						startYdrag = clicky;
						endXdrag = clickw;
						endYdrag = clickh;
					}
					if (vyrez > DRAG_TRESHOLD+1) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					if (clickw<0) {	clickx += clickw; clickw = -clickw;	}
					if (clickh<0) {	clicky += clickh; clickh = -clickh;	}
					if (drag_okno) // right click to wnd
					{
						if (vyrez >DRAG_TRESHOLD && oldb == 2)	type = DRAGRIGHTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD	&& oldb	== 2)
						{
							DoubleClick(type = CLICKRIGHTEVENT);
						}
						else if	(vyrez > DRAG_TRESHOLD && oldb == 1)
						{
							type =	DRAGLEFTEVENT;
						}
						else if	(vyrez <= DRAG_TRESHOLD	&& oldb	== 1)
						{
							DoubleClick(type = CLICKLEFTEVENT);
						}
						w =	clickw;
						h =	clickh;
						curr = drag_okno; // redirect it to right window
						x =	clickx - curr->x - curr->xoff;
						y =	clicky - curr->y - curr->yoff;

					}
					else  // right click to background
					{
						if (vyrez > DRAG_TRESHOLD && oldb == 2)	app.type = DRAGRIGHTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD && oldb	== 2)
						{
							DoubleClick(app.type = CLICKRIGHTEVENT);
							
						}
						else if	(vyrez > DRAG_TRESHOLD && oldb == 1) app.type =	DRAGLEFTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD && oldb	== 1)
						{
							DoubleClick(app.type = CLICKLEFTEVENT);
						}
						app.x =	clickx;
						app.y =	clicky;
						app.w =	clickw;
						app.h =	clickh;
					}
					vyrez =	0;
					break;
				}
				else if	(vyrez && (mb &	3))	// vyrez continue
				{
					if (vyrez == DRAG_TRESHOLD+1)	if (magn) blue_rect2(clickx,	clicky,	clickw,	clickh);
					if (vyrez > DRAG_TRESHOLD) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					clickw += (mx -	oldx);
					clickh += (my -	oldy);
					if (vyrez > DRAG_TRESHOLD) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					vyrez++;
					if (vyrez == DRAG_TRESHOLD+1)
					{
						int type2 =	mb==2?STARTDRAGRIGHTEVENT:STARTDRAGLEFTEVENT;
						if (drag_okno)
						{
							type = type2;
							x =	clickx - curr->x - curr->xoff;
							y =	clicky - curr->y - curr->yoff;
						}
						else
						{
							app.type = type2;
							app.x = clickx;
							app.y = clicky;
						}
					}
					else // mouse move only!
					{
						// when ptr is over the window, its receive relative pos
						// app proc receive EACH movement as absolute pos
						if (drag_okno) type = MOVEEVENT;
						app.type = MOVEEVENT;
   	    	            app.x = x;
               	        app.y = y;
                   	    app.key = 0;
					}
				}
				// most first push button - what next?
				else if	((!action) && (!drag) && (!vyrez) && (!resize) && (mb == 1))
				{
					int	t;

					t =	curr->TitleFind(this);
					if (curr->IsVisible()
					   && ((idcb = Control::ButtonFind(this, 7)) !=	0) && t	!= 4)
					{
						action = 1;
						idcb->ClickDown();
						AutoRepeatStart(idcb);
					}
					else if	(t)
					{
						// windowmove start
						if (t == 1 && (!oldb) && !(curr->GetStatus() & WUNMOVED))
						{
							drag = 1;
							dragx =	dragy =	0;
							if (!fulldrag || curr->GetStatus()&WFASTMOVE)
							{
								blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
							}
						}
						// close window
						else if	(t == 2)
						{
							if (!Window::TestNullHandler())
							{
								delete curr;
								return ;
							}
						}
						// hide
				  		else if	(t==3) curr->WindowIconize();
						// resize
						else if	(t == 4)
						{
							resize = 1;
							resizex	= resizey =	0;
							if (!fulldrag || curr->GetStatus()&WFASTMOVE)
							{
								blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
							}
						}
					}
					else		// tuklo sa vedla titulku
					{
						n =	Window::WindowFind(this);
						// do okna
						// prepni na toto okno ak nieje aktivne
						if ((n != curr)	&& (n))
						{
							if (!(curr->GetStatus()	& WMODAL))
							{
								n->WindowFocus();
								type = tmp_event;
								TranslateUserEvent();
								curr = Window::GetCurrent();
								if (curr != n || !n) return; // window destroyed !!
							}
							else Snd(100,70);
						}
						else
						{
							vyrez =	1;
							drag_okno =	Window::WindowFind(this); // 0-back, 1-wnd
							clickx = x;
							clicky = y;
							clickw = clickh	= 0;
//							FlushInput();
						}
					}
				}
				else if	(oldb == 0 && mb ==	2)	// right click down
				{
					vyrez =	1;
					drag_okno =	Window::WindowFind(this);  // find window for click
					if ((drag_okno != curr) && drag_okno)
					{
						if (!(curr->GetStatus()	& WMODAL))
						{
							drag_okno->WindowFocus();
							type = tmp_event;
							TranslateUserEvent();
							curr = Window::GetCurrent();
							if (curr != drag_okno || !drag_okno) return; // window destroyed !!
						}
						else Snd(100,70);
					}
					if (ininput)  // if input, force ESC
					{
						ininput->inputproc(ESC);
					}
					clickx = x;
					clicky = y;
					clickw = clickh	= 0;
				}
				else if	(mb == 0 &&	oldb ==	0)	// only mouse move
				{
					type = MOVEEVENT;
					app.type = MOVEEVENT;
    	            app.x = x;
           	        app.y = y;
               	    app.key = 0;

					idm	= Control::ButtonFind(this, 3);
					if (idm	&& idm->GetOwner() != curr)
					{
						break;
                    }
					if (idm)
					{
						key = idm->GetId();
						if (idm->GetTrigger()) idm = 0;
					}
					else if	((idw =	Window::WindowFind(this)) != 0)
					{
						key	= idw==curr?0:-1;
						curr = idw;		// sender is changed !!!
						x =	x-idw->GetX()-idw->xoff;
						y =	y-idw->GetY()-idw->yoff;
					}
					else
					{
						type = NOEVENT;
						app.type = MOVEEVENT;
						app.x = x;
						app.y = y;
						app.key = 0;
					}
					if ((!oldidm &&	idm) ||	((idm && oldidm) &&	(idm !=	oldidm)))
					  	idm->MoveIn();
					if ((oldidm	&& !idm) ||	((idm && oldidm) &&	(idm !=	oldidm)))
					  	oldidm->MoveOut();
					if ((!oldidw &&	idw) ||	((idw && oldidw) &&	(idw !=	oldidw)))
						idw->MoveIn();
					if ((oldidw	&& !idw) ||	((idw && oldidw) &&	(idw !=	oldidw)))
						oldidw->MoveOut();
					break;
				}
		}
	else
	{
		switch(type)  // send & mouse keys only
		{
			case KEYEVENT:
				if (isgmkey(key))
					break;	// exit when no window are
			case MOVEEVENT:
				App::SendToApp(this);
		}
		return ;
	}							// no window exists

	if (app.type==MOVEEVENT)
	{
		app.wnd = idw;
		accel   = idm;
    }
	
	if (!action && !drag && !vyrez && !resize && (type==MOVEEVENT || app.type==MOVEEVENT))
	{
		int value=0;
		if (oldx==mx)
		{
			if (oldx==0) value |= 1;
			else if (oldx>=(GetXRes()-1)-1) value |= 2;
		}
		if (oldy==my)
		{
			if (oldy==0) value |= 4;
			if (oldy>=(GetYRes()-1)-1) value |= 8;
		}
		if (value)
		{
			app.type = CURSOROUTEVENT;
			app.key  = value;
		}
	}
	oldx = mx;
	oldy = my;
	oldb = mb;
	oldidm = idm;
	oldidw = idw;

	// there are it's magic place where an events goes into the right place
	if (type == ACCELEVENT) app = *this;
	
	if (app.type !=	NOEVENT)
		App::SendToApp(&app); // send to app if needed

	if (type != NOEVENT)   // send to Window if needed
	{
		if (type == ACCELEVENT && call_handler)
			call_handler->RefreshGroup();
		if (type==MOVEEVENT && key==-1)
		{
			key = 0;
			if (idw) idw->SendToWindow(this);
		}
		else if ((curr=Window::GetCurrent()) != 0)
			curr->SendToWindow(this);
	}

	if (call_handler)
	{
		call_handler->RunHandler(); // call to control handler also
		// at this point, call_handler is already NULL
	}
   // switch to the next Editbox
   if (ebox && ecurr && curr==ecurr)
		ebox->ClickUp(1);
}

void DisableCtrlBreak(void)
{
	ctrl_break ^= 1;
}

FGMouseCursor *__fg_cursor=idc_normal;

int	__MouseDraw(int x, int y,	int	a)
{
	int	old	= mouseflag;
	int	i;
	FGPixel *maska, *mapa;
	static FGPixel work[sizeof(m_save)], tmp[sizeof(m_save)];

	mouseflag =	a;
	if (a) // draw cursor
	{
		VideoToRam(m_save, x-__fg_cursor->xoff, y-__fg_cursor->yoff, 0, 0, __fg_cursor->w, __fg_cursor->h, __fg_cursor->w, __fg_cursor->h);
		memcpy(work, m_save, sizeof(m_save));
		maska =	__fg_cursor->mask;
		mapa  = __fg_cursor->bitmap;
		for	(i = 0;	i <	(__fg_cursor->w * __fg_cursor->h); i++)
		{
			work[i]	&= *maska;
			work[i]	|= ((*mapa++) &	(FGPixel)(~*maska++));
		}
		RamToVideo(work, 0, 0, x-__fg_cursor->xoff, y-__fg_cursor->yoff, __fg_cursor->w, __fg_cursor->h, __fg_cursor->w, __fg_cursor->h, BITBLIT_COPY);
	}
	else // restore background
	{
		VideoToRam(tmp, x-__fg_cursor->xoff, y-__fg_cursor->yoff, 0, 0, __fg_cursor->w, __fg_cursor->h, __fg_cursor->w, __fg_cursor->h);
		if (memcmp(tmp,	work, sizeof(m_save)) != 0) // yes, changed
		{
			for (i = 0; i < (int)sizeof(m_save)/bpp; i++)
			{
				if (tmp[i] != work[i])
				{
					m_save[i] = tmp[i];
				}
			}
		}
		RamToVideo(m_save, 0, 0, x-__fg_cursor->xoff, y-__fg_cursor->yoff, __fg_cursor->w, __fg_cursor->h, __fg_cursor->w, __fg_cursor->h, BITBLIT_COPY);
	}
	return old;
}

FGMouseCursor * App::CursorLoad(FGMouseCursor *cur)
{
	if (cur == __fg_cursor)
		return cur;
	FGMouseCursor *old = __fg_cursor;
	int oldf = RemoveMousePointer();
	__fg_cursor = cur;
	if (oldf)
		ShowMousePointer();
	return old;
}


