// ---------------------------------------------------------------------------
//
//		Windows - DirectDraw frontend driver
//
// ---------------------------------------------------------------------------

#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <ddraw.h>
#include "dll.h"

/*============================================================================
 *
 * DirectDraw Structures
 *
 * Various structures used to invoke DirectDraw.
 *
 *==========================================================================*/


class FGWin32Driver : public FGDriver
{
	protected:
		static HWND         hwnd;
		static HINSTANCE	hInstance;			// Instance handle
		static int win_cursor(int,int,int a)
		{
			return !a;
		}
	public:
    	FGWin32Driver(char *name) : FGDriver(name)
        {
        }
		virtual int get_event(int& type, int& key, int& x, int& y, int& buttons);
		virtual int available(void)
		{
			return 1;
		}
		virtual int set_mouse(void)
		{
			mouse_string = "WINDOWS_MOUSE";
			return TRUE;
		}
		virtual void change_mouse(void)
		{
		}
		virtual void reset_mouse(void)
		{
			ismouse = 0;
		}
		virtual int set_keyboard(void)
		{
			return TRUE;
		}
		virtual void reset_keyboard(void)
		{
			iskeyboard = 0;
		}
};

class FGWin32GDIDriver : public FGWin32Driver
{
		DEVMODE dmode;
	   	HDC     hdc;
	   	HDC     memhdc;
		PAINTSTRUCT ps;
		PAINTSTRUCT memps;
        BITMAP 	bdata;
	public:
    	FGWin32GDIDriver(char *name);
    	~FGWin32GDIDriver();
		virtual int link(void)
		{
			return 1;
		}
		virtual void get_all_modes(vmode *mode) // default
		{
			total_modes = 1;		// set only one virtual videomode [window with this size]
		}
		virtual int set_mode(int ww, int hh);
};

class FGWin32DXDriver : public FGWin32Driver
{
		friend class FGdx;
		typedef
#ifndef _MSC_VER
			HRESULT
#endif
			WINAPI
			(*tDrawCreate)
			(GUID FAR *lpGUID,
			LPDIRECTDRAW FAR *lplpDD,
			IUnknown FAR *pUnkOuter );
		DDSURFACEDESC       desc;
		DDSCAPS             ddscaps;
		tDrawCreate			dDrawCreate;
		Dll					*DDrawDLL;
		LPDIRECTDRAW2		ddraw2;

		int					DX5_Load(void);
		void				DX5_Unload(void);
		int					DX5_UpdateVideoInfo(void);
		virtual int detect_hwaccel(void)
		{
			return FG_DX;
		}

	protected:
		static HRESULT WINAPI EnumModes2(DDSURFACEDESC *desc, VOID *udata);
		static LPDIRECTDRAWPALETTE palette;
		virtual void get_all_modes(vmode *) // default
		{
			int result;
			if(FAILED(result=ddraw2->EnumDisplayModes(0, NULL, this,EnumModes2)))
				SetDDerror("EnumDisplayModes", result);
		}
		static void fillbox(int x, int y, int w, int h, FGPixel color, unsigned ppop)
		{
			LPDIRECTDRAWSURFACE3 dst_surface;
			RECT area;
			DDBLTFX bltfx;
			HRESULT result;
			if (ppop) // fallback
			{
				__fill_box(x,y,w,h,color,ppop);
				return;
			}
#ifdef DDRAW_DEBUG
			fprintf(stderr, "HW accelerated fill at (%d,%d)\n", dstrect->x, dstrect->y);
#endif
			area.top = y;
			area.bottom = y+h;
			area.left = x;
			area.right = x+w;
			bltfx.dwSize = sizeof(bltfx);
			bltfx.dwFillColor = color;
			result = IDirectDrawSurface3_Blt(FGWin32DXDriver::current,
				&area, NULL, NULL, DDBLT_COLORFILL|DDBLT_WAIT, &bltfx);
			if ( result == DDERR_SURFACELOST )
			{
				IDirectDrawSurface3_Restore(FGWin32DXDriver::primary);
				result = IDirectDrawSurface3_Blt(FGWin32DXDriver::current,
					&area, NULL, NULL, DDBLT_COLORFILL|DDBLT_WAIT, &bltfx);
			}
			if ( result != DD_OK )
			{
				FGWin32DXDriver::SetDDerror("IDirectDrawSurface3::Blt", result);
			}
		}
	public:
		static LPDIRECTDRAWSURFACE primary;
		static LPDIRECTDRAWSURFACE secondary;
		static LPDIRECTDRAWSURFACE current;

		static 				void SetDDerror(const char *function, int code);
		static 				void SetDIerror(const char *function, int code);
		static long HandleMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
		FGWin32DXDriver(char *name);
		~FGWin32DXDriver();
		virtual int text_mode(void)
		{
			return TRUE;
		}
		virtual int	link(void);
		virtual int set_mode(int ww, int hh);
		virtual int EnableBuffering(int);
		virtual int set_page(int);
#ifdef INDEX_COLORS
		static void dx_palette(unsigned col, unsigned rgb)
		{
			unsigned rgb2 = ((rgb>>16)&0xFF) + (rgb&0xFF00) + ((rgb<<16)&0xFF0000);
			palette->SetEntries(0, col, 1,(PALETTEENTRY *)&rgb2);
		}
#endif
		virtual void DisableBuffering(void)
		{
			current_buffer=0;
			bufnum = 0;
			::videobase = videobase = buffer[3];
			current = primary;
			offset = 0;
		}
};

